/*
 * TikZConverter.h
 *
 *  Created on: Dec 1, 2015
 *      Author: Jan Travnicek
 */

#ifndef TIKZ_CONVERTER_H_
#define TIKZ_CONVERTER_H_

#include <ostream>
#include <set>
#include <list>
#include <vector>
#include <typeinfo>
#include <map>
#include <utility>

#include <core/multipleDispatch.hpp>
#include <exception/CommonException.h>

#include <automaton/Automaton.h>
#include <automaton/AutomatonFeatures.h>

#include <string/String.h>

#include <automaton/FSM/NFA.h>
#include <automaton/FSM/EpsilonNFA.h>
#include <automaton/FSM/MultiInitialStateNFA.h>
#include <automaton/FSM/DFA.h>
#include <automaton/FSM/ExtendedNFA.h>
#include <automaton/FSM/CompactNFA.h>
#include <automaton/TA/NFTA.h>
#include <automaton/TA/DFTA.h>
#include <automaton/PDA/NPDA.h>
#include <automaton/PDA/SinglePopNPDA.h>
#include <automaton/PDA/InputDrivenDPDA.h>
#include <automaton/PDA/InputDrivenNPDA.h>
#include <automaton/PDA/VisiblyPushdownDPDA.h>
#include <automaton/PDA/VisiblyPushdownNPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicDPDA.h>
#include <automaton/PDA/RealTimeHeightDeterministicNPDA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/PDA/SinglePopDPDA.h>
#include <automaton/TM/OneTapeDTM.h>

#include <factory/StringDataFactory.hpp>

#include "common/converterCommon.hpp"

class TikZConverter : public std::SingleDispatchFirstStaticParam < TikZConverter, void, std::ostream &, const automaton::AutomatonBase & > {
	template < class SymbolType, class StateType >
	static void transitions(const automaton::DFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out);

	template < class SymbolType, class StateType >
	static void transitions(const automaton::NFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out);

	template < class SymbolType, class StateType >
	static void transitions(const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out);

	template < class SymbolType, class EpsilonType, class StateType >
	static void transitions(const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out);

	template < class SymbolType, class StateType >
	static void transitions(const automaton::ExtendedNFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out);

	template < class SymbolType, class StateType >
	static void transitions(const automaton::CompactNFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out);

	template < class SymbolType, class RankType, class StateType >
	static void transitions(const automaton::NFTA < SymbolType, RankType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out);

	template < class SymbolType, class RankType, class StateType >
	static void transitions(const automaton::DFTA < SymbolType, RankType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::NPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & tm, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::SinglePopNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & tm, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::InputDrivenDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::InputDrivenNPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::VisiblyPushdownDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & tm, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::VisiblyPushdownNPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & tm, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & tm, const std::map < StateType, int > & states, std::ostream & out);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void transitions(const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & tm, const std::map < StateType, int > & states, std::ostream & out);

	template < class SymbolType, class StateType >
	static void transitions(const automaton::OneTapeDTM < SymbolType, StateType > & tm, const std::map < StateType, int > & states, std::ostream & out);
public:
	static void convert ( std::ostream & out, const automaton::Automaton & a );

	template < class SymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::DFA < SymbolType, StateType > & a);

	template < class SymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::NFA < SymbolType, StateType > & a);

	template < class SymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::MultiInitialStateNFA < SymbolType, StateType >& a);

	template < class SymbolType, class EpsilonType, class StateType >
	static void convert(std::ostream& out, const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & a);

	template < class SymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::ExtendedNFA < SymbolType, StateType > & a);

	template < class SymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::CompactNFA < SymbolType, StateType > & a);

	template < class SymbolType, class RankType, class StateType >
	static void convert(std::ostream& out, const automaton::NFTA < SymbolType, RankType, StateType > & a);

	template < class SymbolType, class RankType, class StateType >
	static void convert(std::ostream& out, const automaton::DFTA < SymbolType, RankType, StateType > & a);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::NPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::SinglePopNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::InputDrivenDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::InputDrivenNPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::VisiblyPushdownDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::VisiblyPushdownNPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a);

	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a);

	template < class SymbolType, class StateType >
	static void convert(std::ostream& out, const automaton::OneTapeDTM < SymbolType, StateType > & a);
};

template<class SymbolType, class EpsilonType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::MultiInitialStateNFA < SymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialStates ( ).count ( state.first ) )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::NFA < SymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::DFA < SymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::ExtendedNFA < SymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::CompactNFA < SymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class RankType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::NFTA < SymbolType, RankType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class RankType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::DFTA < SymbolType, RankType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::InputDrivenDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::InputDrivenNPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::VisiblyPushdownDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::VisiblyPushdownNPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialStates ( ).count ( state.first ) )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialStates ( ).count ( state.first ) )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::NPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::SinglePopNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class StateType>
void TikZConverter::convert ( std::ostream & out, const automaton::OneTapeDTM < SymbolType, StateType > & a ) {
	out << "\\begin{tikzpicture}\n";
	int cnt = 1;

	 // Map states to indices
	std::map < StateType, int > states;

	for ( const StateType & state : a.getStates ( ) )
		states.insert ( std::make_pair ( state, cnt++ ) );

	 // Print states
	for ( const auto & state : states ) {
		std::string mods;

		if ( a.getFinalStates ( ).count ( state.first ) )
			mods += ",accepting";

		if ( a.getInitialState ( ) == state.first )
			mods += ",initial";

		out << "\\node[state" + mods + "] (" << state.second << ") {" << replace ( alib::StringDataFactory::toString (  state.first  ), "\"", "\\\"" ) << "}\n";
	}

	transitions ( a, states, out );
	out << "\\end{tikzpicture}";
}

template<class SymbolType, class EpsilonType, class StateType>
void TikZConverter::transitions ( const automaton::EpsilonNFA < SymbolType, EpsilonType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	 // put transitions from automaton to "transitions"
	for ( const auto & transition : fsm.getTransitions ( ) ) {
		std::string symbol;

		if ( transition.first.second.template is < EpsilonType > ( ) )
			symbol = "&epsilon;";
		else
			symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ).template get < SymbolType > ( ) ), "\"", "\\\"" );

		for ( const StateType & to : transition.second ) {
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class SymbolType, class StateType>
void TikZConverter::transitions ( const automaton::MultiInitialStateNFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	 // put transitions from automaton to "transitions"
	for ( const auto & transition : fsm.getTransitions ( ) ) {
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		for ( const StateType & to : transition.second ) {
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class SymbolType, class StateType>
void TikZConverter::transitions ( const automaton::NFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	 // put transitions from automaton to "transitions"
	for ( const auto & transition : fsm.getTransitions ( ) ) {
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		for ( const StateType & to : transition.second ) {
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class SymbolType, class StateType>
void TikZConverter::transitions ( const automaton::DFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	 // put transitions from automaton to "transitions"
	for ( const auto & transition : fsm.getTransitions ( ) ) {
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( transition.second )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class SymbolType, class StateType>
void TikZConverter::transitions ( const automaton::ExtendedNFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	 // put transitions from automaton to "transitions"
	for ( const auto & transition : fsm.getTransitions ( ) ) {
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		for ( const StateType & to : transition.second ) {
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class SymbolType, class StateType>
void TikZConverter::transitions ( const automaton::CompactNFA < SymbolType, StateType > & fsm, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	 // put transitions from automaton to "transitions"
	for ( const auto & transition : fsm.getTransitions ( ) ) {
		std::string symbol = replace ( alib::StringDataFactory::toString (  string::stringFrom ( transition.first.second  ) ), "\"", "\\\"" );

		for ( const StateType & to : transition.second ) {
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class SymbolType, class RankType, class StateType>
void TikZConverter::transitions ( const automaton::NFTA < SymbolType, RankType, StateType > & fta, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, std::vector < int > >, std::string > transitions;

	 // put transitions from automaton to "transitions"
	for ( const auto & transition : fta.getTransitions ( ) ) {
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.first.getSymbol (  ) ), "\"", "\\\"" );
		symbol += std::to_string ( transition.first.first.getRank ( ) );

		for ( const StateType & to : transition.second ) {
			std::pair < int, std::vector < int > > key ( states.find ( to )->second, { } );

			for ( const StateType & state : transition.first.second )
				key.second.push_back ( states.find ( state )->second );

			std::map < std::pair < int, std::vector < int > >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol;
			}
		}
	}

	 // Print auxilary dots
	for ( unsigned i = 1; i < transitions.size ( ); i++ )
		out << "\\node[draw=none,fill=none] (" << states.size ( ) + i << ") {}\n";

	out << "\\path[->]";

	 // print the map
	unsigned i = states.size ( ) + 1;

	for ( std::pair < const std::pair < int, std::vector < int > >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << i << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.first << ")\n";

		unsigned j = 0;

		for ( int from : transition.first.second ) {
			out << "(" << from << ") edge [left] node [align=center] ";
			out << "{$" << j << "$}";
			out << "(" << i << ")\n";

			j++;
		}

		i++;
	}
}

template<class SymbolType, class RankType, class StateType>
void TikZConverter::transitions ( const automaton::DFTA < SymbolType, RankType, StateType > & fta, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, std::vector < int > >, std::string > transitions;

	 // put transitions from automaton to "transitions"
	for ( const auto & transition : fta.getTransitions ( ) ) {
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.first.getSymbol (  ) ), "\"", "\\\"" );
		symbol += std::to_string ( transition.first.first.getRank ( ) );

		std::pair < int, std::vector < int > > key ( states.find ( transition.second )->second, { } );

		for ( const StateType & state : transition.first.second )
			key.second.push_back ( states.find ( state )->second );

		std::map < std::pair < int, std::vector < int > >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	 // Print auxilary dots
	for ( unsigned i = 1; i < transitions.size ( ); i++ )
		out << "\\node[draw=none,fill=none] (" << states.size ( ) + i << ") {}\n";

	out << "\\path[->]";

	 // print the map
	unsigned i = states.size ( ) + 1;

	for ( std::pair < const std::pair < int, std::vector < int > >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << i << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.first << ")\n";

		unsigned j = 0;

		for ( int from : transition.first.second ) {
			out << "(" << from << ") edge [left] node [align=center] ";
			out << "{$" << j << "$}";
			out << "(" << i << ")\n";

			j++;
		}

		i++;
	}
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : pda.getTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( std::get < 1 > ( transition.first ).template is < EpsilonType > ( ) )
			symbol = "&epsilon;";
		else
			symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ).template get < InputSymbolType > ( ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		if ( std::get < 2 > ( transition.first ).size ( ) == 0 )
			symbol += " &epsilon;";
		else
			for ( const PushdownStoreSymbolType & symb : std::get < 2 > ( transition.first ) )
				symbol += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

		symbol += " ->";

		 // Push part
		if ( transition.second.second.size ( ) == 0 )
			symbol += " &epsilon;";
		else
			for ( const PushdownStoreSymbolType & symb : transition.second.second )
				symbol += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

		 // Insert into map
		std::pair < int, int > key ( states.find ( std::get < 0 > ( transition.first ) )->second, states.find ( transition.second.first )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::SinglePopDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : pda.getTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( std::get < 1 > ( transition.first ).template is < EpsilonType > ( ) )
			symbol = "&epsilon;";
		else
			symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ).template get < InputSymbolType > ( ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " " + replace ( alib::StringDataFactory::toString (  std::get < 2 > ( transition.first  ) ), "\"", "\\\"" );

		symbol += " ->";

		 // Push part
		if ( transition.second.second.size ( ) == 0 )
			symbol += " &epsilon;";
		else
			for ( const PushdownStoreSymbolType & symb : transition.second.second )
				symbol += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

		 // Insert into map
		std::pair < int, int > key ( states.find ( std::get < 0 > ( transition.first ) )->second, states.find ( transition.second.first )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::InputDrivenDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	const auto & symbolToPDSOperation = pda.getPushdownStoreOperations ( );

	for ( const auto & transition : pda.getTransitions ( ) ) {
		const auto & pop = symbolToPDSOperation.find ( transition.first.second )->second.first;
		const auto & push = symbolToPDSOperation.find ( transition.first.second )->second.second;

		 // input symbol
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		if ( pop.size ( ) == 0 )
			symbol += " &epsilon;";
		else
			for ( const PushdownStoreSymbolType & symb : pop )
				symbol += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

		symbol += " ->";

		const auto & to = transition.second;

		 // Push part
		if ( push.size ( ) == 0 )
			symbol += " &epsilon;";
		else
			for ( const PushdownStoreSymbolType & symb : push )
				symbol += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

		 // Insert into map
		std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::InputDrivenNPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	const auto & symbolToPDSOperation = pda.getPushdownStoreOperations ( );

	for ( const auto & transition : pda.getTransitions ( ) ) {
		const auto & pop = symbolToPDSOperation.find ( transition.first.second )->second.first;
		const auto & push = symbolToPDSOperation.find ( transition.first.second )->second.second;

		 // input symbol
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		if ( pop.size ( ) == 0 )
			symbol += " &epsilon;";
		else
			for ( const PushdownStoreSymbolType & symb : pop )
				symbol += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2 = symbol;

			 // Push part
			if ( push.size ( ) == 0 )
				symbol2 += " &epsilon;";
			else
				for ( const PushdownStoreSymbolType & symb : push )
					symbol2 += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

			 // Insert into map
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::VisiblyPushdownDPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : pda.getCallTransitions ( ) ) {
		 // input symbol
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		symbol += " " + replace ( alib::StringDataFactory::toString (  transition.second.second  ), "\"", "\\\"" );

		 // Insert into map
		std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( transition.second.first )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	for ( const auto & transition : pda.getReturnTransitions ( ) ) {
		 // input symbol
		std::string symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " " + replace ( alib::StringDataFactory::toString (  std::get < 2 > ( transition.first  ) ), "\"", "\\\"" );
		symbol += " ->";

		symbol += " &epsilon;";

		 // Insert into map
		std::pair < int, int > key ( states.find ( std::get < 0 > ( transition.first ) )->second, states.find ( transition.second )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	for ( const auto & transition : pda.getLocalTransitions ( ) ) {
		 // input symbol
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		symbol += " &epsilon;";

		 // Insert into map
		std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( transition.second )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::VisiblyPushdownNPDA < InputSymbolType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : pda.getCallTransitions ( ) ) {
		 // input symbol
		std::string symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2	 = symbol;
			symbol2 += " " + replace ( alib::StringDataFactory::toString (  to.second  ), "\"", "\\\"" );

			 // Insert into map
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to.first )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	for ( const auto & transition : pda.getReturnTransitions ( ) ) {
		 // input symbol
		std::string symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " " + replace ( alib::StringDataFactory::toString (  std::get < 2 > ( transition.first  ) ), "\"", "\\\"" );
		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2	 = symbol;
			symbol2 += " &epsilon;";

			 // Insert into map
			std::pair < int, int > key ( states.find ( std::get < 0 > ( transition.first ) )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	for ( const auto & transition : pda.getLocalTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		symbol = replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2	 = symbol;
			symbol2 += " &epsilon;";

			 // Insert into map
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::RealTimeHeightDeterministicDPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : pda.getCallTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( transition.first.second.template is < EpsilonType > ( ) )
			symbol = "&epsilon";
		else
			symbol = replace ( alib::StringDataFactory::toString (  transition.first.second.template get < InputSymbolType > (  ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		symbol += " " + replace ( alib::StringDataFactory::toString (  transition.second.second  ), "\"", "\\\"" );

		 // Insert into map
		std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( transition.second.first )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	for ( const auto & transition : pda.getReturnTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( std::get < 1 > ( transition.first ).template is < EpsilonType > ( ) )
			symbol = "&epsilon";
		else
			symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ).template get < InputSymbolType > ( ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " " + replace ( alib::StringDataFactory::toString (  std::get < 2 > ( transition.first  ) ), "\"", "\\\"" );
		symbol += " ->";

		symbol += " &epsilon;";

		 // Insert into map
		std::pair < int, int > key ( states.find ( std::get < 0 > ( transition.first ) )->second, states.find ( transition.second )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	for ( const auto & transition : pda.getLocalTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( transition.first.second.template is < EpsilonType > ( ) )
			symbol = "&epsilon";
		else
			symbol = replace ( alib::StringDataFactory::toString (  transition.first.second.template get < InputSymbolType > (  ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		symbol += " &epsilon;";

		 // Insert into map
		std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( transition.second )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : pda.getCallTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( transition.first.second.template is < EpsilonType > ( ) )
			symbol = "&epsilon";
		else
			symbol = replace ( alib::StringDataFactory::toString (  transition.first.second.template get < InputSymbolType > (  ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2 = symbol;
			symbol2 += " " + replace ( alib::StringDataFactory::toString (  to.second  ), "\"", "\\\"" );

			 // Insert into map
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to.first )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	for ( const auto & transition : pda.getReturnTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( std::get < 1 > ( transition.first ).template is < EpsilonType > ( ) )
			symbol = "&epsilon;";
		else
			symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ).template get < InputSymbolType > ( ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " " + replace ( alib::StringDataFactory::toString (  std::get < 2 > ( transition.first  ) ), "\"", "\\\"" );
		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2 = symbol;
			symbol2 += " &epsilon;";

			 // Insert into map
			std::pair < int, int > key ( states.find ( std::get < 0 > ( transition.first ) )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	for ( const auto & transition : pda.getLocalTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( transition.first.second.template is < EpsilonType > ( ) )
			symbol = "&epsilon;";
		else
			symbol = replace ( alib::StringDataFactory::toString (  transition.first.second.template get < InputSymbolType > (  ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " &epsilon;";
		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2 = symbol;
			symbol2 += " &epsilon;";

			 // Insert into map
			std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( to )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::NPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : pda.getTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( std::get < 1 > ( transition.first ).template is < EpsilonType > ( ) )
			symbol = "&epsilon;";
		else
			symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ).template get < InputSymbolType > ( ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		if ( std::get < 2 > ( transition.first ).size ( ) == 0 )
			symbol += " &epsilon;";
		else
			for ( const PushdownStoreSymbolType & symb : std::get < 2 > ( transition.first ) )
				symbol += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2 = symbol;

			 // Push part
			if ( to.second.size ( ) == 0 )
				symbol2 += " &epsilon;";
			else
				for ( const PushdownStoreSymbolType & symb : to.second )
					symbol2 += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

			 // Insert into map
			std::pair < int, int > key ( states.find ( std::get < 0 > ( transition.first ) )->second, states.find ( to.first )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType>
void TikZConverter::transitions ( const automaton::SinglePopNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : pda.getTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		if ( std::get < 1 > ( transition.first ).template is < EpsilonType > ( ) )
			symbol = "&epsilon;";
		else
			symbol = replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.first  ).template get < InputSymbolType > ( ) ), "\"", "\\\"" );

		symbol += " |";

		 // Pop part
		symbol += " " + replace ( alib::StringDataFactory::toString (  std::get < 2 > ( transition.first  ) ), "\"", "\\\"" );

		symbol += " ->";

		std::string symbol2;

		for ( const auto & to : transition.second ) {
			symbol2 = symbol;

			 // Push part
			if ( to.second.size ( ) == 0 )
				symbol2 += " &epsilon;";
			else
				for ( const PushdownStoreSymbolType & symb : to.second )
					symbol2 += " " + replace ( alib::StringDataFactory::toString (  symb  ), "\"", "\\\"" );

			 // Insert into map
			std::pair < int, int > key ( states.find ( std::get < 0 > ( transition.first ) )->second, states.find ( to.first )->second );
			std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

			if ( mapit == transitions.end ( ) ) {
				transitions.insert ( std::make_pair ( key, symbol2 ) );
			} else {
				mapit->second += ",";

				size_t pos = mapit->second.find_last_of ( "\n" );

				if ( pos == std::string::npos ) pos = 0;

				if ( mapit->second.size ( ) - pos > 100 )
					mapit->second += "\n";
				else
					mapit->second += " ";

				mapit->second += symbol2;
			}
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

template<class SymbolType, class StateType>
void TikZConverter::transitions ( const automaton::OneTapeDTM < SymbolType, StateType > & tm, const std::map < StateType, int > & states, std::ostream & out ) {
	std::map < std::pair < int, int >, std::string > transitions;

	for ( const auto & transition : tm.getTransitions ( ) ) {
		std::string symbol;

		 // input symbol
		symbol	= "(";
		symbol += replace ( alib::StringDataFactory::toString (  transition.first.second  ), "\"", "\\\"" );
		symbol += ", ";
		symbol += replace ( alib::StringDataFactory::toString (  std::get < 1 > ( transition.second  ) ), "\"", "\\\"" );
		symbol += " ";

		switch ( std::get < 2 > ( transition.second ) ) {
		case automaton::Shift::LEFT:
			symbol += "&larr;";
			break;

		case automaton::Shift::RIGHT:
			symbol += "&rarr;";
			break;

		case automaton::Shift::NONE:
			symbol += "&times;";
			break;

		default:
			throw exception::CommonException ( "Unexpected shift direction" );
		}

		 // Insert into map
		std::pair < int, int > key ( states.find ( transition.first.first )->second, states.find ( std::get < 0 > ( transition.second ) )->second );
		std::map < std::pair < int, int >, std::string >::iterator mapit = transitions.find ( key );

		if ( mapit == transitions.end ( ) ) {
			transitions.insert ( std::make_pair ( key, symbol ) );
		} else {
			mapit->second += ",";

			size_t pos = mapit->second.find_last_of ( "\n" );

			if ( pos == std::string::npos ) pos = 0;

			if ( mapit->second.size ( ) - pos > 100 )
				mapit->second += "\n";
			else
				mapit->second += " ";

			mapit->second += symbol;
		}
	}

	out << "\\path[->]";

	 // print the map
	for ( std::pair < const std::pair < int, int >, std::string > & transition : transitions ) {
		replaceInplace ( transition.second, "\n", "\\n" );

		out << "(" << transition.first.first << ") edge [left] node [align=center] ";
		out << "{$" << transition.second << "$}";
		out << "(" << transition.first.second << ")\n";
	}
}

#endif /* TIKZ_CONVERTER_H_ */
