/*
 * Normalize.h
 *
 *  Created on: Dec 9, 2013
 *	  Author: Jan Travnicek
 */

#ifndef NORMALIZE_H_
#define NORMALIZE_H_

#include <core/multipleDispatch.hpp>
#include <automaton/FSM/DFA.h>
#include <automaton/PDA/DPDA.h>
#include <automaton/Automaton.h>

namespace automaton {

namespace simplify {

class Normalize : public std::SingleDispatch<Normalize, automaton::Automaton, const automaton::AutomatonBase &> {
public:
	/**
	 * @param dfa automaton to normalize
	 */
	static automaton::Automaton normalize(const automaton::Automaton& dfa);

	template < class SymbolType, class StateType >
	static automaton::DFA < SymbolType, unsigned > normalize(const automaton::DFA < SymbolType, StateType > & dfa);
	template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
	static automaton::DPDA < InputSymbolType, EpsilonType, unsigned, unsigned > normalize(const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & dfa);
};

template < class SymbolType, class StateType >
automaton::DFA < SymbolType, unsigned > Normalize::normalize(const automaton::DFA < SymbolType, StateType > & fsm) {
	unsigned counter = 0;
	std::map < StateType, unsigned > normalizationData;
	std::deque < StateType > processingData;

	normalizationData.insert(std::make_pair(fsm.getInitialState(), counter++));
	processingData.push_back(fsm.getInitialState());

	while(!processingData.empty()) {
		StateType current = std::move ( processingData.front() );
		processingData.pop_front();

		// Transitions are trivialy sorted by input symbol (from state is the same)
		for(const std::pair < const std::pair < StateType, SymbolType >, StateType > & transition : fsm.getTransitionsFromState(current) ) {
			if(normalizationData.find(transition.second) == normalizationData.end()) {
				normalizationData.insert ( std::make_pair ( transition.second, counter++ ) );
				processingData.push_back ( transition.second );
			}
		}
	}

	if(normalizationData.size() != fsm.getStates().size()) {
		throw exception::CommonException("Automaton normalize require minimal deterministic finite automaton");
	}

	automaton::DFA < SymbolType, unsigned > result( normalizationData.find ( fsm.getInitialState ( ) )->second );

	result.setInputAlphabet(fsm.getInputAlphabet());

	for(const StateType & state : fsm.getStates ( ) ) {
		result.addState ( normalizationData.find ( state )->second );
	}

	for(const StateType & finalState : fsm.getFinalStates ( ) ) {
		result.addFinalState( normalizationData.find ( finalState )->second );
	}

	for(const std::pair < const std::pair < StateType, SymbolType >, StateType > & transition : fsm.getTransitions ( ) ) {
		result.addTransition ( normalizationData.find ( transition.first.first )->second, transition.first.second, normalizationData.find ( transition.second )->second );
	}

	return result;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
automaton::DPDA < InputSymbolType, EpsilonType, unsigned, unsigned > Normalize::normalize(const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & pda) {
	unsigned counterState = 0;
	std::map< StateType, unsigned > normalizationDataState;
	unsigned counterSymbol = 0;
	std::map < InputSymbolType, unsigned > normalizationDataSymbol;
	std::deque < StateType > processingData;

	normalizationDataState.insert(std::make_pair(pda.getInitialState(), counterState++));
	normalizationDataSymbol.insert(std::make_pair(pda.getInitialSymbol(), counterSymbol++));
	processingData.push_back(pda.getInitialState());

	while(!processingData.empty()) {
		StateType current = std::move ( processingData.front() );
		processingData.pop_front();

		std::map < std::pair < std::variant < EpsilonType, InputSymbolType >, std::vector < unsigned > >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > transform;
		bool stateFinished = true;
		// For each transition from state current
		for ( const std::pair < const std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, std::vector < PushdownStoreSymbolType > >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > & iter : pda.getTransitionsFromState(current)) {
			// look whether all poped symbols are already transformed
			if(std::all_of(std::get<2>(iter.first).begin(), std::get<2>(iter.first).end(), [&](const PushdownStoreSymbolType& symbol) { return normalizationDataSymbol.find(symbol) != normalizationDataSymbol.end(); })) {
				std::vector < unsigned > transformedSymbols;
				// if so compute vector of transformed poped symbol -- this can be compared to other vectors of transformed symbols and the order can by trusted
				for(const PushdownStoreSymbolType& symbol : std::get<2> ( iter.first ) ) {
					transformedSymbols.push_back(normalizationDataSymbol.find(symbol)->second);
				}

				// then first order is by input symbol and the second is this vector transfomed symbols 
				transform.insert(std::make_pair(std::make_pair(std::get<1>(iter.first), transformedSymbols), iter.second));
			} else {
				stateFinished = false;
			}
		}
		// if there was a single transition with undefined transformation for poped symbol process this state again later
		if(!stateFinished) {
			processingData.push_back(std::move ( current ) );
		}
		// now transitions are trivialy sorted by input symbol and the already transformed symbols

		// for each pair state, pushed symbols in order given by input symbol and transformed poped symbols
		for(const auto& iter : transform) {
			const auto& iters = iter.second;
			// if the state is new asign a unique number to it and mark it for processing
			if( normalizationDataState.find(iters.first) == normalizationDataState.end() ) {
				normalizationDataState.insert(std::make_pair(iters.first, counterState++));
				processingData.push_back(iters.first);
			}
			// if the symbols in order given by order of pushing are new asign a unique number to them
			for(const PushdownStoreSymbolType & iter2 : iters.second) {
				if( normalizationDataSymbol.find(iter2) == normalizationDataSymbol.end() ) {
					normalizationDataSymbol.insert(std::make_pair(iter2, counterSymbol++));
				}
			}
		}
	}

	if(normalizationDataState.size() != pda.getStates().size() || normalizationDataSymbol.size() != pda.getPushdownStoreAlphabet().size()) {
		throw exception::CommonException("Automaton normalize require connected deterministic pushdown automaton");
	}

	automaton::DPDA < InputSymbolType, EpsilonType, unsigned, unsigned > result ( normalizationDataState.find ( pda.getInitialState ( ) )->second, normalizationDataSymbol.find ( pda.getInitialSymbol ( ) )->second );
	result.setInputAlphabet(pda.getInputAlphabet());

	for(const DefaultSymbolType & symbol : pda.getPushdownStoreAlphabet())
		result.addPushdownStoreSymbol(normalizationDataSymbol.find(symbol)->second);

	for(const DefaultStateType & state : pda.getStates())
		result.addState ( normalizationDataState.find(state)->second);

	for(const DefaultStateType & state : pda.getFinalStates())
		result.addFinalState ( normalizationDataState.find(state)->second );

	for(const auto & transition : pda.getTransitions()) {
		std::vector<unsigned> pop;
		for(const auto& elem : std::get<2>(transition.first)) {
			pop.push_back(normalizationDataSymbol.find(elem)->second);
		}
		std::vector<unsigned> push;
		for(const auto& elem : transition.second.second) {
			push.push_back(normalizationDataSymbol.find(elem)->second);
		}
		result.addTransition ( normalizationDataState.find(std::get<0>(transition.first))->second, std::get<1>(transition.first), pop, normalizationDataState.find(transition.second.first)->second, push);
	}

	return result;
}

} /* namespace simplify */

} /* namespace automaton */

#endif /* NORMALIZE_H_ */
