/*
 * AutomataConcatenation.h
 *
 *  Created on: 20. 11. 2014
 *	  Author: Tomas Pecka
 */

#ifndef AUTOMATA_CONCATENATION_H_
#define AUTOMATA_CONCATENATION_H_

#include <core/multipleDispatch.hpp>
#include <automaton/Automaton.h>
#include <automaton/FSM/EpsilonNFA.h>

namespace automaton {

namespace transform {

/**
 * Concatenates two automata.
 *  - For finite automata A1, A2, we create automaton L accepting L(A1).L(A2) (Melichar, 2.78)
 */
class AutomataConcatenation : public std::PromotingDoubleDispatch<AutomataConcatenation, automaton::Automaton, const automaton::AutomatonBase &> {
public:
	static automaton::Automaton concatenation(const automaton::Automaton& first, const automaton::Automaton& second);

	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > concatenation(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second);
	template < class SymbolType, class StateType >
	static automaton::NFA < SymbolType, StateType > concatenation(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second);
};

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > AutomataConcatenation::concatenation(const automaton::NFA < SymbolType, StateType > & first, const automaton::NFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	StateType q01q02(first.getInitialState(), second.getInitialState());
	automaton::NFA < SymbolType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));
	res.addState(q01q02);

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.setInitialState(q01q02);

	for(const auto& t : first.getTransitions()) {
		for(const auto& q : t.second) {
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, q));

			if(first.getFinalStates().count(q) > 0)
				res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(secondDefault, second.getInitialState()));
		}
	}
	for(const auto& t : second.getTransitions())
		for(const auto& q : t.second)
			res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, q));

	for(const auto& t : first.getTransitionsFromState(first.getInitialState()))
		for(const auto& q : t.second)
			res.addTransition(q01q02, t.first.second, StateType(firstDefault, q));
	for(const auto& t : second.getTransitionsFromState(second.getInitialState()))
		for(const auto& q : t.second)
			res.addTransition(q01q02, t.first.second, StateType(secondDefault, q));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));
	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.addFinalState(q01q02);

	return res;
}

template < class SymbolType, class StateType >
automaton::NFA < SymbolType, StateType > AutomataConcatenation::concatenation(const automaton::DFA < SymbolType, StateType > & first, const automaton::DFA < SymbolType, StateType > & second) {
	static StateType firstDefault ( 1 );
	static StateType secondDefault ( 2 );

	StateType q01q02(first.getInitialState(), second.getInitialState());
	automaton::NFA < SymbolType, StateType > res(StateType(firstDefault, first.getInitialState()));

	for(const auto& q : first.getStates())
		res.addState(StateType(firstDefault, q));
	for(const auto& q : second.getStates())
		res.addState(StateType(secondDefault, q));
	res.addState(q01q02);

	for(const auto& symbol : first.getInputAlphabet())
		res.addInputSymbol(symbol);
	for(const auto& symbol : second.getInputAlphabet())
		res.addInputSymbol(symbol);

	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.setInitialState(q01q02);

	for(const auto& t : first.getTransitions()) {
		res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(firstDefault, t.second));

		if(first.getFinalStates().count(t.second) > 0)
			res.addTransition(StateType(firstDefault, t.first.first), t.first.second, StateType(secondDefault, second.getInitialState()));
	}
	for(const auto& t : second.getTransitions())
		res.addTransition(StateType(secondDefault, t.first.first), t.first.second, StateType(secondDefault, t.second));

	for(const auto& t : first.getTransitionsFromState(first.getInitialState()))
		res.addTransition(q01q02, t.first.second, StateType(firstDefault, t.second));
	for(const auto& t : second.getTransitionsFromState(second.getInitialState()))
		res.addTransition(q01q02, t.first.second, StateType(secondDefault, t.second));

	for(const auto& q : second.getFinalStates())
		res.addFinalState(StateType(secondDefault, q));
	if(first.getFinalStates().count(first.getInitialState()) > 0)
		res.addFinalState(q01q02);

	return res;
}

} /* namespace transform */

} /* namespace automaton */

#endif /* AUTOMATA_CONCATENATION_H_ */
