#include "KruskalTest.h"

#include <graph/spanningtree/Kruskal.h>

#include <graph/generate/RandomGraphFactory.h>
#include <exception/CommonException.h>
#include "../TestUtils.h"

CPPUNIT_TEST_SUITE_NAMED_REGISTRATION( GraphKruskalTest, "graph" );
CPPUNIT_TEST_SUITE_REGISTRATION( GraphKruskalTest );

#define GraphDef( edgeCount ) const struct { unsigned nodeCnt; unsigned edgeCnt; struct { unsigned node1; unsigned node2; unsigned weight; } edges [ edgeCount ]; }

GraphDef(1) graph1 = {
	2, 1,
	{
		{ 0, 1,  5 }
	}
};
GraphDef(1) spanningTree1 = {
	2, 1,
	{
		{ 0, 1,   5 }
	}
};

//----------------------------------
GraphDef(3) graph2 = {
	3, 3,
	{
		{ 0, 1,   7 },
		{ 1, 2,   6 },
		{ 1, 2,   5 }
	}
};
GraphDef(2) spanningTree2 = {
	3, 2,
	{
		{ 0, 1,   7 },
		{ 1, 2,   6 },
	}
};

//----------------------------------
GraphDef(5) graph3 = {
	4, 5,
	{
		{ 0, 1,   4 },
		{ 0, 2,   1 },
		{ 0, 3,   2 },
		{ 1, 3,   3 },
		{ 3, 2,   5 }
	}
};
GraphDef(3) spanningTree3 = {
	4, 3,
	{
		{ 0, 2,   1 },
		{ 0, 3,   2 },
		{ 1, 3,   3 }
	}
};

GraphDef(24) graph4 = {
	12, 24,
	{
		{ 0, 1,   1 },
		{ 1, 2,   2 },
		{ 2, 3,   2 },
		{ 0, 4,   2 },
		{ 1, 4,   4 },
		{ 1, 5,   3 },
		{ 2, 5,   1 },
		{ 2, 6,   3 },
		{ 3, 6,   4 },
		{ 3, 7,   5 },
		{ 4, 5,   1 },
		{ 5, 6,   5 },
		{ 6, 7,   2 },
		{ 4, 8,   5 },
		{ 4, 9,   2 },
		{ 5, 9,   3 },
		{ 5,10,   4 },
		{ 6,10,   1 },
		{ 6,11,   2 },
		{ 7,11,   5 },
		{ 8, 9,   6 },
		{ 9,10,   2 },
		{10,11,   4 }
	}
};
GraphDef(11) spanningTree4 = {
	12, 11,
	{
		{ 0, 1,   1 },
		{ 1, 2,   2 },
		{ 2, 3,   2 },
		{ 2, 5,   1 },
		{ 4, 5,   1 },
		{ 6, 7,   2 },
		{ 4, 8,   5 },
		{ 4, 9,   2 },
		{ 6,10,   1 },
		{ 6,11,   3 },
		{ 9,10,   1 }
	}
};



template< typename Type1, typename Type2 >
void testGraph( unsigned graphID, const Type1 * graphDef, const Type2 * spanTreeDef )
{
	std::vector<graph::Node> id2node( graphDef->nodeCnt );
	std::vector<graph::UndirectedEdge> id2edge( graphDef->edgeCnt );
	std::unordered_map<graph::Node,unsigned> node2id;
	std::unordered_map<graph::UndirectedEdge,unsigned> edge2id;

	graph::AdjacencyListUndirectedGraph graph;

	for (unsigned i = 0; i < graphDef->nodeCnt; i++) {
		graph::Node node( std::string("n")+toString(i) );
		graph.addNode( node );
		id2node[ i ] = node;
		node2id[ node ] = i;
	}
	for (unsigned i = 0; i < graphDef->edgeCnt; i++) {
		graph::UndirectedEdge edge( id2node[ graphDef->edges[i].node1 ], id2node[ graphDef->edges[i].node2 ], std::string("e")+toString(i) );
		graph.addEdge( edge, graphDef->edges[i].weight );
		id2edge[ i ] = edge;
		edge2id[ edge ] = i;
	}

	std::cout << "testing spanning tree #" << graphID << std::endl;

	graph::AdjacencyListUndirectedGraph spanningTree = graph::spanningtree::Kruskal::kruskal( graph );

	CPPUNIT_ASSERT_EQUAL_MESSAGE( "invalid node count", spanTreeDef->nodeCnt, (unsigned)spanningTree.nodeCount() );
	CPPUNIT_ASSERT_EQUAL_MESSAGE( "invalid edge count", spanTreeDef->edgeCnt, (unsigned)spanningTree.edgeCount() );
	for (unsigned i = 0; i < spanTreeDef->edgeCnt; i++) {
		CPPUNIT_ASSERT_EQUAL_MESSAGE( "missing edge: "+toString(spanTreeDef->edges[i].node1)+","+toString(spanTreeDef->edges[i].node2),
			true, spanningTree.hasEdge( id2node[ spanTreeDef->edges[i].node1 ], id2node[ spanTreeDef->edges[i].node2 ] ) );
	}
}

void GraphKruskalTest::runTests()
{
	testGraph( 1, &graph1, &spanningTree1 );
	testGraph( 2, &graph2, &spanningTree2 );
	testGraph( 3, &graph3, &spanningTree3 );
	testGraph( 4, &graph4, &spanningTree4 );

	std::cout << "testing random big graphs" << std::endl;
	int nodeCount = 8;
	for (int i = 0; i < 8; i++) {
		graph::AdjacencyListUndirectedGraph graph = graph::generate::RandomGraphFactory::generateUndirectedGraph( nodeCount, nodeCount*2 );
		graph::spanningtree::Kruskal::kruskal( graph );
		nodeCount *= 2;
	}
}
