/*
 * AlgorithmAbstraction.hpp
 *
 *  Created on: 11. 7. 2017
 *	  Author: Jan Travnicek
 */

#ifndef _ALGORITHM_ABSTRACTION_HPP_
#define _ALGORITHM_ABSTRACTION_HPP_

#include <abstraction/ResultClassSelector.hpp>
#include <tuple>
#include <memory>
#include <abstraction/Registry.h>

namespace abstraction {

template < class ReturnType, class ... ParamTypes >
class Callback {
	std::function < ReturnType ( ParamTypes ... ) > m_callback;

public:
	Callback ( std::function < ReturnType ( ParamTypes ... ) > callback ) : m_callback ( callback ) {
	}

	ReturnType * operator ( ) ( ParamTypes ... params ) {
		return new ReturnType ( m_callback ( std::forward < ParamTypes > ( params ) ... ) );
	}
};

template < class ReturnType, class ... ParamTypes >
class Callback < ReturnType *, ParamTypes ... > {
	std::function < ReturnType * ( ParamTypes ... ) > m_callback;

public:
	Callback ( std::function < ReturnType * ( ParamTypes ... ) > callback ) : m_callback ( callback ) {
	}

	ReturnType * operator ( ) ( ParamTypes ... params ) {
		return m_callback ( std::forward < ParamTypes > ( params ) ... );
	}
};

template < class ReturnType, class ... ParamTypes >
class AlgorithmAbstractionBase : public ResultClassSelector < ReturnType >::OutputClass {
protected:
	std::tuple < std::shared_ptr < ValueOperationAbstraction < typename std::decay < ParamTypes >::type > > ... > inputs;

	virtual bool attachInput ( const std::shared_ptr < OperationAbstraction > & input, unsigned index ) override {
		auto attachCallback = [ & ] ( auto & param ) {
			if ( param != nullptr )
				return false;

			typename std::decay < decltype ( param )>::type validData = std::dynamic_pointer_cast < typename std::decay < decltype ( param ) >::type::element_type > ( input );
			if ( validData ) {
				param = validData;
				return true;
			} else {
				return false;
			}
		};
		return std::call_on_nth < bool > ( inputs, index, attachCallback );
	}

	virtual bool detachInput ( const std::shared_ptr < OperationAbstraction > & input ) override {
		bool res = false;
		std::foreach ( inputs, [ & ] ( auto & param ) {
				if ( param == input ) {
					param = nullptr;
					res = true;
				}
			} );
		return res;
	}

public:
	virtual bool inputsReady ( ) const override {
		auto readyCallback = [ ] ( const auto & param ) {
			return ( bool ) param && param->isReady ( );
		};

		return all_of ( inputs, readyCallback );
	}

	virtual bool inputsAttached ( ) const override {
		auto attachedCallback = [ ] ( const auto & param ) {
			return ( bool ) param;
		};

		return all_of ( inputs, attachedCallback );
	}

	virtual bool eval ( ) override {
		if ( ! inputsAttached ( ) )
			return false;

		if ( this->cached ( ) )
			return true;

		auto evalCallback = [ ] ( const auto & param ) {
			return param->eval ( );
		};

		if ( ! all_of ( inputs, evalCallback ) )
			return false;

		return this->run ( );
	}

	virtual unsigned numberOfParams ( ) const override {
		return sizeof ... ( ParamTypes );
	}

	virtual std::shared_ptr < OperationAbstraction > getImmediateValueFromParam ( unsigned index, const std::string & value ) const override {
		auto callback = [ & ] ( auto & param ) {
			std::string paramType = std::type_name < typename std::decay < decltype ( param ) >::type::element_type::return_type > ( );
			return Registry::getImmediateAbstraction ( std::move ( paramType ), value );
		};
		return std::call_on_nth < std::shared_ptr < OperationAbstraction > > ( inputs, index, callback );
	}

	virtual std::shared_ptr < OperationAbstraction > getXmlParserFromParam ( unsigned index, std::deque < sax::Token > tokens ) const override {
		auto callback = [ & ] ( auto & param ) {
			std::string paramType = std::type_name < typename std::decay < decltype ( param ) >::type::element_type::return_type > ( );
			return Registry::getXmlParserAbstraction ( std::move ( paramType ), std::move ( tokens ) );
		};
		return std::call_on_nth < std::shared_ptr < OperationAbstraction > > ( inputs, index, callback );
	}
};

template < class ReturnType, class ... ParamTypes >
class AlgorithmAbstraction : public AlgorithmAbstractionBase < ReturnType, ParamTypes ... > {
	Callback < ReturnType, ParamTypes ... > m_callback;

public:
	AlgorithmAbstraction ( std::function < ReturnType ( ParamTypes ... ) > callback ) : m_callback ( callback ) {
	}

	virtual bool run ( ) override {
		if ( ! this->inputsReady ( ) )
			return false;

		if ( this->cached ( ) )
			return true;

		this->run_helper ( m_callback, this->inputs, std::make_index_sequence < sizeof ... ( ParamTypes ) > { } );
		return true;
	}

};

template < class ReturnType, class ... ParamTypes >
class AlgorithmAbstraction < ReturnType *, ParamTypes ... > : public AlgorithmAbstractionBase < ReturnType, ParamTypes ... > {
	Callback < ReturnType *, ParamTypes ... > m_callback;

public:
	AlgorithmAbstraction ( std::function < ReturnType * ( ParamTypes ... ) > callback ) : m_callback ( callback ) {
	}

	virtual bool run ( ) override {
		if ( ! this->inputsReady ( ) )
			return false;

		if ( this->cached ( ) )
			return true;

		this->run_helper ( m_callback, this->inputs, std::make_index_sequence < sizeof ... ( ParamTypes ) > { } );
		return true;
	}

};

} /* namespace abstraction */

#endif /* _ALGORITHM_ABSTRACTION_HPP_ */
