/*
 * CommonBase.h
 *
 *  Created on: Apr 05, 2014
 *      Author: Jan Travnicek
 */

#ifndef COMMON_BASE_H_
#define COMMON_BASE_H_

#include <typeinfo>
#include <typeindex>
#include <ostream>
#include <stdexcept>

namespace alib {

/**
 * Most common base for every datatype in algorithms library toolkit.
 */
class CommonBaseBase {
public:
	/**
	 * Virtual destructor just to have it.
	 */
	virtual ~CommonBaseBase ( ) noexcept { }

};

/**
 * Middle common base specifies usual operations of all datatypes in algorithms library toolkit.
 *
 * \tparam T type that inherits from the commmon base.
 */
template < typename T >
class CommonBaseMiddle : public CommonBaseBase {
public:
	/**
	 * Virtual copy constructor
	 */
	virtual T * clone ( ) const = 0;

	/**
	 * Virtual move constructor
	 */
	virtual T * plunder ( ) && = 0;

	/**
	 * Compares this object with other for inequality.
	 *
	 * \returns true if this and other objects are not equal, false othervise
	 */
	bool operator !=( const T & other ) const {
		return this->compare ( other ) != 0;
	}

	/**
	 * Compares this object with other for equality.
	 *
	 * \returns true if this and other objects are equal, false othervise
	 */
	bool operator ==( const T & other ) const {
		return this->compare ( other ) == 0;
	}

	/**
	 * Determines whether this object is strictly smaller in relation to the other object.
	 *
	 * \returns true if this object is strictly smaller than other object are not equal, false othervise
	 */
	bool operator <( const T & other ) const {
		return this->compare ( other ) < 0;
	}

	/**
	 * Determines whether this object is smaller or equal in relation to the other object.
	 *
	 * \returns true if this object is smaller or equal than other object are not equal, false othervise
	 */
	bool operator <=( const T & other ) const {
		return this->compare ( other ) <= 0;
	}

	/**
	 * Determines whether this object is strictly bigger in relation to the other object.
	 *
	 * \returns true if this object is strictly bigger than other object are not equal, false othervise
	 */
	bool operator >( const T & other ) const {
		return this->compare ( other ) > 0;
	}

	/**
	 * Determines whether this object is bigger or equal in relation to the other object.
	 *
	 * \returns true if this object is bigger or equal than other object are not equal, false othervise
	 */
	bool operator >=( const T & other ) const {
		return this->compare ( other ) >= 0;
	}

	/**
	 * \brief Comparison helper method evaluating allowing possibly deeper comparison of this with other class of the same type.
	 *
	 * \details If the other class is of different type the relative order is computer by means of type_index.
	 *
	 * \param other the other class to compare with
	 *
	 * \returns result of actual comparison if type of this class and other class is the same, result of difference of type indexes othervise.
	 */
	virtual int compare ( const T & other ) const = 0;

	/**
	 * Print this object as raw representation to ostream.
	 *
	 * \param os ostream where to print
	 * \param instance object to print
	 *
	 * \returns modified output stream
	 */
	friend std::ostream & operator <<( std::ostream & os, const T & instance ) {
		instance >> os;
		return os;
	}

	/**
	 * Print this instance as raw representation to ostream.
	 *
	 * \param os ostream where to print
	 */
	virtual void operator >>( std::ostream & ) const = 0;

	/**
	 * Casts this instance to as compact as possible string representation.
	 *
	 * \returns string representation of the object
	 */
	virtual explicit operator std::string ( ) const = 0;

};

/**
 * The actual common base that is to be used as a base class for concrete type hierarchy object types.
 *
 * \tparam T type that inherits from the commmon base.
 */
template< typename T >
class CommonBase : public CommonBaseMiddle < T > {
public:

	/**
	 * Helper for normalisation of templated datatypes in algorithms library toolkit to the same types but with default template parameters.
	 *
	 * \returns new instance of the grammar with default template parameters or unmodified instance if the template parameters were already default ones
	 */
	virtual T * normalize ( ) && = 0;
};

} /* namespace alib */

namespace std {

/**
 * Helper for comparing everything that inherits from CommonBaseBase.
 */
template < class T >
struct compare < T, typename std::enable_if < std::is_base_of < alib::CommonBaseBase, typename std::decay < T >::type >::value >::type > {

	/**
	 * Compare operator that determines relation between first and second object.
	 *
	 * \returns value < 0 if first < second
	 *          value == 0 if first == second
	 *          value > 0 if first > second
	 */
	int operator()(const T& first, const T& second) const {
		return first.compare(second);
	}
};

} /* namespace std */

#endif /* COMMON_BASE_H_ */
