/*
 * WrapperBase.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef WRAPPER_BASE_H_
#define WRAPPER_BASE_H_

#include <ostream>
#include <memory>

namespace alib {

/**
 * Most common base for every datatype wrapper in algorithms library toolkit.
 */
class WrapperBaseBase {
public:
	/**
	 * Virtual destructor just to have it.
	 */
	virtual ~WrapperBaseBase ( ) noexcept { }

};

/**
 * Middle wrapper base specifies usual operations of all datatypes wrappers in algorithms library toolkit.
 *
 * \tparam T type that is a base class of the concrete wrapper that inherits from this.
 */
template < typename T >
class WrapperBaseMiddle : public WrapperBaseBase {
protected:
	std::cow_shared_ptr < T > m_data;

private:
	void unify ( WrapperBaseMiddle & other ) {
		if ( this->m_data.getUseCount ( ) > other.m_data.getUseCount ( ) )
			other.m_data = this->m_data;
		else
			this->m_data = other.m_data;
	}

public:
	/**
	 * Disable construction from integer so that it is not confused with other constructors.
	 */
	WrapperBaseMiddle ( int ) = delete;

	/**
	 * Constructor that wraps raw pointer. Takes ownership of the pointer.
	 */
	explicit WrapperBaseMiddle ( T * data ) : m_data ( data ) {
	}

	/**
	 * Constructor that wraps an object given by constant reference. Uses clone of the parameter internally.
	 */
	explicit WrapperBaseMiddle ( const T & data ) : m_data ( data.clone ( ) ) {
	}

	/**
	 * Constructor that wraps an object given by r-value reference. Uses plunder of the parameter internally.
	 */
	explicit WrapperBaseMiddle ( T && data ) : m_data ( std::move ( data ).plunder ( ) ) {
	}

	/**
	 * Gets the wrapped object.
	 *
	 * \returns wrapped object.
	 */
	const T & getData ( ) const {
		return * m_data;
	}

	/**
	 * Gets the wrapped object.
	 *
	 * \returns wrapped object.
	 */
	T & getData ( ) {
		return * m_data;
	}

	/**
	 * Sets the wrapped object from raw pointer. Takes ownership of the pointer.
	 *
	 * Internally handles situation like storing the same value that is already wrapped. Usefull in normalisation.
	 */
	void setData ( T * data ) {
		if ( & getData ( ) == data )
			return;

		this->m_data = std::cow_shared_ptr < T > ( data );
	}

	/**
	 * Sets the wrapped object from constant reference. Uses clone of the parameter internally.
	 */
	void setData ( const T & data ) {
		setData ( data.clone ( ) );
	}

	/**
	 * Sets the wrapped object from r-value reference. Uses plunder of the parameter internally.
	 */
	void setData ( T && data ) {
		setData ( std::move ( data ).plunder ( ) );
	}

	/**
	 * Determines whether this object is bigger or equal in relation to the other object.
	 *
	 * \returns true if this object is bigger or equal than other object are not equal, false othervise
	 */
	bool operator >=( const WrapperBaseMiddle & other ) const {
		return this->compare ( other ) >= 0;
	}

	/**
	 * Determines whether this object is smaller or equal in relation to the other object.
	 *
	 * \returns true if this object is smaller or equal than other object are not equal, false othervise
	 */
	bool operator <=( const WrapperBaseMiddle & other ) const {
		return this->compare ( other ) <= 0;
	}

	/**
	 * Determines whether this object is strictly bigger in relation to the other object.
	 *
	 * \returns true if this object is strictly bigger than other object are not equal, false othervise
	 */
	bool operator >( const WrapperBaseMiddle & other ) const {
		return this->compare ( other ) > 0;
	}

	/**
	 * Determines whether this object is strictly smaller in relation to the other object.
	 *
	 * \returns true if this object is strictly smaller than other object are not equal, false othervise
	 */
	bool operator <( const WrapperBaseMiddle & other ) const {
		return this->compare ( other ) < 0;
	}

	/**
	 * Compares this object with other for inequality.
	 *
	 * \returns true if this and other objects are not equal, false othervise
	 */
	bool operator !=( const WrapperBaseMiddle & other ) const {
		return this->compare ( other ) != 0;
	}

	/**
	 * Compares this object with other for equality.
	 *
	 * \returns true if this and other objects are equal, false othervise
	 */
	bool operator ==( const WrapperBaseMiddle & other ) const {
		return this->compare ( other ) == 0;
	}

	/**
	 * \brief Compares this wrapper with another, establishing relationship between them.
	 *
	 * \returns result of actual comparison of wrapped objects returning:
	 *          value < 0 if this < other
	 *          value == 0 if this == other
	 *          value > 0 if this > other
	 */
	int compare ( const WrapperBaseMiddle & other ) const {
		if ( this->m_data.get ( ) == other.m_data.get ( ) ) return 0;

		int res = ( * this->m_data ).compare ( * other.m_data );

		if ( res == 0 ) const_cast < WrapperBaseMiddle * > ( this )->unify ( const_cast < WrapperBaseMiddle & > ( other ) );

		return res;
	}

	/**
	 * Print the wrapped object as raw representation to ostream.
	 *
	 * \param os ostream where to print
	 * \param instance wrapper to print
	 *
	 * \returns modified output stream
	 */
	friend std::ostream & operator <<( std::ostream & os, const WrapperBaseMiddle & instance ) {
		os << instance.getData ( );
		return os;
	}

	/**
	 * Casts the wrapped object to as compact as possible string representation.
	 *
	 * \returns string representation of the wrapped object
	 */
	explicit operator std::string ( ) const {
		return ( std::string ) * m_data;
	}

	/**
	 * Internaly increments the wrapped object by one to make different from the original one.
	 *
	 * \returns reference to this
	 */
	WrapperBaseMiddle < T > & operator ++ ( ) {
		T * res = std::move ( this->getData ( ) ).inc ( );

		if ( res != NULL )
			this->setData ( res );

		return *this;
	}
};

/**
 * The actual wrapper base that is to be used as a base class for concrete type hierarchy object types.
 *
 * \tparam T type that is a base class of the concrete wrapper that inherits from this.
 */
template < typename T >
class WrapperBase : public WrapperBaseMiddle < T > {
public:
	using WrapperBaseMiddle < T >::WrapperBaseMiddle;

	/**
	 * Helper for normalisation of wrapped templated datatypes in algorithms library toolkit to the same types but with default template parameters.
	 *
	 * This method actually changes the stored object to the normalized one.
	 */
	void normalize ( ) {
		this->setData ( std::move ( this->getData ( ) ).normalize ( ) );
	}

};

} /* namespace alib */

namespace std {

/**
 * Helper for comparing everything that inherits from WrapperBaseBase.
 */
template < class T >
struct compare < T, typename std::enable_if < std::is_base_of < alib::WrapperBaseBase, T >::value >::type > {

	/**
	 * Compare operator that determines relation between first and second object.
	 *
	 * \returns value < 0 if first < second
	 *          value == 0 if first == second
	 *          value > 0 if first > second
	 */
	int operator ()( const T & first, const T & second ) const {
		return first.getData ( ).compare ( second.getData ( ) );
	}

};

} /* namespace std */

#endif /* WRAPPER_BASE_H_ */
