/*
 * ObjectsTrie.h
 *
 * Created on: Apr 1, 2013
 * Author: Jan Travnicek
 */

#ifndef OBJECTS_TRIE_H_
#define OBJECTS_TRIE_H_

#include <trie>
#include <string>
#include <sstream>

#include "ContainerBase.h"
#include "ContainerFeatures.h"

#include <core/xmlApi.hpp>
#include "../object/UniqueObject.h"
#include <object/AnyObject.h>

namespace container {

/**
 * Basic container from which are derived all other containers.
 * Contains reason why the container occured.
 */
template < class KeyType, class ValueType >
class ObjectsTrie final : public std::trie < KeyType, ValueType >, public ContainerBase {
public:
	explicit ObjectsTrie ( std::trie < KeyType, ValueType > );

	virtual ContainerBase * clone ( ) const;

	virtual ContainerBase * plunder ( ) &&;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ObjectsTrie & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "Trie";

		return xmlTagName;
	}

	static ObjectsTrie parse ( std::deque < sax::Token >::iterator & input );
	static std::map < KeyType, std::trie < KeyType, ValueType > > parseChildren ( std::deque < sax::Token >::iterator & input );
	static std::trie < KeyType, ValueType > parseRaw ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;
	static void composeChildren ( std::deque < sax::Token > & out, const std::map < KeyType, std::trie < KeyType, ValueType > > & children );
	static void compose ( std::deque < sax::Token > & out, const std::trie < KeyType, ValueType > & input );

	virtual alib::ObjectBase * inc ( ) &&;

	std::trie < alib::Object, alib::Object > normalizeRaw ( std::trie < KeyType, ValueType > && node ) {
		std::map < alib::Object, std::trie < alib::Object, alib::Object > > children;

		for ( std::pair < KeyType, std::trie < KeyType, ValueType > > && child : std::make_moveable_map ( node.getChildren ( ) ) ) {
			children.insert ( std::make_pair ( alib::Object ( alib::AnyObject < KeyType > ( std::move ( child.first ) ) ), normalizeRaw ( std::move ( child.second ) ) ) );
		}

		return std::trie < alib::Object, alib::Object > ( alib::Object ( alib::AnyObject < ValueType > ( std::move ( node.getData ( ) ) ) ), std::move ( children ) );
	}

	typedef ObjectsTrie < > normalized_type;

	virtual ContainerBase * normalize ( ) && {
		return new ObjectsTrie < alib::Object > ( normalizeRaw ( std::move ( * this ) ) );
	}
};

template < class KeyType, class ValueType >
ObjectsTrie < KeyType, ValueType >::ObjectsTrie ( std::trie < KeyType, ValueType > raw ) : std::trie < KeyType, ValueType > ( std::move ( raw ) ) {
}

template < class KeyType, class ValueType >
ContainerBase * ObjectsTrie < KeyType, ValueType >::clone ( ) const {
	return new ObjectsTrie ( * this );
}

template < class KeyType, class ValueType >
ContainerBase * ObjectsTrie < KeyType, ValueType >::plunder ( ) && {
	return new ObjectsTrie ( std::move ( * this ) );
}

template < class KeyType, class ValueType >
int ObjectsTrie < KeyType, ValueType >::compare ( const ObjectsTrie & other ) const {
	static std::compare < std::trie < KeyType, ValueType > > comp;

	return comp ( static_cast < const std::trie < KeyType, ValueType > & > ( * this ), static_cast < const std::trie < KeyType, ValueType > & > ( other ) );
}

template < class KeyType, class ValueType >
void ObjectsTrie < KeyType, ValueType >::operator >>( std::ostream & os ) const {
	os << "(ObjectsTrie " << static_cast < const std::trie < KeyType, ValueType > & > ( * this ) << ")";
}

template < class KeyType, class ValueType >
ObjectsTrie < KeyType, ValueType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return std::move ( ss ).str ( );
}

template < class KeyType, class ValueType >
ObjectsTrie <KeyType, ValueType > ObjectsTrie < KeyType, ValueType >::parse ( std::deque < sax::Token >::iterator & input ) {
	return ObjectsTrie < KeyType, ValueType > ( parseRaw ( input ) );
}

template < class KeyType, class ValueType >
void ObjectsTrie < KeyType, ValueType >::compose ( std::deque < sax::Token > & out ) const {
	compose ( out, * this );
}

template < class KeyType, class ValueType >
std::map < KeyType, std::trie < KeyType, ValueType > > ObjectsTrie < KeyType, ValueType >::parseChildren ( std::deque < sax::Token >::iterator & input ) {
	std::map < KeyType, std::trie < KeyType, ValueType > > children;

	while ( sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, "Child" ) ) {
		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "Child" );

		KeyType key = alib::xmlApi < KeyType >::parse ( input );
		ValueType value = alib::xmlApi < ValueType >::parse ( input );
		std::map < KeyType, std::trie < KeyType, ValueType > > innerChildren = parseChildren ( input );

		children.insert ( std::make_pair ( std::move ( key ), std::trie < KeyType, ValueType > ( std::move ( value ), std::move ( innerChildren ) ) ) );

		sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "Child" );
	}

	return children;
}

template < class KeyType, class ValueType >
std::trie < KeyType, ValueType > ObjectsTrie < KeyType, ValueType >::parseRaw ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ObjectsTrie < >::getXmlTagName ( ) );
	ValueType value = alib::xmlApi < ValueType >::parse ( input );
	std::map < KeyType, std::trie < KeyType, ValueType > > children = parseChildren ( input );

	std::trie < KeyType, ValueType > trie ( std::move ( value ), std::move ( children ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ObjectsTrie < >::getXmlTagName ( ) );
	return trie;
}

template < class KeyType, class ValueType >
void ObjectsTrie < KeyType, ValueType >::composeChildren ( std::deque < sax::Token > & out, const std::map < KeyType, std::trie < KeyType, ValueType > > & children ) {
	for ( const std::pair < const KeyType, std::trie < KeyType, ValueType > > & child : children ) {
		out.emplace_back ( "Child", sax::Token::TokenType::START_ELEMENT );

		alib::xmlApi < KeyType >::compose ( out, child.first );
		alib::xmlApi < ValueType >::compose ( out, child.second.getData ( ) );
		composeChildren ( out, child.second.getChildren ( ) );

		out.emplace_back ( "Child", sax::Token::TokenType::END_ELEMENT );
	}
}

template < class KeyType, class ValueType >
void ObjectsTrie < KeyType, ValueType >::compose ( std::deque < sax::Token > & out, const std::trie < KeyType, ValueType > & input ) {
	out.emplace_back ( ObjectsTrie < >::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < ValueType >::compose ( out, input.getData ( ) );
	composeChildren ( out, input.getChildren ( ) );
	out.emplace_back ( ObjectsTrie < >::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < class KeyType, class ValueType >
alib::ObjectBase* ObjectsTrie < KeyType, ValueType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace container */

namespace alib {

template < typename T, typename R >
struct xmlApi < std::trie < T, R > > {
	static std::trie < T, R > parse ( std::deque < sax::Token >::iterator & input );
	static bool first ( const std::deque < sax::Token >::const_iterator & input );
	static std::string xmlTagName ( );
	static void compose ( std::deque < sax::Token > & output, const std::trie < T, R > & data );
};

template < typename T, typename R >
std::trie < T, R > xmlApi < std::trie < T, R > >::parse ( std::deque < sax::Token >::iterator & input ) {
	return container::ObjectsTrie < T, R >::parseRaw ( input );
}

template < typename T, typename R >
bool xmlApi < std::trie < T, R > >::first ( const std::deque < sax::Token >::const_iterator & input ) {
	return sax::FromXMLParserHelper::isToken ( input, sax::Token::TokenType::START_ELEMENT, xmlTagName ( ) );
}

template < typename T, typename R >
std::string xmlApi < std::trie < T, R > >::xmlTagName ( ) {
	return container::ObjectsTrie < >::getXmlTagName ( );
}

template < typename T, typename R >
void xmlApi < std::trie < T, R > >::compose ( std::deque < sax::Token > & output, const std::trie < T, R > & input ) {
	return container::ObjectsTrie < T, R >::compose ( output, input );
}

} /* namespace alib */

#endif /* OBJECTS_TRIE_H_ */
