/*
 * XmlDataFactory.hpp
 *
 *  Created on: Jan 1, 2014
 *      Author: Jan Travnicek
 */

#ifndef XML_DATA_FACTORY_HPP_
#define XML_DATA_FACTORY_HPP_

#include <string>
#include <deque>
#include <measure>

#include <sax/Token.h>
#include <sax/SaxParseInterface.h>
#include <sax/SaxComposeInterface.h>
#include <core/xmlApi.hpp>
#include <exception/CommonException.h>

namespace alib {

/**
 * Data builder.
 */
class XmlDataFactory {
public:

	class fromFile {
		const std::string & filename;

	public:
		fromFile ( const std::string & file ) : filename ( file ) {
		}

		/**
		 * Parses the XML in file and returns the String.
		 * @param filename path to the file
		 * @return String
		 */
		template < class T >
		operator T ( ) {
			std::deque < sax::Token > tokens;
			sax::SaxParseInterface::parseFile ( filename, tokens );

			return fromTokens ( std::move( tokens ) );
		}
	};

	class fromString {
		const std::string & string;

	public:
		fromString ( const std::string & str ) : string ( str ) {
		}

		/**
		 * Parses the XML and returns the String.
		 * @param str string containing the XML
		 * @return String
		 */
		template < class T >
		operator T ( ) {
			std::deque < sax::Token > tokens;
			sax::SaxParseInterface::parseMemory ( string, tokens );

			return fromTokens ( std::move ( tokens ) );
		}
	};

	class fromStdin {
	public:
		/**
		 * Parses the XML from stdin and returns the String.
		 * @return String
		 */
		template < class T >
		operator T ( ) {
			std::deque < sax::Token > tokens;
			sax::SaxParseInterface::parseStdin ( tokens );

			return fromTokens ( std::move ( tokens ) );
		}
	};

	class fromStream {
		std::istream & in;

	public:
		fromStream ( std::istream & i ) : in ( i ) {
		}

		/**
		 * Parses the XML from stream and returns the String.
		 * @return String
		 */
		template < class T >
		operator T ( ) {
			std::deque < sax::Token > tokens;
			sax::SaxParseInterface::parseStream ( in, tokens );

			return fromTokens ( std::move ( tokens ) );
		}
	};

	class fromTokens {
		std::deque < sax::Token > tokens;

	public:
		fromTokens ( std::deque < sax::Token > && toks ) : tokens ( std::move ( toks ) ) {
		}

		/**
		 * Parses the String from list of tokens.
		 * @param tokens XML represented as list of tokens
		 * @return parsed String
		 */
		template < class T >
		operator T ( ) {
			xmlApiInputContext context ( tokens.begin ( ) );

			if ( context == tokens.end ( ) ) throw exception::CommonException ( "Empty tokens list" );

			if ( alib::xmlApi < exception::CommonException >::first ( context ) ) throw alib::xmlApi < exception::CommonException >::parse ( context );

			measurements::start ( "XML Parser", measurements::Type::INIT );
			T res = alib::xmlApi < T >::parse ( context );
			measurements::end ( );

			if ( context != tokens.end ( ) ) throw exception::CommonException ( "Unexpeted tokens at the end of the xml" );

			return res;
		}
	};

	/**
	 * Determines whether the input token stream contains given type
	 * @param tokens
	 * @return bool
	 */
	template < class T >
	static bool first ( const std::deque < sax::Token > & tokens ) {
		if( tokens.empty ( ) ) throw exception::CommonException ( "Empty tokens list" );

		return alib::xmlApi < exception::CommonException >::first ( tokens.begin ( ) ) || alib::xmlApi < T >::first ( tokens.begin ( ) );
	}

	/**
	 * Parses the XML in file and returns the String.
	 * @param filename path to the file
	 * @return String
	 */
	template < class T >
	static void toFile ( const T & data, const std::string & filename ) {
		std::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printFile ( filename, tokens );
	}

	/**
	 * Parses the XML and returns the String.
	 * @param str string containing the XML
	 * @return String
	 */
	template < class T >
	static std::string toString ( const T & data ) {
		std::deque < sax::Token > tokens = toTokens < T > ( data );
		std::string str;
		sax::SaxComposeInterface::printMemory ( str, tokens );

		return str;
	}

	/**
	 * Parses the XML from stdin and returns the String.
	 * @return String
	 */
	template < class T >
	static void toStdout ( const T & data ) {
		std::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printStdout ( tokens );
	}

	/**
	 * Parses the XML from stream and returns the String.
	 * @return String
	 */
	template < class T >
	static void toStream ( const T & data, std::ostream & out ) {
		std::deque < sax::Token > tokens = toTokens < T > ( data );
		sax::SaxComposeInterface::printStream ( out, tokens );
	}

	/**
	 * Parses the String from list of tokens.
	 * @param tokens XML represented as list of tokens
	 * @return parsed String
	 */
	template < class T >
	static std::deque < sax::Token > toTokens ( const T & data ) {
		xmlApiOutputContext context;

		measurements::start ( "XML Composer", measurements::Type::FINALIZE );
		alib::xmlApi < T >::compose ( context, data );
		measurements::end ( );

		return context;
	}

};

template < >
inline bool XmlDataFactory::first < exception::CommonException > ( const std::deque < sax::Token > & tokens ) {
	return alib::xmlApi < exception::CommonException >::first ( tokens.begin ( ) );
}

template < >
inline XmlDataFactory::fromTokens::operator exception::CommonException ( ) {
	std::deque < sax::Token >::iterator iter = tokens.begin ( );

	if ( iter == tokens.end ( ) ) throw exception::CommonException ( "Empty tokens list" );

	exception::CommonException res = alib::xmlApi < exception::CommonException >::parse ( iter );

	if ( iter != tokens.end ( ) ) throw exception::CommonException ( "Unexpeted tokens at the end of the xml" );

	return res;
}

} /* namespace alib */

#endif /* XML_DATA_FACTORY_HPP_ */
