/*
 * ExtendedNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef EXTENDED_NFA_H_
#define EXTENDED_NFA_H_

#include <map>
#include <ostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include <core/components.hpp>
#include <common/createUnique.hpp>
#include <object/UniqueObject.h>

#include "../AutomatonException.h"
#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

#include "../../label/InitialStateLabel.h"
#include "../../regexp/unbounded/UnboundedRegExpStructure.h"
#include "../../regexp/unbounded/UnboundedRegExpConcatenation.h"
#include "../../regexp/unbounded/UnboundedRegExpSymbol.h"

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class InputAlphabet;
class States;
class FinalStates;
class InitialState;

template<class SymbolType, class StateType >
class ExtendedNFA final : public AutomatonBase, public std::Components < ExtendedNFA < SymbolType, StateType >, SymbolType, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > transitions;

public:
	explicit ExtendedNFA ( StateType initialState );
	explicit ExtendedNFA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, StateType initialState, std::set < StateType > finalStates );
	explicit ExtendedNFA ( const CompactNFA < SymbolType, StateType > & other );
	template < class EpsilonType >
	explicit ExtendedNFA ( const EpsilonNFA < SymbolType, EpsilonType, StateType > & other );
	explicit ExtendedNFA ( const MultiInitialStateNFA < SymbolType, StateType > & other );
	explicit ExtendedNFA ( const NFA < SymbolType, StateType > & other );
	explicit ExtendedNFA ( const DFA < SymbolType, StateType > & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( StateType current, regexp::UnboundedRegExpStructure < SymbolType > input, StateType next );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addTransitions ( StateType current, regexp::UnboundedRegExpStructure < SymbolType > input, std::set < StateType > next );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const StateType & current, const regexp::UnboundedRegExpStructure < SymbolType > & input, const StateType & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > getTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > getTransitionsToState ( const StateType & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const ExtendedNFA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "ExtendedNFA";

		return xmlTagName;
	}

	static ExtendedNFA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, ExtendedNFA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef ExtendedNFA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( ExtendedNFA < > ) == typeid ( ExtendedNFA < SymbolType, StateType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		std::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		DefaultStateType initialState = AutomatonNormalize::normalizeState ( std::move ( this->template accessElement < InitialState > ( ).get ( ) ) );
		std::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		ExtendedNFA < > * res = new ExtendedNFA < > ( std::move ( states ), std::move ( alphabet ), std::move ( initialState ), std::move ( finalStates ) );

		for ( std::pair < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > && transition : std::make_moveable_map ( transitions ) ) {
			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			regexp::UnboundedRegExpStructure < DefaultSymbolType > input = AutomatonNormalize::normalizeRegExp ( std::move ( transition.first.second ) );
			std::set < DefaultStateType > targets = AutomatonNormalize::normalizeStates ( std::move ( transition.second ) );

			res->addTransitions ( std::move ( from ), std::move ( input ), std::move ( targets ) );
		}

		return res;
	}
};

} /* namespace automaton */

#include "DFA.h"
#include "NFA.h"
#include "CompactNFA.h"
#include "EpsilonNFA.h"
#include "MultiInitialStateNFA.h"

namespace automaton {

template<class SymbolType, class StateType >
ExtendedNFA < SymbolType, StateType >::ExtendedNFA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, StateType initialState, std::set < StateType > finalStates ) : std::Components < ExtendedNFA, SymbolType, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template<class SymbolType, class StateType >
ExtendedNFA < SymbolType, StateType >::ExtendedNFA ( StateType initialState ) : ExtendedNFA ( std::set < StateType > { initialState }, std::set < SymbolType > { }, initialState, std::set < StateType > { } ) {
}

template<class SymbolType, class StateType >
ExtendedNFA < SymbolType, StateType >::ExtendedNFA ( const CompactNFA < SymbolType, StateType > & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		regexp::UnboundedRegExpConcatenation < SymbolType > con;

		for ( auto & symbol : transition.first.second )
			con.appendElement ( regexp::UnboundedRegExpSymbol < SymbolType > ( symbol ) );

		std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( transition.first.first, regexp::UnboundedRegExpStructure < SymbolType > ( std::move ( con ) ) );
		transitions[key] = transition.second;
	}
}

template < class SymbolType, class StateType >
template < class EpsilonType >
ExtendedNFA < SymbolType, StateType >::ExtendedNFA ( const EpsilonNFA < SymbolType, EpsilonType, StateType > & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		if ( transition.first.second.template is < EpsilonType > ( ) ) {
			std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( transition.first.first, regexp::UnboundedRegExpStructure < SymbolType > ( ) );
			transitions[key] = transition.second;
		} else {
			std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( transition.first.first, regexp::UnboundedRegExpStructure < SymbolType > ( regexp::UnboundedRegExpSymbol < SymbolType > ( transition.first.second.template get < SymbolType > ( ) ) ) );
			transitions[key] = transition.second;
		}
	}
}

template<class SymbolType, class StateType >
ExtendedNFA < SymbolType, StateType >::ExtendedNFA ( const MultiInitialStateNFA < SymbolType, StateType > & other ) : ExtendedNFA ( other.getStates ( ) + std::set < StateType > { common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), other.getStates ( ) ) }, other.getInputAlphabet ( ), common::createUnique ( label::InitialStateLabel::instance < StateType > ( ), other.getStates ( ) ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( transition.first.first, regexp::UnboundedRegExpStructure < SymbolType > ( regexp::UnboundedRegExpSymbol < SymbolType > ( transition.first.second ) ) );
		transitions[key] = transition.second;
	}

	std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( this->getInitialState ( ), regexp::UnboundedRegExpStructure < SymbolType > ( ) );
	transitions[key] = other.getInitialStates ( );
}

template<class SymbolType, class StateType >
ExtendedNFA < SymbolType, StateType >::ExtendedNFA ( const NFA < SymbolType, StateType > & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( transition.first.first, regexp::UnboundedRegExpStructure < SymbolType > ( regexp::UnboundedRegExpSymbol < SymbolType > ( transition.first.second ) ) );
		transitions[key] = transition.second;
	}
}

template<class SymbolType, class StateType >
ExtendedNFA < SymbolType, StateType >::ExtendedNFA ( const DFA < SymbolType, StateType > & other ) : ExtendedNFA ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) ) {
		std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( transition.first.first, regexp::UnboundedRegExpStructure < SymbolType > ( regexp::UnboundedRegExpSymbol < SymbolType > ( transition.first.second ) ) );
		transitions[key].insert ( transition.second );
	}
}

template<class SymbolType, class StateType >
AutomatonBase * ExtendedNFA < SymbolType, StateType >::clone ( ) const {
	return new ExtendedNFA ( * this );
}

template<class SymbolType, class StateType >
AutomatonBase * ExtendedNFA < SymbolType, StateType >::plunder ( ) && {
	return new ExtendedNFA ( std::move ( * this ) );
}

template<class SymbolType, class StateType >
bool ExtendedNFA < SymbolType, StateType >::addTransition ( StateType from, regexp::UnboundedRegExpStructure < SymbolType > input, StateType to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + std::to_string ( from ) + "\" doesn't exist." );

	std::set < SymbolType > inputRegExpAlphabet = input.getStructure ( ).computeMinimalAlphabet ( );

	 // Transition regexp's alphabet must be subset of automaton's alphabet
	if ( !std::includes ( getInputAlphabet ( ).begin ( ), getInputAlphabet ( ).end ( ), inputRegExpAlphabet.begin ( ), inputRegExpAlphabet.end ( ) ) )
		throw AutomatonException ( "Input string is over different alphabet than automaton" );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + std::to_string ( to ) + "\" doesn't exist." );

	std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

template<class SymbolType, class StateType >
void ExtendedNFA < SymbolType, StateType >::addTransitions ( StateType from, regexp::UnboundedRegExpStructure < SymbolType > input, std::set < StateType > to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + std::to_string ( from ) + "\" doesn't exist." );

	std::set < SymbolType > inputRegExpAlphabet = input.getStructure ( ).computeMinimalAlphabet ( );

	 // Transition regexp's alphabet must be subset of automaton's alphabet
	if ( !std::includes ( getInputAlphabet ( ).begin ( ), getInputAlphabet ( ).end ( ), inputRegExpAlphabet.begin ( ), inputRegExpAlphabet.end ( ) ) )
		throw AutomatonException ( "Input string is over different alphabet than automaton" );

	if ( !std::includes ( getStates ( ).begin ( ), getStates ( ).end ( ), to.begin ( ), to.end ( ) ) )
		throw AutomatonException ( "Some target states don't exist." );

	std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	transitions [ std::move ( key ) ].insert ( std::make_moveable_set ( to ).begin ( ), std::make_moveable_set ( to ).end ( ) );
}

template<class SymbolType, class StateType >
bool ExtendedNFA < SymbolType, StateType >::removeTransition ( const StateType & from, const regexp::UnboundedRegExpStructure < SymbolType > & input, const StateType & to ) {
	std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > > key = std::make_pair ( from, input );

	return transitions[key].erase ( to );
}

template<class SymbolType, class StateType >
const std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > & ExtendedNFA < SymbolType, StateType >::getTransitions ( ) const {
	return transitions;
}

template<class SymbolType, class StateType >
std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > ExtendedNFA < SymbolType, StateType >::getTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + std::to_string ( from ) + "\" doesn't exist" );

	std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > transitionsFromState;

	for ( const std::pair < const std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState.insert ( make_pair ( transition.first, transition.second ) );

	return transitionsFromState;
}

template<class SymbolType, class StateType >
std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > ExtendedNFA < SymbolType, StateType >::getTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + std::to_string ( to ) + "\" doesn't exist" );

	std::map < std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > transitionsToState;

	for ( const std::pair < const std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState.insert ( make_pair ( transition.first, transition.second ) );

	return transitionsToState;
}

template<class SymbolType, class StateType >
int ExtendedNFA < SymbolType, StateType >::compare ( const ExtendedNFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialState ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialState ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	static std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template<class SymbolType, class StateType >
void ExtendedNFA < SymbolType, StateType >::operator >>( std::ostream & out ) const {
	out << "(ExtendedNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialState = " << getInitialState ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

template<class SymbolType, class StateType >
ExtendedNFA < SymbolType, StateType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template<class SymbolType, class StateType >
ExtendedNFA < SymbolType, StateType > ExtendedNFA < SymbolType, StateType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, ExtendedNFA::getXmlTagName() );

	std::set < StateType > states = AutomatonFromXMLParser::parseStates < StateType > ( input );
	std::set < SymbolType > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet < SymbolType > ( input );
	StateType initialState = AutomatonFromXMLParser::parseInitialState < StateType > ( input );
	std::set < StateType > finalStates = AutomatonFromXMLParser::parseFinalStates < StateType > ( input );

	ExtendedNFA < SymbolType, StateType > automaton ( std::move ( initialState ) );

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < ExtendedNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, ExtendedNFA::getXmlTagName() );
	return automaton;
}

template<class SymbolType, class StateType >
void ExtendedNFA < SymbolType, StateType >::parseTransition ( std::deque < sax::Token >::iterator & input, ExtendedNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	StateType from = AutomatonFromXMLParser::parseTransitionFrom < StateType > ( input );
	regexp::UnboundedRegExpStructure < SymbolType > inputRegexp = AutomatonFromXMLParser::parseTransitionInputRegexp < SymbolType > ( input );
	StateType to = AutomatonFromXMLParser::parseTransitionTo < StateType > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputRegexp ), std::move ( to ) );
}

template<class SymbolType, class StateType >
void ExtendedNFA < SymbolType, StateType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( ExtendedNFA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialState ( out, this->getInitialState ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( ExtendedNFA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template<class SymbolType, class StateType >
void ExtendedNFA < SymbolType, StateType >::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputRegexp ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class StateType >
alib::ObjectBase* ExtendedNFA < SymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template<class SymbolType, class StateType >
class ComponentConstraint< automaton::ExtendedNFA < SymbolType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::ExtendedNFA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for ( const std::pair < const std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > & transition : automaton.getTransitions ( ) ) {
			std::set < SymbolType > alphabet = transition.first.second.getStructure ( ).computeMinimalAlphabet ( );
			if ( alphabet.count ( symbol ) )
				return true;
		}

		return false;
	}

	static bool available ( const automaton::ExtendedNFA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::ExtendedNFA < SymbolType, StateType > &, const SymbolType & ) {
	}
};

template<class SymbolType, class StateType >
class ComponentConstraint< automaton::ExtendedNFA < SymbolType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::ExtendedNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < StateType, regexp::UnboundedRegExpStructure < SymbolType > >, std::set < StateType > > & transition : automaton.getTransitions ( ) )
			if ( ( transition.first.first == state ) || ( transition.second.find ( state ) != transition.second.end ( ) ) )
				return true;

		return false;
	}

	static bool available ( const automaton::ExtendedNFA < SymbolType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::ExtendedNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

template<class SymbolType, class StateType >
class ComponentConstraint< automaton::ExtendedNFA < SymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::ExtendedNFA < SymbolType, StateType > &, const StateType & ) {
	return false;
}

	static bool available ( const automaton::ExtendedNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::ExtendedNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

template<class SymbolType, class StateType >
class ElementConstraint< automaton::ExtendedNFA < SymbolType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::ExtendedNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::ExtendedNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* EXTENDED_NFA_H_ */
