/*
 * MultiInitialStateNFA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef MULTI_INITIAL_STATE_NFA_H_
#define MULTI_INITIAL_STATE_NFA_H_

#include <map>
#include <ostream>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include <core/components.hpp>
#include <object/UniqueObject.h>

#include "../AutomatonException.h"
#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

namespace automaton {

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
class InputAlphabet;
class States;
class FinalStates;
class InitialStates;

template < class SymbolType, class StateType >
class MultiInitialStateNFA final : public AutomatonBase, public std::Components < MultiInitialStateNFA < SymbolType, StateType >, SymbolType, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > {
protected:
	std::map < std::pair < StateType, SymbolType >, std::set < StateType > > transitions;

public:
	explicit MultiInitialStateNFA ( );
	explicit MultiInitialStateNFA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < StateType > initialStates, std::set < StateType > finalStates );
	explicit MultiInitialStateNFA ( const NFA < SymbolType, StateType > & other );
	explicit MultiInitialStateNFA ( const DFA < SymbolType, StateType > & other );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < StateType > & getInitialStates ( ) const {
		return this->template accessComponent < InitialStates > ( ).get ( );
	}

	bool addInitialState ( StateType state ) {
		return this->template accessComponent < InitialStates > ( ).add ( std::move ( state ) );
	}

	void setInitialStates ( std::set < StateType > states ) {
		this->template accessComponent < InitialStates > ( ).set ( std::move ( states ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < SymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( SymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const SymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( StateType current, SymbolType input, StateType next );

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addTransitions ( StateType current, SymbolType input, std::set < StateType > next );

	/**
	 * Removes transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const StateType & current, const SymbolType & input, const StateType & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < StateType, SymbolType >, std::set < StateType > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::map < std::pair < StateType, SymbolType >, std::set < StateType > > getTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::pair < StateType, SymbolType >, std::set < StateType > > getTransitionsToState ( const StateType & from ) const;

	/**
	 * Determines whether MultiInitialStateNFA is deterministic.
	 * FA is deterministic if and only if:
	 * \li \c contains only 1 initial state.
	 * \li \c is epsilon free. Trivial for this class
	 * \li \c size of transition function \delta (from state, input symbol) \leq 1
	 * @return true when automaton is deterministic, false otherwise
	 */
	bool isDeterministic ( ) const;

	/**
	 * Determines whether MultiInitialStateNFA is total deterministic
	 * FA is deterministic if and only if
	 * \li \c it is deterministic
	 * \li \c size of transition function \forall states and \forall input symbols \delta (from state, input symbol) = 1
	 * @return true when automaton is total deterministic, false otherwise
	 */
	bool isTotal ( ) const;

	unsigned transitionsSize ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const MultiInitialStateNFA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "MultiInitialStateNFA";

		return xmlTagName;
	}

	static MultiInitialStateNFA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, MultiInitialStateNFA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef MultiInitialStateNFA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( MultiInitialStateNFA < > ) == typeid ( MultiInitialStateNFA < SymbolType, StateType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		std::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		std::set < DefaultStateType > initialStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < InitialStates > ( ).get ( ) ) );
		std::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		MultiInitialStateNFA < > * res = new MultiInitialStateNFA < > ( std::move ( states ), std::move ( alphabet ), std::move ( initialStates ), std::move ( finalStates ) );

		for ( std::pair < std::pair < StateType, SymbolType >, std::set < StateType > > && transition : std::make_moveable_map ( transitions ) ) {
			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			DefaultSymbolType input = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( transition.first.second ) );
			std::set < DefaultStateType > targets = AutomatonNormalize::normalizeStates ( std::move ( transition.second ) );

			res->addTransitions ( std::move ( from ), std::move ( input ), std::move ( targets ) );
		}

		return res;
	}
};

} /* namespace automaton */

#include "NFA.h"
#include "DFA.h"

namespace automaton {

template < class SymbolType, class StateType >
MultiInitialStateNFA < SymbolType, StateType >::MultiInitialStateNFA ( std::set < StateType > states, std::set < SymbolType > inputAlphabet, std::set < StateType > initialStates, std::set < StateType > finalStates ) : std::Components < MultiInitialStateNFA, SymbolType, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( initialStates ), std::move ( finalStates ) ), std::tuple < > ( ) ) {
}

template < class SymbolType, class StateType >
MultiInitialStateNFA < SymbolType, StateType >::MultiInitialStateNFA ( ) : MultiInitialStateNFA ( std::set < StateType > { }, std::set < SymbolType > { }, std::set < StateType > { }, std::set < StateType > { } ) {
}

template < class SymbolType, class StateType >
MultiInitialStateNFA < SymbolType, StateType >::MultiInitialStateNFA ( const DFA < SymbolType, StateType > & other ) : MultiInitialStateNFA ( other.getStates ( ), other.getInputAlphabet ( ), { other.getInitialState ( ) }, other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) )
		transitions[transition.first].insert ( transition.second );
}

template < class SymbolType, class StateType >
MultiInitialStateNFA < SymbolType, StateType >::MultiInitialStateNFA ( const NFA < SymbolType, StateType > & other ) : MultiInitialStateNFA ( other.getStates ( ), other.getInputAlphabet ( ), { other.getInitialState ( ) }, other.getFinalStates ( ) ) {
	for ( const auto & transition : other.getTransitions ( ) )
		transitions[transition.first] = transition.second;
}

template < class SymbolType, class StateType >
AutomatonBase * MultiInitialStateNFA < SymbolType, StateType >::clone ( ) const {
	return new MultiInitialStateNFA ( * this );
}

template < class SymbolType, class StateType >
AutomatonBase * MultiInitialStateNFA < SymbolType, StateType >::plunder ( ) && {
	return new MultiInitialStateNFA ( std::move ( * this ) );
}

template < class SymbolType, class StateType >
bool MultiInitialStateNFA < SymbolType, StateType >::addTransition ( StateType from, SymbolType input, StateType to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + std::to_string ( from ) + "\" doesn't exist." );

	if ( !getInputAlphabet ( ).count ( input ) )
		throw AutomatonException ( "Input symbol \"" + std::to_string ( input ) + "\" doesn't exist." );

	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + std::to_string ( to ) + "\" doesn't exist." );

	std::pair < StateType, SymbolType > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	return transitions[std::move ( key )].insert ( std::move ( to ) ).second;
}

template < class SymbolType, class StateType >
void MultiInitialStateNFA < SymbolType, StateType >::addTransitions ( StateType from, SymbolType input, std::set < StateType > to ) {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + std::to_string ( from ) + "\" doesn't exist." );

	if ( !getInputAlphabet ( ).count ( input ) )
		throw AutomatonException ( "Input symbol \"" + std::to_string ( input ) + "\" doesn't exist." );

	if ( !std::includes ( getStates ( ).begin ( ), getStates ( ).end ( ), to.begin ( ), to.end ( ) ) )
		throw AutomatonException ( "Some target states don't exist." );

	std::pair < StateType, SymbolType > key = std::make_pair ( std::move ( from ), std::move ( input ) );

	transitions [ std::move ( key ) ].insert ( std::make_moveable_set ( to ).begin ( ), std::make_moveable_set ( to ).end ( ) );
}

template < class SymbolType, class StateType >
bool MultiInitialStateNFA < SymbolType, StateType >::removeTransition ( const StateType & from, const SymbolType & input, const StateType & to ) {
	std::pair < StateType, SymbolType > key = std::make_pair ( from, input );

	return transitions[key].erase ( to );
}

template < class SymbolType, class StateType >
const std::map < std::pair < StateType, SymbolType >, std::set < StateType > > & MultiInitialStateNFA < SymbolType, StateType >::getTransitions ( ) const {
	return transitions;
}

template < class SymbolType, class StateType >
std::map < std::pair < StateType, SymbolType >, std::set < StateType > > MultiInitialStateNFA < SymbolType, StateType >::getTransitionsFromState ( const StateType & from ) const {
	if ( !getStates ( ).count ( from ) )
		throw AutomatonException ( "State \"" + std::to_string ( from ) + "\" doesn't exist" );

	std::map < std::pair < StateType, SymbolType >, std::set < StateType > > transitionsFromState;

	for ( const std::pair < const std::pair < StateType, SymbolType >, std::set < StateType > > & transition : transitions )
		if ( transition.first.first == from )
			transitionsFromState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsFromState;
}

template < class SymbolType, class StateType >
std::map < std::pair < StateType, SymbolType >, std::set < StateType > > MultiInitialStateNFA < SymbolType, StateType >::getTransitionsToState ( const StateType & to ) const {
	if ( !getStates ( ).count ( to ) )
		throw AutomatonException ( "State \"" + std::to_string ( to ) + "\" doesn't exist" );

	std::map < std::pair < StateType, SymbolType >, std::set < StateType > > transitionsToState;

	for ( const std::pair < const std::pair < StateType, SymbolType >, std::set < StateType > > & transition : transitions )
		if ( transition.second.find ( to ) != transition.second.end ( ) )
			transitionsToState[transition.first].insert ( transition.second.begin ( ), transition.second.end ( ) );

	return transitionsToState;
}

template < class SymbolType, class StateType >
bool MultiInitialStateNFA < SymbolType, StateType >::isDeterministic ( ) const {
	for ( const std::pair < const std::pair < StateType, SymbolType >, std::set < StateType > > & transition : transitions )
		if ( transition.second.size ( ) > 1 )
			return false;

	return getInitialStates ( ).size ( ) == 1;
}

template < class SymbolType, class StateType >
bool MultiInitialStateNFA < SymbolType, StateType >::isTotal ( ) const {
	return isDeterministic ( ) && transitionsSize ( ) == getInputAlphabet ( ).size ( ) * getStates ( ).size ( );
}

template < class SymbolType, class StateType >
unsigned MultiInitialStateNFA < SymbolType, StateType >::transitionsSize ( ) const {
	int res = 0;

	for ( const auto & transition : transitions )
		res += transition.second.size ( );

	return res;
}

template < class SymbolType, class StateType >
int MultiInitialStateNFA < SymbolType, StateType >::compare ( const MultiInitialStateNFA & other ) const {
	auto first = std::tie ( getStates ( ), getInputAlphabet ( ), getInitialStates ( ), getFinalStates ( ), transitions );
	auto second = std::tie ( other.getStates ( ), other.getInputAlphabet ( ), other.getInitialStates ( ), other.getFinalStates ( ), other.getTransitions ( ) );

	static std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class StateType >
void MultiInitialStateNFA < SymbolType, StateType >::operator >>( std::ostream & out ) const {
	out << "(MultiInitialStateNFA "
	    << "states = " << getStates ( )
	    << "inputAlphabet = " << getInputAlphabet ( )
	    << "initialStates = " << getInitialStates ( )
	    << "finalStates = " << getFinalStates ( )
	    << "transitions = " << transitions
	    << ")";
}

template < class SymbolType, class StateType >
MultiInitialStateNFA < SymbolType, StateType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType, class StateType >
MultiInitialStateNFA < SymbolType, StateType > MultiInitialStateNFA < SymbolType, StateType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, MultiInitialStateNFA < SymbolType, StateType >::getXmlTagName() );

	std::set < StateType > states = AutomatonFromXMLParser::parseStates < StateType > ( input );
	std::set < SymbolType > inputSymbols = AutomatonFromXMLParser::parseInputAlphabet < SymbolType > ( input );
	std::set < StateType > initialStates = AutomatonFromXMLParser::parseInitialStates < StateType > ( input );
	std::set < StateType > finalStates = AutomatonFromXMLParser::parseFinalStates < StateType > ( input );

	MultiInitialStateNFA < SymbolType, StateType > automaton;

	automaton.setStates ( std::move ( states ) );
	automaton.setInputAlphabet ( std::move ( inputSymbols ) );
	automaton.setInitialStates ( std::move ( initialStates ) );
	automaton.setFinalStates ( std::move ( finalStates ) );

	AutomatonFromXMLParser::parseTransitions < MultiInitialStateNFA > ( input, automaton );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, MultiInitialStateNFA < SymbolType, StateType >::getXmlTagName() );
	return automaton;
}

template < class SymbolType, class StateType >
void MultiInitialStateNFA < SymbolType, StateType >::parseTransition ( std::deque < sax::Token >::iterator & input, MultiInitialStateNFA & automaton ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, "transition" );
	StateType from = AutomatonFromXMLParser::parseTransitionFrom < StateType > ( input );
	SymbolType inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol < SymbolType > ( input );
	StateType to = AutomatonFromXMLParser::parseTransitionTo < StateType > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, "transition" );

	automaton.addTransition ( std::move ( from ), std::move ( inputSymbol ), std::move ( to ) );
}

template < class SymbolType, class StateType >
void MultiInitialStateNFA < SymbolType, StateType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( MultiInitialStateNFA < SymbolType, StateType >::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	AutomatonToXMLComposer::composeStates ( out, this->getStates ( ) );
	AutomatonToXMLComposer::composeInputAlphabet ( out, this->getInputAlphabet ( ) );
	AutomatonToXMLComposer::composeInitialStates ( out, this->getInitialStates ( ) );
	AutomatonToXMLComposer::composeFinalStates ( out, this->getFinalStates ( ) );
	composeTransitions ( out );

	out.emplace_back ( MultiInitialStateNFA < SymbolType, StateType >::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class StateType >
void MultiInitialStateNFA < SymbolType, StateType >::composeTransitions ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "transitions", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & transition : this->getTransitions ( ) )
		for ( const auto & targetState : transition.second ) {
			out.emplace_back ( "transition", sax::Token::TokenType::START_ELEMENT );

			AutomatonToXMLComposer::composeTransitionFrom ( out, transition.first.first );
			AutomatonToXMLComposer::composeTransitionInputSymbol ( out, transition.first.second );
			AutomatonToXMLComposer::composeTransitionTo ( out, targetState );

			out.emplace_back ( "transition", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "transitions", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class StateType >
alib::ObjectBase* MultiInitialStateNFA < SymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::MultiInitialStateNFA < SymbolType, StateType >, SymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::MultiInitialStateNFA < SymbolType, StateType > & automaton, const SymbolType & symbol ) {
		for ( const std::pair < const std::pair < StateType, SymbolType >, std::set < StateType > > & transition : automaton.getTransitions ( ) )
			if ( transition.first.second == symbol )
				return true;

		return false;
	}

	static bool available ( const automaton::MultiInitialStateNFA < SymbolType, StateType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const automaton::MultiInitialStateNFA < SymbolType, StateType > &, const SymbolType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::MultiInitialStateNFA < SymbolType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::MultiInitialStateNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialStates ( ).count ( state ) )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for ( const std::pair < const std::pair < StateType, SymbolType >, std::set < StateType > > & transition : automaton.getTransitions ( ) )
			if ( ( transition.first.first == state ) || ( transition.second.find ( state ) != transition.second.end ( ) ) )
				return true;

		return false;
	}

	static bool available ( const automaton::MultiInitialStateNFA < SymbolType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::MultiInitialStateNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::MultiInitialStateNFA < SymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::MultiInitialStateNFA < SymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::MultiInitialStateNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::MultiInitialStateNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

template < class SymbolType, class StateType >
class ComponentConstraint< automaton::MultiInitialStateNFA < SymbolType, StateType >, StateType, automaton::InitialStates > {
public:
	static bool used ( const automaton::MultiInitialStateNFA < SymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::MultiInitialStateNFA < SymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::MultiInitialStateNFA < SymbolType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* MULTI_INITIAL_STATE_NFA_H_ */
