/*
 * DPDA.h
 *
 *  Created on: Apr 10, 2013
 *      Author: Jan Travnicek
 */

#ifndef DPDA_H_
#define DPDA_H_

#include <set>
#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class InitialSymbol;
class States;
class FinalStates;
class InitialState;

/**
 * Push Down Automaton
 *
 * $|\delta (q, a, \gamma)| \leq 1$, $\forall q, a, \gamma, q \in Q, a \in (T \cup \{\varepsilon\}), \gamma \in G^*.$
 * if $\delta (q, a, \alpha) \neq \emptyset$, $\delta (q, a, \beta) \neq \emptyset$ and $\alpha \neq \beta$, then $\alpha$ is not suffix of $\beta$ and $\beta$ is not suffix of $\alpha$ (formally $\gamma \alpha \neq \beta and \alpha \neq \gamma \beta$).
 * if $\delta(q, a, \alpha) \neq \emptyset$, $\delta (q, \varepsilon, \beta) \neq \emptyset$, then $\alpha$ is not suffix of $\beta$ and $\beta$ is not suffix of $\alpha$ (fornally $\gamma \alpha \neq \beta and \alpha \neq \gamma \beta$).
 */
template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class DPDA final : public AutomatonBase, public std::Components < DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, InputSymbolType, std::tuple < InputAlphabet>, std::tuple < >, PushdownStoreSymbolType, std::tuple < PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > {
protected:
	std::map < std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, std::vector < PushdownStoreSymbolType > >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > transitions;

public:
	explicit DPDA ( std::set < StateType > states, std::set < InputSymbolType > inputAlphabet, std::set < PushdownStoreSymbolType > pushdownStoreSymbol, StateType initialState, PushdownStoreSymbolType initialPushdownSymbol, std::set < StateType > finalStates );
	explicit DPDA ( StateType initialState, PushdownStoreSymbolType initialPushdownSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const StateType & getInitialState ( ) const {
		return this->template accessElement < InitialState > ( ).get ( );
	}

	bool setInitialState ( StateType state ) {
		return this->template accessElement < InitialState > ( ).set ( std::move ( state ) );
	}

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < PushdownStoreSymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( PushdownStoreSymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < PushdownStoreSymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < PushdownStoreSymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const PushdownStoreSymbolType & symbol ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const PushdownStoreSymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( PushdownStoreSymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < InputSymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( InputSymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < InputSymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < InputSymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const InputSymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition to the DPDA.
	 * @param transition transition to add
	 * @throws AutomatonException when some part of the transition is not present
	 * in the DPDA (state, input symbol, stack symbol) or when transition already exists
	 */
	bool addTransition ( StateType from, std::variant < EpsilonType, InputSymbolType > input, std::vector < PushdownStoreSymbolType > pop, StateType to, std::vector < PushdownStoreSymbolType > push );

	bool addTransition ( StateType from, InputSymbolType input, std::vector < PushdownStoreSymbolType > pop, StateType to, std::vector < PushdownStoreSymbolType > push );

	bool addTransition ( StateType from, std::vector < PushdownStoreSymbolType > pop, StateType to, std::vector < PushdownStoreSymbolType > push );

	/**
	 * Removes the transition from the DPDA.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition is not present in the DPDA
	 */
	bool removeTransition ( const StateType & from, const std::variant < EpsilonType, InputSymbolType > & input, const std::vector < PushdownStoreSymbolType > & pop, const StateType & to, const std::vector < PushdownStoreSymbolType > & push );

	bool removeTransition ( const StateType & from, const InputSymbolType & input, const std::vector < PushdownStoreSymbolType > & pop, const StateType & to, const std::vector < PushdownStoreSymbolType > & push );

	bool removeTransition ( const StateType & from, const std::vector < PushdownStoreSymbolType > & pop, const StateType & to, const std::vector < PushdownStoreSymbolType > & push );

	/**
	 * @return DPDA transitions
	 */
	const std::map < std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, std::vector < PushdownStoreSymbolType > >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > & getTransitions ( ) const;

	/**
	 * @return automaton transitions from state
	 */
	std::map < std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, std::vector < PushdownStoreSymbolType > >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > getTransitionsFromState ( const StateType & from ) const;

	/**
	 * @return automaton transitions to state
	 */
	std::map < std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, std::vector < PushdownStoreSymbolType > >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > getTransitionsToState ( const StateType & from ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const DPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "DPDA";

		return xmlTagName;
	}

	static DPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, DPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef DPDA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( DPDA < > ) == typeid ( DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		std::set < DefaultSymbolType > pushdownAlphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < PushdownStoreAlphabet > ( ).get ( ) ) );
		DefaultSymbolType initialSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( this->template accessElement < InitialSymbol > ( ).get ( ) ) );
		std::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		DefaultStateType initialState = AutomatonNormalize::normalizeState ( std::move ( this->template accessElement < InitialState > ( ).get ( ) ) );
		std::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		DPDA < > * res = new DPDA < > ( std::move ( states ), std::move ( alphabet ), std::move ( pushdownAlphabet ), std::move ( initialState ), std::move ( initialSymbol ), std::move ( finalStates ) );

		for ( std::pair < std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, std::vector < PushdownStoreSymbolType > >, std::pair < StateType, std::vector < PushdownStoreSymbolType > > > && transition : std::make_moveable_map ( transitions ) ) {
			std::vector < DefaultSymbolType > pop = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( std::get < 2 > ( transition.first ) ) );
			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( std::get < 0 > ( transition.first ) ) );
			std::variant < DefaultEpsilonType, DefaultSymbolType > input = AutomatonNormalize::normalizeSymbolEpsilon ( std::move ( std::get < 1 > ( transition.first ) ) );

			DefaultStateType to = AutomatonNormalize::normalizeState ( std::move ( transition.second.first ) );
			std::vector < DefaultSymbolType > push = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( transition.second.second ) );

			res->addTransition ( std::move ( from ), std::move ( input ), std::move ( pop ), std::move ( to ), std::move ( push ) );
		}

		return res;
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::DPDA ( std::set < StateType > states, std::set < InputSymbolType > inputAlphabet, std::set < PushdownStoreSymbolType > pushdownStoreAlphabet, StateType initialState, PushdownStoreSymbolType initialSymbol, std::set < StateType > finalStates ) : std::Components < DPDA, InputSymbolType, std::tuple < InputAlphabet >, std::tuple < >, PushdownStoreSymbolType, std::tuple < PushdownStoreAlphabet >, std::tuple < InitialSymbol >, StateType, std::tuple < States, FinalStates >, std::tuple < InitialState > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::make_tuple ( std::move ( initialState ) ) ) {
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::DPDA ( StateType initialState, PushdownStoreSymbolType initialPushdownSymbol) : DPDA ( std::set < StateType > { initialState }, std::set < InputSymbolType > { }, std::set < PushdownStoreSymbolType > { initialPushdownSymbol }, initialState, initialPushdownSymbol, std::set < StateType > { }) {
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
AutomatonBase* DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::clone() const {
	return new DPDA(*this);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
AutomatonBase* DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::plunder() && {
	return new DPDA(std::move(*this));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addTransition(StateType from, std::variant<EpsilonType, InputSymbolType> input, std::vector < PushdownStoreSymbolType > pop, StateType to, std::vector < PushdownStoreSymbolType > push) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is < InputSymbolType > ( ) && ! getInputAlphabet().count ( input.template get < InputSymbolType > ( ) ) ) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get < InputSymbolType > ( ) ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	for(const PushdownStoreSymbolType & popSymbol : pop) {
		if (! getPushdownStoreAlphabet().count(popSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( popSymbol ) + "\" doesn't exist.");
		}
	}

	for(const PushdownStoreSymbolType & pushSymbol : push) {
		if (! getPushdownStoreAlphabet().count(pushSymbol)) {
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pushSymbol ) + "\" doesn't exist.");
		}
	}

	std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> > key(std::move(from), std::move(input), std::move(pop));
	std::pair<StateType, std::vector<PushdownStoreSymbolType> > value = std::make_pair(std::move(to), std::move(push));

	if (transitions.find(key) != transitions.end()) {
		if(transitions.find(key)->second == value)
			return false;
		else if (std::get<1>(key).template is<InputSymbolType>())
			throw AutomatonException(
				"Transition (\"" + std::to_string ( std::get<0>(key) ) + "\", \"" + std::to_string ( std::get<1>(key).template get<InputSymbolType>() ) + "\", \"pop\") -> ?? already exists.");
		else
			throw AutomatonException(
				"Transition (\"" + std::to_string ( std::get<0>(key) ) + "\", \"" + std::to_string ( std::get<1>(key).template get<EpsilonType>() ) + "\", \"pop\") -> ?? already exists.");
	}

	if(std::get<1>(key).template is<EpsilonType>()) {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<const std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > > & transition) {
				if(std::get<0>(transition.first) == std::get<0>(key)) {
					const std::vector<PushdownStoreSymbolType>& alpha = std::get<2>(transition.first);
					const std::vector<PushdownStoreSymbolType>& beta = std::get<2>(key);

					const std::vector<PushdownStoreSymbolType>& shorter = (alpha.size() < beta.size()) ? alpha : beta;
					const std::vector<PushdownStoreSymbolType>& longer = (alpha.size() < beta.size()) ? beta : alpha;

					for(auto itS = shorter.begin(), itL = longer.begin(); itS != shorter.end(); ++itS, ++itL) {
						if(*itS != *itL) return false;
					}
					return true;
				}
				return false;
			}))
				throw AutomatonException("Conflicting transition");
	} else {
		if(std::any_of(transitions.begin(), transitions.end(), [&](const std::pair<const std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > > & transition) {
				if(std::get<0>(transition.first) == std::get<0>(key) && ( std::get<1>(transition.first) == std::get<1>(key) || std::get<1>(transition.first).template is<EpsilonType>() )) {
					const std::vector<PushdownStoreSymbolType>& alpha = std::get<2>(transition.first);
					const std::vector<PushdownStoreSymbolType>& beta = std::get<2>(key);

					const std::vector<PushdownStoreSymbolType>& shorter = (alpha.size() < beta.size()) ? alpha : beta;
					const std::vector<PushdownStoreSymbolType>& longer = (alpha.size() < beta.size()) ? beta : alpha;

					for(auto itS = shorter.begin(), itL = longer.begin(); itS != shorter.end(); ++itS, ++itL) {
						if(*itS != *itL) return false;
					}
					return true;
				}
				return false;
			}))
				throw AutomatonException("Conflicting transition");
	}

	transitions.insert(std::make_pair(std::move(key), std::move(value)) );
	return true;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addTransition(StateType from, InputSymbolType input, std::vector < PushdownStoreSymbolType > pop, StateType to, std::vector < PushdownStoreSymbolType > push) {
	std::variant<EpsilonType, InputSymbolType> inputVariant(std::move(input));
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addTransition(StateType from, std::vector < PushdownStoreSymbolType > pop, StateType to, std::vector < PushdownStoreSymbolType > push) {
	auto inputVariant = std::variant<EpsilonType, InputSymbolType>::template from < EpsilonType > ( );
	return addTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to), std::move(push));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeTransition(const StateType& from, const std::variant<EpsilonType, InputSymbolType>& input, const std::vector < PushdownStoreSymbolType > & pop, const StateType& to, const std::vector < PushdownStoreSymbolType > & push) {
	std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector < PushdownStoreSymbolType > > key(from, input, pop);
	std::pair<StateType, std::vector < PushdownStoreSymbolType > > value = std::make_pair(to, push);

	if (transitions.find(key) == transitions.end())
		return false;

	if(transitions.find(key)->second != value) {
		if (input.template is < InputSymbolType > ( ) )
			throw AutomatonException(
				"Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input.template get < InputSymbolType > ( ) ) + "\") -> \"to\" doesn't exist.");
		else
			throw AutomatonException(
				"Transition (\"" + std::to_string ( from ) + "\", \"" + std::to_string ( input.template get<EpsilonType>() ) + "\") -> \"to\" doesn't exist.");
	}

	transitions.erase(key);
	return true;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeTransition(const StateType& from, const InputSymbolType& input, const std::vector < PushdownStoreSymbolType > & pop, const StateType& to, const std::vector < PushdownStoreSymbolType > & push) {
	std::variant<EpsilonType, InputSymbolType > inputVariant(input);
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeTransition(const StateType& from, const std::vector < PushdownStoreSymbolType > & pop, const StateType& to, const std::vector < PushdownStoreSymbolType > & push) {
	auto inputVariant = std::variant<EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return removeTransition(from, inputVariant, pop, to, push);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
const std::map<std::tuple<StateType, std::variant<EpsilonType, InputSymbolType >, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::getTransitions() const {
	return transitions;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
std::map<std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > > DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::getTransitionsFromState(const StateType& from) const {
	if( ! getStates().count(from) )
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist");

	std::map<std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > > transitionsFromState;
	for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& transition : transitions) {
		if (std::get<0>(transition.first) == from) {
			transitionsFromState.insert(std::make_pair(transition.first, transition.second));
		}
	}

	return transitionsFromState;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
std::map<std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > > DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::getTransitionsToState(const StateType& to) const {
	if( ! getStates().count(to))
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist");

	std::map<std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > > transitionsToState;
	for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& transition : transitions) {
		if (transition.second.first == to) {
			transitionsToState.insert(std::make_pair(transition.first, transition.second));
		}
	}

	return transitionsToState;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
int DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::compare(const DPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialState(), getFinalStates(), getPushdownStoreAlphabet(), getInitialSymbol(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialState(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getInitialSymbol(), other.transitions);

	static std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::operator>>(std::ostream& out) const {
	out << "(DPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialState = " << getInitialState()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "initialSymbol = " << getInitialSymbol()
		<< "transitions = " << transitions
		<< ")";
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<InputSymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<InputSymbolType>(input);
	std::set<PushdownStoreSymbolType> pushdownStoreSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<PushdownStoreSymbolType>(input);
	StateType initialState = AutomatonFromXMLParser::parseInitialState<StateType>(input);
	PushdownStoreSymbolType initialPushdownStoreSymbol = AutomatonFromXMLParser::parseInitialPushdownStoreSymbol<PushdownStoreSymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	DPDA automaton(std::move(initialState), std::move(initialPushdownStoreSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(pushdownStoreSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<DPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DPDA::getXmlTagName());
	return automaton;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, DPDA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
	std::variant<EpsilonType, InputSymbolType> inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, InputSymbolType>(input);
	std::vector<PushdownStoreSymbolType> pop = AutomatonFromXMLParser::parseTransitionPop<PushdownStoreSymbolType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	std::vector<PushdownStoreSymbolType> push = AutomatonFromXMLParser::parseTransitionPush<PushdownStoreSymbolType>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to), std::move(push));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(DPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialState(out, this->getInitialState());
	AutomatonToXMLComposer::composeInitialPushdownStoreSymbol(out, this->getInitialSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(DPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back("transition", sax::Token::TokenType::START_ELEMENT);

		AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
		AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
		AutomatonToXMLComposer::composeTransitionPop(out, std::get<2>(transition.first));
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second.first);
		AutomatonToXMLComposer::composeTransitionPush(out, transition.second.second);

		out.emplace_back("transition", sax::Token::TokenType::END_ELEMENT);
	}

	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
alib::ObjectBase* DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, InputSymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const InputSymbolType & symbol ) {
		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& transition : automaton.getTransitions())
			if (std::get<1>(transition.first).template is<InputSymbolType>() && symbol == std::get<1>(transition.first).template get<InputSymbolType>())
				return true;

		return false;
	}

	static bool available ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const InputSymbolType & ) {
		return true;
	}

	static void valid ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const InputSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, PushdownStoreSymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const PushdownStoreSymbolType & symbol ) {
		if(automaton.getInitialSymbol() == symbol)
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& transition : automaton.getTransitions()) {
			const auto & popSymbols = std::get<2>(transition.first);
			const auto & pushSymbols = transition.second.second;
			if(std::contains(popSymbols.begin(), popSymbols.end(), symbol ) || std::contains(pushSymbols.begin(), pushSymbols.end(), symbol))
				return true;
		}

		return false;
	}

	static bool available ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
		return true;
	}

	static void valid ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ElementConstraint< automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, PushdownStoreSymbolType, automaton::InitialSymbol > {
public:
	static bool available ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const PushdownStoreSymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialState ( ) == state )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::tuple<StateType, std::variant<EpsilonType, InputSymbolType>, std::vector<PushdownStoreSymbolType> >, std::pair<StateType, std::vector<PushdownStoreSymbolType> > >& transition : automaton.getTransitions())
			if ( state == std::get<0>(transition.first) || transition.second.first == state )
				return true;

		return false;
	}

	static bool available ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ElementConstraint< automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::InitialState > {
public:
	static bool available ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* DPDA_H_ */
