/*
 * REAL_TIME_HEIGHT_DETERMINISTIC_NPDA.h
 *
 *  Created on: Mar 25, 2013
 *      Author: Jan Travnicek
 */

#ifndef REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_
#define REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_

#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../AutomatonBase.h"
#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

namespace automaton {

class InputAlphabet;
class PushdownStoreAlphabet;
class BottomOfTheStackSymbol;
class States;
class FinalStates;
class InitialStates;

/**
 * Represents Finite Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class RealTimeHeightDeterministicNPDA final : public AutomatonBase, public std::Components < RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, InputSymbolType, std::tuple < InputAlphabet >, std::tuple < >, PushdownStoreSymbolType, std::tuple < PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, StateType, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > {
protected:
	std::map < std::pair < StateType, std::variant < EpsilonType, InputSymbolType > >, std::set < std::pair < StateType, PushdownStoreSymbolType > > > callTransitions;
	std::map < std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType >, std::set < StateType > > returnTransitions;
	std::map < std::pair < StateType, std::variant < EpsilonType, InputSymbolType > >, std::set < StateType > > localTransitions;

public:
	explicit RealTimeHeightDeterministicNPDA ( std::set < StateType > states, std::set < InputSymbolType > inputAlphabet, std::set < PushdownStoreSymbolType > pushdownStoreSymbol, std::set < StateType > initialStates, PushdownStoreSymbolType bottomOfTheStackSymbol, std::set < StateType > finalStates );
	explicit RealTimeHeightDeterministicNPDA ( PushdownStoreSymbolType bottomOfTheStackSymbol );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getInitialStates ( ) const {
		return this->template accessComponent < InitialStates > ( ).get ( );
	}

	bool addInitialState ( StateType state ) {
		return this->template accessComponent < InitialStates > ( ).add ( std::move ( state ) );
	}

	void setInitialStates ( std::set < StateType > states ) {
		this->template accessComponent < InitialStates > ( ).set ( std::move ( states ) );
	}

	void removeInitialState ( const StateType & state ) {
		this->template accessComponent < InitialStates > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < PushdownStoreSymbolType > & getPushdownStoreAlphabet ( ) const {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).get ( );
	}

	bool addPushdownStoreSymbol ( PushdownStoreSymbolType symbol ) {
		return this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addPushdownStoreSymbols ( std::set < PushdownStoreSymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setPushdownStoreAlphabet ( std::set < PushdownStoreSymbolType > symbols ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removePushdownStoreSymbol ( const PushdownStoreSymbolType & symbol ) {
		this->template accessComponent < PushdownStoreAlphabet > ( ).remove ( symbol );
	}

	const PushdownStoreSymbolType & getBottomOfTheStackSymbol ( ) const {
		return this->template accessElement < BottomOfTheStackSymbol > ( ).get ( );
	}

	bool setBottomOfTheStackSymbol ( PushdownStoreSymbolType symbol ) {
		return this->template accessElement < BottomOfTheStackSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < InputSymbolType > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( InputSymbolType symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < InputSymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < InputSymbolType > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const InputSymbolType & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addCallTransition ( StateType current, std::variant < EpsilonType, InputSymbolType > input, StateType next, PushdownStoreSymbolType push );
	bool addCallTransition ( StateType current, StateType next, PushdownStoreSymbolType push );
	bool addCallTransition ( StateType current, InputSymbolType input, StateType next, PushdownStoreSymbolType push );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addReturnTransition ( StateType current, std::variant < EpsilonType, InputSymbolType > input, PushdownStoreSymbolType pop, StateType next );
	bool addReturnTransition ( StateType current, PushdownStoreSymbolType pop, StateType next );
	bool addReturnTransition ( StateType current, InputSymbolType input, PushdownStoreSymbolType pop, StateType next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addLocalTransition ( StateType current, std::variant < EpsilonType, InputSymbolType > input, StateType next );
	bool addLocalTransition ( StateType current, StateType next );
	bool addLocalTransition ( StateType current, InputSymbolType input, StateType next );

	/**
	 * Adds call transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addCallTransitions ( StateType current, std::variant < EpsilonType, InputSymbolType > input, std::set < std::pair < StateType, PushdownStoreSymbolType > > targets );
	void addCallTransitions ( StateType current, std::set < std::pair < StateType, PushdownStoreSymbolType > > targets );
	void addCallTransitions ( StateType current, InputSymbolType input, std::set < std::pair < StateType, PushdownStoreSymbolType > > targets );

	/**
	 * Adds return transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addReturnTransitions ( StateType current, std::variant < EpsilonType, InputSymbolType > input, PushdownStoreSymbolType pop, std::set < StateType > next );
	void addReturnTransitions ( StateType current, PushdownStoreSymbolType pop, std::set < StateType > next );
	void addReturnTransitions ( StateType current, InputSymbolType input, PushdownStoreSymbolType pop, std::set < StateType > next );

	/**
	 * Adds local transition defined by parameters to the automaton.
	 * @param current current state
	 * @param input input symbol
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	void addLocalTransitions ( StateType current, std::variant < EpsilonType, InputSymbolType > input, std::set < StateType > next );
	void addLocalTransitions ( StateType current, std::set < StateType > next );
	void addLocalTransitions ( StateType current, InputSymbolType input, std::set < StateType > next );

	/**
	 * Removes call transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeCallTransition ( const StateType & current, const std::variant < EpsilonType, InputSymbolType > & input, const StateType & next, const PushdownStoreSymbolType & push );
	bool removeCallTransition ( const StateType & current, const StateType & next, const PushdownStoreSymbolType & push );
	bool removeCallTransition ( const StateType & current, const InputSymbolType & input, const StateType & next, const PushdownStoreSymbolType & push );

	/**
	 * Removes return transition from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeReturnTransition ( const StateType & current, const std::variant < EpsilonType, InputSymbolType > & input, const PushdownStoreSymbolType & pop, const StateType & next );
	bool removeReturnTransition ( const StateType & current, const PushdownStoreSymbolType & pop, const StateType & next );
	bool removeReturnTransition ( const StateType & current, const InputSymbolType & input, const PushdownStoreSymbolType & pop, const StateType & next );

	/**
	 * Removes transition local from the automaton.
	 * @param transition transition to remove
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeLocalTransition ( const StateType & current, const std::variant < EpsilonType, InputSymbolType > & input, const StateType & next );
	bool removeLocalTransition ( const StateType & current, const StateType & next );
	bool removeLocalTransition ( const StateType & current, const InputSymbolType & input, const StateType & next );

	const std::map < std::pair < StateType, std::variant < EpsilonType, InputSymbolType > >, std::set < std::pair < StateType, PushdownStoreSymbolType > > > & getCallTransitions ( ) const;

	const std::map < std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType >, std::set < StateType > > & getReturnTransitions ( ) const;

	const std::map < std::pair < StateType, std::variant < EpsilonType, InputSymbolType > >, std::set < StateType > > & getLocalTransitions ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const RealTimeHeightDeterministicNPDA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "RealTimeHeightDeterministicNPDA";

		return xmlTagName;
	}

	static RealTimeHeightDeterministicNPDA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, RealTimeHeightDeterministicNPDA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef RealTimeHeightDeterministicNPDA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( RealTimeHeightDeterministicNPDA < > ) == typeid ( RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		std::set < DefaultSymbolType > pushdownAlphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < PushdownStoreAlphabet > ( ).get ( ) ) );
		DefaultSymbolType bottomSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( this->template accessElement < BottomOfTheStackSymbol > ( ).get ( ) ) );
		std::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		std::set < DefaultStateType > initialStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < InitialStates > ( ).get ( ) ) );
		std::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		RealTimeHeightDeterministicNPDA < > * res = new RealTimeHeightDeterministicNPDA < > ( std::move ( states ), std::move ( alphabet ), std::move ( pushdownAlphabet ), std::move ( initialStates ), std::move ( bottomSymbol ), std::move ( finalStates ) );

		for ( std::pair < std::pair < StateType, std::variant < EpsilonType, InputSymbolType > >, std::set < std::pair < StateType, PushdownStoreSymbolType > > > && transition : std::make_moveable_map ( callTransitions ) ) {
			std::set < std::pair < DefaultStateType, DefaultSymbolType > > targets;
			for ( std::pair < StateType, PushdownStoreSymbolType > && target : std::make_moveable_set ( transition.second ) )
				targets.insert ( std::make_pair ( AutomatonNormalize::normalizeState ( std::move ( target.first ) ), alphabet::SymbolNormalize::normalizeSymbol ( std::move ( target.second ) ) ) );

			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			std::variant < DefaultEpsilonType, DefaultSymbolType > input = AutomatonNormalize::normalizeSymbolEpsilon ( std::move ( transition.first.second ) );
			res->addCallTransitions ( std::move ( from ), std::move ( input ), std::move ( targets ) );
		}

		for ( std::pair < std::tuple < StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType >, std::set < StateType > > && transition : std::make_moveable_map ( returnTransitions ) ) {
			std::set < DefaultStateType > targets = AutomatonNormalize::normalizeStates ( std::move ( transition.second ) );

			DefaultSymbolType pop = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( std::get < 2 > ( transition.first ) ) );
			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( std::get < 0 > ( transition.first ) ) );
			std::variant < DefaultEpsilonType, DefaultSymbolType > input = AutomatonNormalize::normalizeSymbolEpsilon ( std::move ( std::get < 1 > ( transition.first ) ) );

			res->addReturnTransitions ( std::move ( from ), std::move ( input ), std::move ( pop ), std::move ( targets ) );
		}

		for ( std::pair < std::pair < StateType, std::variant < EpsilonType, InputSymbolType > >, std::set < StateType > > && transition : std::make_moveable_map ( localTransitions ) ) {
			std::set < DefaultStateType > targets = AutomatonNormalize::normalizeStates ( std::move ( transition.second ) );

			DefaultStateType from = AutomatonNormalize::normalizeState ( std::move ( transition.first.first ) );
			std::variant < DefaultEpsilonType, DefaultSymbolType > input = AutomatonNormalize::normalizeSymbolEpsilon ( std::move ( transition.first.second ) );

			res->addLocalTransitions ( std::move ( from ), std::move ( input ), std::move ( targets ) );
		}

		return res;
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::RealTimeHeightDeterministicNPDA ( std::set < StateType > states, std::set < InputSymbolType > inputAlphabet, std::set < PushdownStoreSymbolType > pushdownStoreAlphabet, std::set < StateType > initialStates, PushdownStoreSymbolType bottomOfTheStackSymbol, std::set < StateType > finalStates ) : std::Components < RealTimeHeightDeterministicNPDA, InputSymbolType, std::tuple < InputAlphabet >, std::tuple < >, PushdownStoreSymbolType, std::tuple < PushdownStoreAlphabet >, std::tuple < BottomOfTheStackSymbol >, StateType, std::tuple < States, InitialStates, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( pushdownStoreAlphabet ) ), std::make_tuple ( std::move ( bottomOfTheStackSymbol ) ), std::make_tuple ( std::move ( states ), std::move ( initialStates ), std::move ( finalStates ) ), std::tuple < > { } ) {
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::RealTimeHeightDeterministicNPDA(PushdownStoreSymbolType bottomOfTheStackSymbol) : RealTimeHeightDeterministicNPDA ( std::set < StateType > { }, std::set < InputSymbolType > { }, std::set < PushdownStoreSymbolType > { bottomOfTheStackSymbol }, std::set < StateType > {}, bottomOfTheStackSymbol, std::set < StateType > { }) {
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
AutomatonBase* RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::clone() const {
	return new RealTimeHeightDeterministicNPDA(*this);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
AutomatonBase* RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::plunder() && {
	return new RealTimeHeightDeterministicNPDA(std::move(*this));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addCallTransition(StateType from, std::variant < EpsilonType, InputSymbolType > input, StateType to, PushdownStoreSymbolType push) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is < InputSymbolType >() && ! getInputAlphabet().count(input.template get < InputSymbolType >())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get < InputSymbolType >() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (! getPushdownStoreAlphabet().count(push)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( push ) + "\" doesn't exist.");
	}

	std::pair<StateType, std::variant < EpsilonType, InputSymbolType >> key(std::move(from), std::move(input));
	std::pair<StateType, PushdownStoreSymbolType> value = std::make_pair(std::move(to), std::move(push));

	return callTransitions[std::move(key)].insert(std::move(value)).second;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addCallTransition(StateType from, StateType to, PushdownStoreSymbolType push) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return addCallTransition(std::move(from), std::move(inputVariant), std::move(to), std::move(push));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addCallTransition(StateType from, InputSymbolType input, StateType to, PushdownStoreSymbolType push) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(input);
	return addCallTransition(std::move(from), std::move(inputVariant), std::move(to), std::move(push));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addReturnTransition(StateType from, std::variant < EpsilonType, InputSymbolType > input, PushdownStoreSymbolType pop, StateType to) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is < InputSymbolType >() && !getInputAlphabet().count(input.template get < InputSymbolType >())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get < InputSymbolType >() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	if (! getPushdownStoreAlphabet().count(pop)) {
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pop ) + "\" doesn't exist.");
	}

	std::tuple<StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType> key(std::move(from), std::move(input), std::move(pop));

	return returnTransitions[std::move(key)].insert(std::move(to)).second;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addReturnTransition(StateType from, PushdownStoreSymbolType pop, StateType to) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return addReturnTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addReturnTransition(StateType from, InputSymbolType input, PushdownStoreSymbolType pop, StateType to) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(std::move(input));
	return addReturnTransition(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addLocalTransition(StateType from, std::variant < EpsilonType, InputSymbolType > input, StateType to) {
	if (! getStates().count(from)) {
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");
	}

	if (input.template is < InputSymbolType >() && ! getInputAlphabet().count(input.template get < InputSymbolType >())) {
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get < InputSymbolType >() ) + "\" doesn't exist.");
	}

	if (! getStates().count(to)) {
		throw AutomatonException("State \"" + std::to_string ( to ) + "\" doesn't exist.");
	}

	std::pair<StateType, std::variant < EpsilonType, InputSymbolType >> key(std::move(from), std::move(input));

	return localTransitions[std::move(key)].insert(std::move(to)).second;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addLocalTransition(StateType from, StateType to) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return addLocalTransition(std::move(from), std::move(inputVariant), std::move(to));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addLocalTransition(StateType from, InputSymbolType input, StateType to) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(std::move(input));
	return addLocalTransition(std::move(from), std::move(inputVariant), std::move(to));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addCallTransitions(StateType from, std::variant < EpsilonType, InputSymbolType > input, std::set < std::pair < StateType, PushdownStoreSymbolType > > targets ) {
	if (! getStates().count(from))
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");

	if (input.template is < InputSymbolType >() && ! getInputAlphabet().count(input.template get < InputSymbolType >()))
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get < InputSymbolType >() ) + "\" doesn't exist.");

	for ( const std::pair < StateType, PushdownStoreSymbolType > & target : targets ) {
		if ( ! getStates().count ( target.first ) )
			throw AutomatonException("State \"" + std::to_string ( target.first ) + "\" doesn't exist.");

		if ( ! getPushdownStoreAlphabet().count ( target.second ) )
			throw AutomatonException("Pushdown store symbol \"" + std::to_string ( target.second ) + "\" doesn't exist.");
	}

	std::pair < StateType, std::variant < EpsilonType, InputSymbolType > > key ( std::move ( from ), std::move ( input ) );
	callTransitions [ std::move ( key ) ].insert ( std::make_moveable_set ( targets ).begin ( ), std::make_moveable_set ( targets ).end ( ) );
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addCallTransitions(StateType from, std::set < std::pair < StateType, PushdownStoreSymbolType > > targets ) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	addCallTransitions(std::move(from), std::move(inputVariant), std::move ( targets ) );
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addCallTransitions(StateType from, InputSymbolType input, std::set < std::pair < StateType, PushdownStoreSymbolType > > targets ) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(input);
	addCallTransitions(std::move(from), std::move(inputVariant), std::move ( targets ) );
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addReturnTransitions(StateType from, std::variant < EpsilonType, InputSymbolType > input, PushdownStoreSymbolType pop, std::set < StateType > to) {
	if (! getStates().count(from))
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");

	if (input.template is < InputSymbolType >() && !getInputAlphabet().count(input.template get < InputSymbolType >()))
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get < InputSymbolType >() ) + "\" doesn't exist.");

	if ( !std::includes ( getStates ( ).begin ( ), getStates ( ).end ( ), to.begin ( ), to.end ( ) ) )
		throw AutomatonException ( "Some target states don't exist." );

	if ( ! getPushdownStoreAlphabet().count ( pop ) )
		throw AutomatonException("Pushdown store symbol \"" + std::to_string ( pop ) + "\" doesn't exist.");

	std::tuple<StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType> key(std::move(from), std::move(input), std::move(pop));
	returnTransitions[std::move(key)].insert ( std::make_moveable_set ( to ).begin ( ), std::make_moveable_set ( to ).end ( ) );
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addReturnTransitions(StateType from, PushdownStoreSymbolType pop, std::set < StateType > to) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	addReturnTransitions(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addReturnTransitions(StateType from, InputSymbolType input, PushdownStoreSymbolType pop, std::set < StateType > to) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(std::move(input));
	addReturnTransitions(std::move(from), std::move(inputVariant), std::move(pop), std::move(to));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addLocalTransitions(StateType from, std::variant < EpsilonType, InputSymbolType > input, std::set < StateType > to) {
	if (! getStates().count(from))
		throw AutomatonException("State \"" + std::to_string ( from ) + "\" doesn't exist.");

	if (input.template is < InputSymbolType >() && ! getInputAlphabet().count(input.template get < InputSymbolType >()))
		throw AutomatonException("Input symbol \"" + std::to_string ( input.template get < InputSymbolType >() ) + "\" doesn't exist.");

	if ( !std::includes ( getStates ( ).begin ( ), getStates ( ).end ( ), to.begin ( ), to.end ( ) ) )
		throw AutomatonException ( "Some target states don't exist." );

	std::pair<StateType, std::variant < EpsilonType, InputSymbolType >> key(std::move(from), std::move(input));
	localTransitions[std::move(key)].insert ( std::make_moveable_set ( to ).begin ( ), std::make_moveable_set ( to ).end ( ) );
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addLocalTransitions(StateType from, std::set < StateType > to) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	addLocalTransitions(std::move(from), std::move(inputVariant), std::move(to));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::addLocalTransitions(StateType from, InputSymbolType input, std::set < StateType > to) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(std::move(input));
	addLocalTransitions(std::move(from), std::move(inputVariant), std::move(to));
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeCallTransition(const StateType& from, const std::variant < EpsilonType, InputSymbolType >& input, const StateType& to, const PushdownStoreSymbolType& push) {
	std::pair<StateType, std::variant < EpsilonType, InputSymbolType >> key(from, input);
	std::pair<StateType, PushdownStoreSymbolType> value = std::make_pair(to, push);

	return callTransitions[key].erase(value);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeCallTransition(const StateType& from, const StateType& to, const PushdownStoreSymbolType& push) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return removeCallTransition(from, inputVariant, to, push);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeCallTransition(const StateType& from, const InputSymbolType& input, const StateType& to, const PushdownStoreSymbolType& push) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(input);
	return removeCallTransition(from, inputVariant, to, push);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeReturnTransition(const StateType& from, const std::variant < EpsilonType, InputSymbolType >& input, const PushdownStoreSymbolType& pop, const StateType& to) {
	std::tuple<StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType> key(from, input, pop);

	return returnTransitions[key].erase(to);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeReturnTransition(const StateType& from, const PushdownStoreSymbolType& pop, const StateType& to) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return removeReturnTransition(from, inputVariant, pop, to);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeReturnTransition(const StateType& from, const InputSymbolType& input, const PushdownStoreSymbolType& pop, const StateType& to) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(input);
	return removeReturnTransition(from, inputVariant, pop, to);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeLocalTransition(const StateType& from, const std::variant < EpsilonType, InputSymbolType >& input, const StateType& to) {
	std::pair<StateType, std::variant < EpsilonType, InputSymbolType >> key(from, input);

	return localTransitions[key].erase(to);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeLocalTransition(const StateType& from, const StateType& to) {
	auto inputVariant = std::variant < EpsilonType, InputSymbolType >::template from < EpsilonType > ( );
	return removeLocalTransition(from, inputVariant, to);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
bool RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::removeLocalTransition(const StateType& from, const InputSymbolType& input, const StateType& to) {
	std::variant < EpsilonType, InputSymbolType > inputVariant(input);
	return removeLocalTransition(from, inputVariant, to);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
const std::map<std::pair<StateType, std::variant < EpsilonType, InputSymbolType >>, std::set<std::pair<StateType, PushdownStoreSymbolType> > >& RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::getCallTransitions() const {
	return callTransitions;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
const std::map<std::tuple<StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::set<StateType> >& RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::getReturnTransitions() const {
	return returnTransitions;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
const std::map<std::pair<StateType, std::variant < EpsilonType, InputSymbolType >>, std::set<StateType> >& RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::getLocalTransitions() const {
	return localTransitions;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
int RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::compare(const RealTimeHeightDeterministicNPDA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getInitialStates(), getFinalStates(), getPushdownStoreAlphabet(), getBottomOfTheStackSymbol(), callTransitions, returnTransitions, localTransitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getInitialStates(), other.getFinalStates(), other.getPushdownStoreAlphabet(), other.getBottomOfTheStackSymbol(), other.callTransitions, other.returnTransitions, other.localTransitions);

	static std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::operator>>(std::ostream& out) const {
	out << "(RealTimeHeightDeterministicNPDA "
		<< "states = " << getStates()
		<< "inputAlphabet = " << getInputAlphabet()
		<< "initialStates = " << getInitialStates()
		<< "finalStates = " << getFinalStates()
		<< "pushdownStoreAlphabet = " << getPushdownStoreAlphabet()
		<< "bottomOfTheStackSymbol = " << getBottomOfTheStackSymbol()
		<< "callTransitions = " << callTransitions
		<< "returnTransitions = " << returnTransitions
		<< "localTransitions = " << localTransitions
		<< ")";
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, RealTimeHeightDeterministicNPDA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<InputSymbolType> inputSymbols = AutomatonFromXMLParser::parseInputAlphabet<InputSymbolType>(input);
	std::set<PushdownStoreSymbolType> stackSymbols = AutomatonFromXMLParser::parsePushdownStoreAlphabet<PushdownStoreSymbolType>(input);
	std::set<StateType> initialStates = AutomatonFromXMLParser::parseInitialStates<StateType>(input);
	PushdownStoreSymbolType bottomOfTheStackSymbol = AutomatonFromXMLParser::parseBottomOfTheStackSymbol<PushdownStoreSymbolType>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	RealTimeHeightDeterministicNPDA automaton(std::move(bottomOfTheStackSymbol));
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setPushdownStoreAlphabet(std::move(stackSymbols));
	automaton.setInitialStates(std::move(initialStates));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<RealTimeHeightDeterministicNPDA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, RealTimeHeightDeterministicNPDA::getXmlTagName());
	return automaton;
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, RealTimeHeightDeterministicNPDA& automaton) {
	if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "callTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		std::variant < EpsilonType, InputSymbolType > inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, InputSymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
		PushdownStoreSymbolType push = AutomatonFromXMLParser::parseTransitionSinglePush<PushdownStoreSymbolType>(input);

		automaton.addCallTransition(std::move(from), std::move(inputSymbol), std::move(to), std::move(push));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "callTransition");
	} else if(sax::FromXMLParserHelper::isToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition")) {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "returnTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		std::variant < EpsilonType, InputSymbolType > inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, InputSymbolType>(input);
		PushdownStoreSymbolType pop = AutomatonFromXMLParser::parseTransitionSinglePop<PushdownStoreSymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);

		automaton.addReturnTransition(std::move(from), std::move(inputSymbol), std::move(pop), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "returnTransition");
	} else {
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "localTransition");
		StateType from = AutomatonFromXMLParser::parseTransitionFrom<StateType>(input);
		std::variant < EpsilonType, InputSymbolType > inputSymbol = AutomatonFromXMLParser::parseTransitionInputEpsilonSymbol<EpsilonType, InputSymbolType>(input);
		StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);

		automaton.addLocalTransition(std::move(from), std::move(inputSymbol), std::move(to));
		sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "localTransition");
	}
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(RealTimeHeightDeterministicNPDA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composePushdownStoreAlphabet(out, this->getPushdownStoreAlphabet());
	AutomatonToXMLComposer::composeInitialStates(out, this->getInitialStates());
	AutomatonToXMLComposer::composeBottomOfTheStackSymbol(out, this->getBottomOfTheStackSymbol());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(RealTimeHeightDeterministicNPDA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
void RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back("transitions", sax::Token::TokenType::START_ELEMENT);
	for(const auto& transition : this->getCallTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("callTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target.first);
			AutomatonToXMLComposer::composeTransitionSinglePush(out, target.second);

			out.emplace_back("callTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getReturnTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("returnTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, std::get<0>(transition.first));
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, std::get<1>(transition.first));
			AutomatonToXMLComposer::composeTransitionSinglePop(out, std::get<2>(transition.first));
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("returnTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	for(const auto& transition : this->getLocalTransitions()) {
		for(const auto& target: transition.second) {
			out.emplace_back("localTransition", sax::Token::TokenType::START_ELEMENT);

			AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.first);
			AutomatonToXMLComposer::composeTransitionInputEpsilonSymbol(out, transition.first.second);
			AutomatonToXMLComposer::composeTransitionTo(out, target);

			out.emplace_back("localTransition", sax::Token::TokenType::END_ELEMENT);
		}
	}
	out.emplace_back("transitions", sax::Token::TokenType::END_ELEMENT);
}

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
alib::ObjectBase* RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, InputSymbolType, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const InputSymbolType & symbol ) {
		for (const std::pair<const std::pair<StateType, std::variant < EpsilonType, InputSymbolType >>, std::set<std::pair<StateType, PushdownStoreSymbolType> > >& callTransition : automaton.getCallTransitions())
			if (callTransition.first.second.template is < InputSymbolType >() && symbol == callTransition.first.second.template get < InputSymbolType >())
				return true;

		for (const std::pair<const std::tuple<StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::set<StateType> >& returnTransition : automaton.getReturnTransitions())
			if (std::get<1>(returnTransition.first).template is < InputSymbolType >() && symbol == std::get<1>(returnTransition.first).template get < InputSymbolType >())
				return true;

		for (const std::pair<const std::pair<StateType, std::variant < EpsilonType, InputSymbolType >>, std::set<StateType> >& localTransition : automaton.getLocalTransitions())
			if (localTransition.first.second.template is < InputSymbolType >() && symbol == localTransition.first.second.template get < InputSymbolType >())
				return true;

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const InputSymbolType & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const InputSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, PushdownStoreSymbolType, automaton::PushdownStoreAlphabet > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const PushdownStoreSymbolType & symbol ) {
		if(automaton.getBottomOfTheStackSymbol() == symbol)
			return true;

		for (const std::pair<const std::pair<StateType, std::variant < EpsilonType, InputSymbolType >>, std::set<std::pair<StateType, PushdownStoreSymbolType> > >& callTransition : automaton.getCallTransitions())
			for(const std::pair<StateType, PushdownStoreSymbolType>& to : callTransition.second)
				if (symbol == to.second)
					return true;

		for (const std::pair<const std::tuple<StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::set<StateType> >& returnTransition : automaton.getReturnTransitions())
			if (symbol == std::get<2>(returnTransition.first))
				return true;

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ElementConstraint< automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, PushdownStoreSymbolType, automaton::BottomOfTheStackSymbol > {
public:
	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const PushdownStoreSymbolType & symbol ) {
		return automaton.template accessComponent < automaton::PushdownStoreAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const PushdownStoreSymbolType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getInitialStates ( ).count ( state ) )
			return true;

		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<StateType, std::variant < EpsilonType, InputSymbolType >>, std::set<std::pair<StateType, PushdownStoreSymbolType> > >& callTransition : automaton.getCallTransitions()) {
			if (state == callTransition.first.first)
				return true;

			for(const std::pair<StateType, PushdownStoreSymbolType>& target : callTransition.second)
				if(target.first == state)
					return true;
		}

		for (const std::pair<const std::tuple<StateType, std::variant < EpsilonType, InputSymbolType >, PushdownStoreSymbolType>, std::set<StateType> >& returnTransition : automaton.getReturnTransitions()) {
			if (state == std::get<0>(returnTransition.first))
				return true;

			for(const StateType& target : returnTransition.second)
				if(target == state)
					return true;
		}

		for (const std::pair<const std::pair<StateType, std::variant < EpsilonType, InputSymbolType >>, std::set<StateType> >& localTransition : automaton.getLocalTransitions()) {
			if (state == localTransition.first.first)
				return true;

			for(const StateType& target : localTransition.second)
				if(target == state)
					return true;
		}

		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

template < class InputSymbolType, class EpsilonType, class PushdownStoreSymbolType, class StateType >
class ComponentConstraint< automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType >, StateType, automaton::InitialStates > {
public:
	static bool used ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::RealTimeHeightDeterministicNPDA < InputSymbolType, EpsilonType, PushdownStoreSymbolType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* REAL_TIME_HEIGHT_DETERMINISTIC_NPDA_H_ */
