/*
 * DFTA.h
 *
 *  Created on: Apr 14, 2015
 *      Author: Stepan Plachy
 */

#ifndef DFTA_H_
#define DFTA_H_

#include <map>
#include <vector>
#include <ostream>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include "alphabet/ranked_symbol.hpp"

#include "../AutomatonFeatures.h"
#include "../AutomatonException.h"
#include "../AutomatonBase.h"
#include "../common/AutomatonFromXMLParser.h"
#include "../common/AutomatonToXMLComposer.h"
#include "../common/AutomatonNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

namespace automaton {

class InputAlphabet;
class States;
class FinalStates;

/**
 * Represents Finite Tree Automaton.
 * Can store nondeterministic finite automaton without epsilon transitions.
 */
template<class SymbolType, class RankType, class StateType >
class DFTA final : public AutomatonBase, public std::Components < DFTA < SymbolType, RankType, StateType >, std::ranked_symbol < SymbolType, RankType >, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, FinalStates >, std::tuple < > > {
	std::map < std::pair < std::ranked_symbol < SymbolType, RankType >, std::vector < StateType > >, StateType > transitions;

public:
	explicit DFTA ( );
	explicit DFTA ( std::set < StateType > states, std::set < std::ranked_symbol < SymbolType, RankType > > inputAlphabet, std::set < StateType > finalStates );

	virtual AutomatonBase * clone ( ) const;

	virtual AutomatonBase * plunder ( ) &&;

	const std::set < StateType > & getStates ( ) const {
		return this->template accessComponent < States > ( ).get ( );
	}

	bool addState ( StateType state ) {
		return this->template accessComponent < States > ( ).add ( std::move ( state ) );
	}

	void setStates ( std::set < StateType > states ) {
		this->template accessComponent < States > ( ).set ( std::move ( states ) );
	}

	void removeState ( const StateType & state ) {
		this->template accessComponent < States > ( ).remove ( state );
	}

	const std::set < StateType > & getFinalStates ( ) const {
		return this->template accessComponent < FinalStates > ( ).get ( );
	}

	bool addFinalState ( StateType state ) {
		return this->template accessComponent < FinalStates > ( ).add ( std::move ( state ) );
	}

	void setFinalStates ( std::set < StateType > states ) {
		this->template accessComponent < FinalStates > ( ).set ( std::move ( states ) );
	}

	void removeFinalState ( const StateType & state ) {
		this->template accessComponent < FinalStates > ( ).remove ( state );
	}

	const std::set < std::ranked_symbol < SymbolType, RankType > > & getInputAlphabet ( ) const {
		return this->template accessComponent < InputAlphabet > ( ).get ( );
	}

	bool addInputSymbol ( std::ranked_symbol < SymbolType, RankType > symbol ) {
		return this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void addInputSymbols ( std::set < std::ranked_symbol < SymbolType, RankType > > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).add ( std::move ( symbols ) );
	}

	void setInputAlphabet ( std::set < std::ranked_symbol < SymbolType, RankType > > symbols ) {
		this->template accessComponent < InputAlphabet > ( ).set ( std::move ( symbols ) );
	}

	void removeInputSymbol ( const std::ranked_symbol < SymbolType, RankType > & symbol ) {
		this->template accessComponent < InputAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Adds transition defined by parameters to the automaton.
	 * @param current current symbol in node
	 * @param children states of children of current node
	 * @param next next state
	 * @throws AutomatonException when transition already exists or when transition contains state or symbol not present in the automaton
	 */
	bool addTransition ( std::ranked_symbol < SymbolType, RankType > current, std::vector < StateType > children, StateType next );

	/**
	 * Removes transition from the automaton.
	 * @throws AutomatonException when transition doesn't exists.
	 */
	bool removeTransition ( const std::ranked_symbol < SymbolType, RankType > symbol, const std::vector < StateType > & states, const StateType & next );

	/**
	 * @return automaton transitions
	 */
	const std::map < std::pair < std::ranked_symbol < SymbolType, RankType >, std::vector < StateType > >, StateType > & getTransitions ( ) const { return transitions; }

	std::set < StateType > getTransitionRightSide ( const std::ranked_symbol < SymbolType, RankType > symbol, const std::vector < StateType > & states );

	unsigned transitionsSize ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const DFTA & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "DFTA";

		return xmlTagName;
	}

	static DFTA parse ( std::deque < sax::Token >::iterator & input );
	static void parseTransition ( std::deque < sax::Token >::iterator & input, DFTA & automaton );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeTransitions ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef DFTA < > normalized_type;

	virtual AutomatonBase * normalize ( ) && {
		if ( typeid ( DFTA < > ) == typeid ( DFTA < SymbolType, RankType, StateType > ) )
			return this;

		std::set < std::ranked_symbol < DefaultSymbolType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( this->template accessComponent < InputAlphabet > ( ).get ( ) ) );
		std::set < DefaultStateType > states = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < States > ( ).get ( ) ) );
		std::set < DefaultStateType > finalStates = AutomatonNormalize::normalizeStates ( std::move ( this->template accessComponent < FinalStates > ( ).get ( ) ) );

		DFTA < > * res = new DFTA < > ( std::move ( states ), std::move ( alphabet ), std::move ( finalStates ) );

		for ( std::pair < std::pair < std::ranked_symbol < SymbolType, RankType >, std::vector < StateType > >, StateType > && transition : std::make_moveable_map ( transitions ) ) {
			std::ranked_symbol < DefaultSymbolType, DefaultRankType > input = alphabet::SymbolNormalize::normalizeRankedSymbol ( std::move ( transition.first.first ) );
			std::vector < DefaultStateType > from = AutomatonNormalize::normalizeStates ( std::move ( transition.first.second ) );
			DefaultStateType to = AutomatonNormalize::normalizeState ( std::move ( transition.second ) );

			res->addTransition ( std::move ( input ), std::move ( from ), std::move ( to ) );
		}

		return res;
	}
};

} /* namespace automaton */

namespace automaton {

template<class SymbolType, class RankType, class StateType >
DFTA < SymbolType, RankType, StateType >::DFTA ( std::set < StateType > states, std::set < std::ranked_symbol < SymbolType, RankType > > inputAlphabet, std::set < StateType > finalStates ) : std::Components < DFTA, std::ranked_symbol < SymbolType, RankType >, std::tuple < InputAlphabet >, std::tuple < >, StateType, std::tuple < States, FinalStates >, std::tuple < > > ( std::make_tuple ( std::move ( inputAlphabet ) ), std::tuple < > ( ), std::make_tuple ( std::move ( states ), std::move ( finalStates ) ), std::tuple < > ( ) ) {
}

template<class SymbolType, class RankType, class StateType >
DFTA < SymbolType, RankType, StateType >::DFTA() : DFTA ( std::set < StateType > { }, std::set < std::ranked_symbol < SymbolType, RankType > > { }, std::set < StateType > { } ) {
}

template<class SymbolType, class RankType, class StateType >
AutomatonBase* DFTA < SymbolType, RankType, StateType >::clone() const {
	return new DFTA(*this);
}

template<class SymbolType, class RankType, class StateType >
AutomatonBase* DFTA < SymbolType, RankType, StateType >::plunder() && {
	return new DFTA(std::move(*this));
}

template<class SymbolType, class RankType, class StateType >
bool DFTA < SymbolType, RankType, StateType >::addTransition ( std::ranked_symbol < SymbolType, RankType > symbol, std::vector<StateType> prevStates, StateType next) {
	if ( prevStates.size() != ( size_t ) symbol.getRank() )
		throw AutomatonException("Number of states doesn't match rank of the symbol");

	if (! getInputAlphabet().count(symbol))
		throw AutomatonException("Input symbol \"" + std::to_string ( symbol ) + "\" doesn't exist.");

	if (! getStates().count(next))
		throw AutomatonException("State \"" + std::to_string ( next ) + "\" doesn't exist.");

	for ( const StateType & it : prevStates) {
		if (! getStates().count(it))
			throw AutomatonException("State \"" + std::to_string ( it ) + "\" doesn't exist.");
	}

	std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> > key = std::make_pair ( std::move ( symbol ), std::move ( prevStates ) );
	if ( transitions.find ( key ) != transitions.end ( ) ) {
		if ( transitions.find ( key )->second == next )
			return false;
		else
			throw AutomatonException("Transition already exists");
	}

	transitions.insert ( std::make_pair ( std::move ( key ), std::move ( next ) ) );
	return true;
}

template<class SymbolType, class RankType, class StateType >
bool DFTA < SymbolType, RankType, StateType >::removeTransition(const std::ranked_symbol < SymbolType, RankType > symbol, const std::vector<StateType> & states, const StateType & next) {
	std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> > key = std::make_pair(symbol, states);

	if ( transitions.find ( key ) == transitions.end ( ) )
		return false;

	if ( transitions.find ( key )->second != next )
		throw AutomatonException("Transition does not exist");

	transitions.erase(key);
	return true;
}

template<class SymbolType, class RankType, class StateType >
int DFTA < SymbolType, RankType, StateType >::compare(const DFTA& other) const {
	auto first = std::tie(getStates(), getInputAlphabet(), getFinalStates(), transitions);
	auto second = std::tie(other.getStates(), other.getInputAlphabet(), other.getFinalStates(), other.transitions);

	static std::compare<decltype(first)> comp;
	return comp(first, second);
}

template<class SymbolType, class RankType, class StateType >
void DFTA < SymbolType, RankType, StateType >::operator>>(std::ostream& out) const {
	out << "(DFTA "
		<< " states = " << getStates()
		<< " inputAlphabet = " << getInputAlphabet()
		<< " finalStates = " << getFinalStates()
		<< " transitions = " << transitions
		<< ")";
}

template<class SymbolType, class RankType, class StateType >
DFTA < SymbolType, RankType, StateType >::operator std::string () const {
	std::stringstream ss;
	ss << *this;
	return ss.str();
}

template<class SymbolType, class RankType, class StateType >
DFTA < SymbolType, RankType, StateType > DFTA < SymbolType, RankType, StateType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, DFTA::getXmlTagName());

	std::set<StateType> states = AutomatonFromXMLParser::parseStates<StateType>(input);
	std::set<std::ranked_symbol < SymbolType, RankType >> inputSymbols = AutomatonFromXMLParser::parseRankedInputAlphabet<std::ranked_symbol < SymbolType, RankType >>(input);
	std::set<StateType> finalStates = AutomatonFromXMLParser::parseFinalStates<StateType>(input);

	DFTA < SymbolType, RankType, StateType > automaton;
	automaton.setStates(std::move(states));
	automaton.setInputAlphabet(std::move(inputSymbols));
	automaton.setFinalStates(std::move(finalStates));

	AutomatonFromXMLParser::parseTransitions<DFTA>(input, automaton);

	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, DFTA::getXmlTagName());
	return automaton;
}

template<class SymbolType, class RankType, class StateType >
void DFTA < SymbolType, RankType, StateType >::parseTransition(std::deque<sax::Token>::iterator& input, DFTA& automaton) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, "transition");
	std::ranked_symbol < SymbolType, RankType > inputSymbol = AutomatonFromXMLParser::parseTransitionInputSymbol<std::ranked_symbol < SymbolType, RankType >>(input);
	std::vector<StateType> from = AutomatonFromXMLParser::parseTransitionFromMultiple<StateType>(input);
	StateType to = AutomatonFromXMLParser::parseTransitionTo<StateType>(input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, "transition");

	automaton.addTransition(std::move(inputSymbol), std::move(from), std::move(to));
}

template<class SymbolType, class RankType, class StateType >
void DFTA < SymbolType, RankType, StateType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(DFTA::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);

	AutomatonToXMLComposer::composeStates(out, this->getStates());
	AutomatonToXMLComposer::composeRankedInputAlphabet(out, this->getInputAlphabet());
	AutomatonToXMLComposer::composeFinalStates(out, this->getFinalStates());
	composeTransitions(out);

	out.emplace_back(DFTA::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template<class SymbolType, class RankType, class StateType >
void DFTA < SymbolType, RankType, StateType >::composeTransitions(std::deque<sax::Token>& out) const {
	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::START_ELEMENT));
	for(const auto& transition : this->getTransitions()) {
		out.emplace_back(sax::Token("transition", sax::Token::TokenType::START_ELEMENT));

		AutomatonToXMLComposer::composeTransitionInputSymbol(out, transition.first.first);
		AutomatonToXMLComposer::composeTransitionFrom(out, transition.first.second);
		AutomatonToXMLComposer::composeTransitionTo(out, transition.second);

		out.emplace_back(sax::Token("transition", sax::Token::TokenType::END_ELEMENT));
	}

	out.emplace_back(sax::Token("transitions", sax::Token::TokenType::END_ELEMENT));
}

template < class SymbolType, class RankType, class StateType >
alib::ObjectBase* DFTA < SymbolType, RankType, StateType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace automaton */

namespace std {

template<class SymbolType, class RankType, class StateType >
class ComponentConstraint< automaton::DFTA < SymbolType, RankType, StateType >, std::ranked_symbol < SymbolType, RankType >, automaton::InputAlphabet > {
public:
	static bool used ( const automaton::DFTA < SymbolType, RankType, StateType > & automaton, const std::ranked_symbol < SymbolType, RankType > & symbol ) {
		for (const std::pair<const std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> >, StateType>& t : automaton.getTransitions())
			if (t.first.first == symbol)
				return true;

		return false;
	}

	static bool available ( const automaton::DFTA < SymbolType, RankType, StateType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const automaton::DFTA < SymbolType, RankType, StateType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
	}
};

template<class SymbolType, class RankType, class StateType >
class ComponentConstraint< automaton::DFTA < SymbolType, RankType, StateType >, StateType, automaton::States > {
public:
	static bool used ( const automaton::DFTA < SymbolType, RankType, StateType > & automaton, const StateType & state ) {
		if ( automaton.getFinalStates ( ).count ( state ) )
			return true;

		for (const std::pair<const std::pair<std::ranked_symbol < SymbolType, RankType >, std::vector<StateType> >, StateType>& t : automaton.getTransitions())
			if(std::contains(t.first.second.begin(), t.first.second.end(), state ) || t.second == state)
				return true;

		return false;
	}

	static bool available ( const automaton::DFTA < SymbolType, RankType, StateType > &, const StateType & ) {
		return true;
	}

	static void valid ( const automaton::DFTA < SymbolType, RankType, StateType > &, const StateType & ) {
	}
};

template<class SymbolType, class RankType, class StateType >
class ComponentConstraint< automaton::DFTA < SymbolType, RankType, StateType >, StateType, automaton::FinalStates > {
public:
	static bool used ( const automaton::DFTA < SymbolType, RankType, StateType > &, const StateType & ) {
		return false;
	}

	static bool available ( const automaton::DFTA < SymbolType, RankType, StateType > & automaton, const StateType & state ) {
		return automaton.template accessComponent < automaton::States > ( ).get ( ).count ( state );
	}

	static void valid ( const automaton::DFTA < SymbolType, RankType, StateType > &, const StateType & ) {
	}
};

} /* namespace std */

#endif /* DFTA_H_ */
