/*
 * LeftLG.h
 *
 *  Created on: Nov 17, 2013
 *      Author: Jan Travnicek
 */

#ifndef LEFT_LG_H_
#define LEFT_LG_H_

#include <map>
#include <vector>
#include <variant>
#include <algorithm>
#include <sstream>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "../GrammarBase.h"
#include "../GrammarFeatures.h"
#include "../GrammarException.h"
#include "../common/GrammarFromXMLParser.h"
#include "../common/GrammarToXMLComposer.h"
#include "../common/GrammarNormalize.h"
#include <alphabet/common/SymbolNormalize.h>

namespace grammar {

/**
 * Left linear grammar in chomsky normal form. Type 3 in Chomsky hierarchy. Produces regular languages.
 */
class TerminalAlphabet;
class NonterminalAlphabet;
class InitialSymbol;

template < class SymbolType >
class LeftLG final : public GrammarBase, public std::Components < LeftLG < SymbolType >, SymbolType, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > {
	std::map < SymbolType, std::set < std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > > > rules;

public:
	explicit LeftLG ( SymbolType initialSymbol );

	explicit LeftLG ( std::set < SymbolType > nonTerminalSymbols, std::set < SymbolType > terminalSymbols, SymbolType initialSymbol );

	virtual GrammarBase * clone ( ) const;

	virtual GrammarBase * plunder ( ) &&;

	bool addRule ( SymbolType leftHandSide, std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > rightHandSide );
	bool addRule ( SymbolType leftHandSide, std::vector < SymbolType > rightHandSide );
	bool addRule ( SymbolType leftHandSide, std::pair < SymbolType, std::vector < SymbolType > > rightHandSide );

	void addRules ( SymbolType leftHandSide, std::set < std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > > rightHandSide );

	const std::map < SymbolType, std::set < std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > > > & getRules ( ) const;

	bool removeRule ( const SymbolType & leftHandSide, const std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > & rightHandSide );
	bool removeRule ( const SymbolType & leftHandSide, const std::vector < SymbolType > & rightHandSide );
	bool removeRule ( const SymbolType & leftHandSide, const std::pair < SymbolType, std::vector < SymbolType > > & rightHandSide );

	bool addRawRule ( SymbolType leftHandSide, std::vector < SymbolType > rightHandSide );

	std::map < SymbolType, std::set < std::vector < SymbolType > > > getRawRules ( ) const;

	bool removeRawRule ( const SymbolType & leftHandSide, const std::vector < SymbolType > & rightHandSide );

	const SymbolType & getInitialSymbol ( ) const {
		return this->template accessElement < InitialSymbol > ( ).get ( );
	}

	bool setInitialSymbol ( SymbolType symbol ) {
		return this->template accessElement < InitialSymbol > ( ).set ( std::move ( symbol ) );
	}

	const std::set < SymbolType > & getNonterminalAlphabet ( ) const {
		return this->template accessComponent < NonterminalAlphabet > ( ).get ( );
	}

	bool addNonterminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < NonterminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setNonterminalAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < NonterminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	const std::set < SymbolType > & getTerminalAlphabet ( ) const {
		return this->template accessComponent < TerminalAlphabet > ( ).get ( );
	}

	bool addTerminalSymbol ( SymbolType symbol ) {
		return this->template accessComponent < TerminalAlphabet > ( ).add ( std::move ( symbol ) );
	}

	void setTerminalAlphabet ( std::set < SymbolType > symbols ) {
		this->template accessComponent < TerminalAlphabet > ( ).set ( std::move ( symbols ) );
	}

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const LeftLG & other ) const;

	virtual void operator >>( std::ostream & os ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "LeftLG";

		return xmlTagName;
	}

	static LeftLG parse ( std::deque < sax::Token >::iterator & input );
	static void parseRule ( std::deque < sax::Token >::iterator & input, LeftLG & grammar );

	void compose ( std::deque < sax::Token > & out ) const;
	void composeRules ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef LeftLG < > normalized_type;

	virtual GrammarBase * normalize ( ) && {
		if ( typeid ( LeftLG < > ) == typeid ( LeftLG < SymbolType > ) )
			return this;

		std::set < DefaultSymbolType > nonterminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < NonterminalAlphabet > ( ).get ( ) ) );
		std::set < DefaultSymbolType > terminals = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < TerminalAlphabet > ( ).get ( ) ) );
		DefaultSymbolType initialSymbol = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( this->template accessElement < InitialSymbol > ( ).get ( ) ) );

		LeftLG < > * res = new LeftLG < > ( std::move ( nonterminals ), std::move ( terminals ), std::move ( initialSymbol ) );

		for ( std::pair < SymbolType, std::set < std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > > > && rule : std::make_moveable_map ( rules ) ) {

			std::set < std::variant < std::vector < DefaultSymbolType >, std::pair < DefaultSymbolType, std::vector < DefaultSymbolType > > > > rhs;
			for ( std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > && target : std::make_moveable_set ( rule.second ) )
				rhs.insert ( GrammarNormalize::normalizeRHS ( std::move ( target ) ) );

			DefaultSymbolType lhs = alphabet::SymbolNormalize::normalizeSymbol ( std::move ( rule.first ) );

			res->addRules ( std::move ( lhs ), std::move ( rhs ) );
		}

		return res;
	}

};

template < class SymbolType >
LeftLG < SymbolType >::LeftLG ( SymbolType initialSymbol ) : LeftLG ( std::set < SymbolType > { initialSymbol }, std::set < SymbolType > ( ), initialSymbol ) {
}

template < class SymbolType >
LeftLG < SymbolType >::LeftLG ( std::set < SymbolType > nonterminalAlphabet, std::set < SymbolType > terminalAlphabet, SymbolType initialSymbol ) : std::Components < LeftLG, SymbolType, std::tuple < TerminalAlphabet, NonterminalAlphabet >, std::tuple < InitialSymbol > > ( std::make_tuple ( std::move ( terminalAlphabet ), std::move ( nonterminalAlphabet ) ), std::make_tuple ( std::move ( initialSymbol ) ) ) {
}

template < class SymbolType >
GrammarBase * LeftLG < SymbolType >::clone ( ) const {
	return new LeftLG ( * this );
}

template < class SymbolType >
GrammarBase * LeftLG < SymbolType >::plunder ( ) && {
	return new LeftLG ( std::move ( * this ) );
}

template < class SymbolType >
bool LeftLG < SymbolType >::addRule ( SymbolType leftHandSide, std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	if ( rightHandSide.template is < std::vector < SymbolType > > ( ) ) {
		for ( const auto & symbol : rightHandSide.template get < std::vector < SymbolType > > ( ) )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + std::to_string ( symbol ) + " is not a terminal symbol" );
	} else {
		const std::pair < SymbolType, std::vector < SymbolType > > & rhs = rightHandSide.template get < std::pair < SymbolType, std::vector < SymbolType > > > ( );

		if ( !getNonterminalAlphabet ( ).count ( rhs.first ) )
			throw GrammarException ( "Symbol " + std::to_string ( rhs.first ) + " is not a nonterminal symbol" );

		for ( const auto & symbol : rhs.second )
			if ( !getTerminalAlphabet ( ).count ( symbol ) )
				throw GrammarException ( "Symbol " + std::to_string ( symbol ) + " is not a terminal symbol" );
	}

	return rules [ std::move ( leftHandSide ) ].insert ( std::move ( rightHandSide ) ).second;
}

template < class SymbolType >
bool LeftLG < SymbolType >::addRule ( SymbolType leftHandSide, std::vector < SymbolType > rightHandSide ) {
	std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

template < class SymbolType >
bool LeftLG < SymbolType >::addRule ( SymbolType leftHandSide, std::pair < SymbolType, std::vector < SymbolType > > rightHandSide ) {
	std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > rhs ( std::move ( rightHandSide ) );

	return addRule ( std::move ( leftHandSide ), std::move ( rhs ) );
}

template < class SymbolType >
void LeftLG < SymbolType >::addRules ( SymbolType leftHandSide, std::set < std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > > rightHandSide ) {
	if ( !getNonterminalAlphabet ( ).count ( leftHandSide ) )
		throw GrammarException ( "Rule must rewrite nonterminal symbol" );

	for ( const std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > & element : rightHandSide ) {
		if ( element.template is < std::vector < SymbolType > > ( ) ) {
			for ( const auto & symbol : element.template get < std::vector < SymbolType > > ( ) )
				if ( !getTerminalAlphabet ( ).count ( symbol ) )
					throw GrammarException ( "Symbol " + std::to_string ( symbol ) + " is not a terminal symbol" );

		} else {
			const std::pair < SymbolType, std::vector < SymbolType > > & rhs = element.template get < std::pair < SymbolType, std::vector < SymbolType > > > ( );

			if ( !getNonterminalAlphabet ( ).count ( rhs.first ) )
				throw GrammarException ( "Symbol " + std::to_string ( rhs.first ) + " is not a nonterminal symbol" );

			for ( const auto & symbol : rhs.second )
				if ( !getTerminalAlphabet ( ).count ( symbol ) )
					throw GrammarException ( "Symbol " + std::to_string ( symbol ) + " is not a terminal symbol" );

		}
	}

	rules [ std::move ( leftHandSide ) ].insert ( std::make_moveable_set ( rightHandSide ).begin ( ), std::make_moveable_set ( rightHandSide ).end ( ) );
}

template < class SymbolType >
const std::map < SymbolType, std::set < std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > > > & LeftLG < SymbolType >::getRules ( ) const {
	return rules;
}

template < class SymbolType >
bool LeftLG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > & rightHandSide ) {
	return rules[leftHandSide].erase ( rightHandSide );
}

template < class SymbolType >
bool LeftLG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const std::vector < SymbolType > & rightHandSide ) {
	std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

template < class SymbolType >
bool LeftLG < SymbolType >::removeRule ( const SymbolType & leftHandSide, const std::pair < SymbolType, std::vector < SymbolType > > & rightHandSide ) {
	std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > rhs ( rightHandSide );

	return removeRule ( leftHandSide, rhs );
}

template < class SymbolType >
bool LeftLG < SymbolType >::addRawRule ( SymbolType leftHandSide, std::vector < SymbolType > rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 )
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
	else if ( getNonterminalAlphabet ( ).count ( rightHandSide[0] ) )
		return addRule ( std::move ( leftHandSide ), std::make_pair ( std::move ( rightHandSide[0] ), std::vector < SymbolType > ( std::make_move_iterator ( rightHandSide.begin ( ) ), std::make_move_iterator ( rightHandSide.end ( ) ) ) ) );
	else
		return addRule ( std::move ( leftHandSide ), std::move ( rightHandSide ) );
}

template < class SymbolType >
std::map < SymbolType, std::set < std::vector < SymbolType > > > LeftLG < SymbolType >::getRawRules ( ) const {
	std::map < SymbolType, std::set < std::vector < SymbolType > > > res;

	for ( const auto & rule : getRules ( ) )
		for ( const auto & rhs : rule.second ) {
			if ( rhs.template is < std::vector < SymbolType > > ( ) ) {
				res[rule.first].insert ( rhs.template get < std::vector < SymbolType > > ( ) );
			} else {
				const auto & rhsTuple = rhs.template get < std::pair < SymbolType, std::vector < SymbolType > > > ( );
				std::vector < SymbolType > tmp { rhsTuple.first };
				tmp.insert ( tmp.end ( ), rhsTuple.second.begin ( ), rhsTuple.second.end ( ) );
				res[rule.first].insert ( std::move ( tmp ) );
			}
		}

	return res;
}

template < class SymbolType >
bool LeftLG < SymbolType >::removeRawRule ( const SymbolType & leftHandSide, const std::vector < SymbolType > & rightHandSide ) {
	if ( rightHandSide.size ( ) == 0 )
		return removeRule ( leftHandSide, rightHandSide );
	else if ( getNonterminalAlphabet ( ).count ( rightHandSide[0] ) )
		return removeRule ( leftHandSide, std::make_pair ( rightHandSide[0], std::vector < SymbolType > ( rightHandSide.begin ( ) + 1, rightHandSide.end ( ) ) ) );
	else
		return removeRule ( leftHandSide, rightHandSide );
}

template < class SymbolType >
int LeftLG < SymbolType >::compare ( const LeftLG & other ) const {
	auto first = std::tie ( getTerminalAlphabet ( ), getNonterminalAlphabet ( ), getInitialSymbol ( ), rules );
	auto second = std::tie ( other.getTerminalAlphabet ( ), other.getNonterminalAlphabet ( ), other.getInitialSymbol ( ), other.rules );

	static std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
void LeftLG < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(LeftLG " << "nonterminalAlphabet = "
	    << getNonterminalAlphabet ( ) << "terminalAlphabet = "
	    << getTerminalAlphabet ( ) << "initialSymbol = "
	    << getInitialSymbol ( ) << "rules = "
	    << rules << ")";
}

template < class SymbolType >
LeftLG < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
LeftLG < SymbolType > LeftLG < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, LeftLG::getXmlTagName() );

	std::set < SymbolType > nonterminalAlphabet = GrammarFromXMLParser::parseNonterminalAlphabet < SymbolType > ( input );
	std::set < SymbolType > terminalAlphabet = GrammarFromXMLParser::parseTerminalAlphabet < SymbolType > ( input );
	SymbolType initialSymbol = GrammarFromXMLParser::parseInitialSymbol < SymbolType > ( input );

	LeftLG < SymbolType > grammar ( std::move ( initialSymbol ) );

	grammar.setNonterminalAlphabet ( std::move ( nonterminalAlphabet ) );
	grammar.setTerminalAlphabet ( std::move ( terminalAlphabet ) );

	GrammarFromXMLParser::parseRules ( input, grammar );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, LeftLG::getXmlTagName() );
	return grammar;
}

template < class SymbolType >
void LeftLG < SymbolType >::parseRule ( std::deque < sax::Token >::iterator & input, LeftLG & grammar ) {
	SymbolType lhs = GrammarFromXMLParser::parseRuleSingleSymbolLHS < SymbolType > ( input );
	std::vector < SymbolType > rhs = GrammarFromXMLParser::parseRuleRHS < SymbolType > ( input );

	grammar.addRawRule ( std::move ( lhs ), std::move ( rhs ) );
}

template < class SymbolType >
void LeftLG < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( LeftLG::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );

	GrammarToXMLComposer::composeNonterminalAlphabet ( out, this->getNonterminalAlphabet ( ) );
	GrammarToXMLComposer::composeTerminalAlphabet ( out, this->getTerminalAlphabet ( ) );
	GrammarToXMLComposer::composeInitialSymbol ( out, this->getInitialSymbol ( ) );
	composeRules ( out );

	out.emplace_back ( LeftLG::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
void LeftLG < SymbolType >::composeRules ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( "rules", sax::Token::TokenType::START_ELEMENT );

	for ( const auto & rule : this->getRules ( ) )

		for ( const auto & rhs : rule.second ) {
			out.emplace_back ( "rule", sax::Token::TokenType::START_ELEMENT );

			GrammarToXMLComposer::composeRuleSingleSymbolLHS ( out, rule.first );
			GrammarToXMLComposer::composeRuleLeftLGRHS ( out, rhs );

			out.emplace_back ( "rule", sax::Token::TokenType::END_ELEMENT );
		}

	out.emplace_back ( "rules", sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* LeftLG < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace grammar */

namespace std {

template < class SymbolType >
class ComponentConstraint< grammar::LeftLG < SymbolType >, SymbolType, grammar::TerminalAlphabet > {
public:
	static bool used ( const grammar::LeftLG < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, std::set < std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > > > & rule : grammar.getRules ( ) ) {
			for ( const std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > & rhsTmp : rule.second )
				if ( rhsTmp.template is < std::vector < SymbolType > > ( ) ) {
					const std::vector < SymbolType > & rhs = rhsTmp.template get < std::vector < SymbolType > > ( );

					if ( std::find ( rhs.begin ( ), rhs.end ( ), symbol ) != rhs.end ( ) )
						return true;
				} else {
					const std::pair < SymbolType, std::vector < SymbolType > > & rhs = rhsTmp.template get < std::pair < SymbolType, std::vector < SymbolType > > > ( );

					if ( std::find ( rhs.second.begin ( ), rhs.second.end ( ), symbol ) != rhs.second.end ( ) )
						return true;
				}

		}

		return false;
	}

	static bool available ( const grammar::LeftLG < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::LeftLG < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + std::to_string ( symbol ) + "cannot be in terminal alphabet since it is already nonterminal alphabet" );
	}
};

template < class SymbolType >
class ComponentConstraint< grammar::LeftLG < SymbolType >, SymbolType, grammar::NonterminalAlphabet > {
public:
	static bool used ( const grammar::LeftLG < SymbolType > & grammar, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, std::set < std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > > > & rule : grammar.getRules ( ) ) {
			if ( rule.first == symbol )
				return true;

			for ( const std::variant < std::vector < SymbolType >, std::pair < SymbolType, std::vector < SymbolType > > > & rhsTmp : rule.second )
				if ( rhsTmp.template is < std::pair < SymbolType, std::vector < SymbolType > > > ( ) ) {
					const std::pair < SymbolType, std::vector < SymbolType > > & rhs = rhsTmp.template get < std::pair < SymbolType, std::vector < SymbolType > > > ( );

					if ( rhs.first == symbol )
						return true;
				}

		}

		if ( grammar.template accessElement < grammar::InitialSymbol > ( ).get ( ) == symbol )
			return true;

		return false;
	}

	static bool available ( const grammar::LeftLG < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const grammar::LeftLG < SymbolType > & grammar, const SymbolType & symbol ) {
		if ( grammar.template accessComponent < grammar::TerminalAlphabet > ( ).get ( ).count ( symbol ) )
			throw grammar::GrammarException ( "Symbol " + std::to_string ( symbol ) + "cannot be in nonterminal alphabet since it is already in terminal alphabet" );
	}
};

template < class SymbolType >
class ElementConstraint< grammar::LeftLG < SymbolType >, SymbolType, grammar::InitialSymbol > {
public:
	static bool available ( const grammar::LeftLG < SymbolType > & grammar, const SymbolType & symbol ) {
		return grammar.template accessComponent < grammar::NonterminalAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const grammar::LeftLG < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* LEFT_LG_H_ */
