/*
 * CompressedBitParallelTreeIndex.h
 *
 *  Created on: Jan 8, 2017
 *      Author: Jan Travnicek
 */

#ifndef ARBOLOGY_COMPRESSED_BIT_PARALLEL_INDEX_H_
#define ARBOLOGY_COMPRESSED_BIT_PARALLEL_INDEX_H_

#include <string>
#include <iostream>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsMap.h>
#include <container/ObjectsVector.h>
#include <common/SparseBoolVector.hpp>
#include <primitive/Bool.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace arbology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class CompressedBitParallelTreeIndex final : public alib::ObjectBase, public std::Components < CompressedBitParallelTreeIndex < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
protected:
	std::map < SymbolType, common::SparseBoolVector > m_vectors;
	std::vector < int > m_jumpTable;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit CompressedBitParallelTreeIndex ( std::set < SymbolType > alphabet, std::map < SymbolType, common::SparseBoolVector > vectors, std::vector < int > jumpTable );

	/**
	 * @return Root node of the trie
	 */
	const std::map < SymbolType, common::SparseBoolVector > & getData ( ) const;

	/**
	 * @return subtree jump table
	 */
	const std::vector < int > & getJumps ( ) const;

	std::vector < SymbolType > getString ( ) const;

	const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Sets the compressedBit vector for given symbol
	 * @param tree root node to set
	 */
	void setCompressedBitVectorForSymbol ( SymbolType symbol, common::SparseBoolVector data );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const CompressedBitParallelTreeIndex & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "CompressedBitParallelTreeIndex";

		return xmlTagName;
	}

	static CompressedBitParallelTreeIndex parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef CompressedBitParallelTreeIndex < > normalized_type;

	virtual ObjectBase * normalize ( ) && {
		if ( typeid ( CompressedBitParallelTreeIndex < > ) == typeid ( CompressedBitParallelTreeIndex < SymbolType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		std::map < DefaultSymbolType, common::SparseBoolVector > vectors;
		for ( std::pair < SymbolType, common::SparseBoolVector > && vector : std::make_moveable_map ( m_vectors ) )
			vectors.insert ( std::make_pair ( alphabet::SymbolNormalize::normalizeSymbol ( std::move ( vector.first ) ), std::move ( vector.second ) ) );

		return new CompressedBitParallelTreeIndex < > ( std::move ( alphabet ), std::move ( vectors ), std::move ( m_jumpTable ) );
	}
};

} /* namespace arbology */

} /* namespace indexes */

namespace indexes {

namespace arbology {

template < class SymbolType >
CompressedBitParallelTreeIndex < SymbolType >::CompressedBitParallelTreeIndex ( std::set < SymbolType > alphabet, std::map < SymbolType, common::SparseBoolVector > vectors, std::vector < int > jumpTable ) : std::Components < CompressedBitParallelTreeIndex, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ), m_vectors ( std::move ( vectors ) ), m_jumpTable ( jumpTable ) {
}

template < class SymbolType >
alib::ObjectBase * CompressedBitParallelTreeIndex < SymbolType >::clone ( ) const {
	return new CompressedBitParallelTreeIndex ( * this );
}

template < class SymbolType >
alib::ObjectBase * CompressedBitParallelTreeIndex < SymbolType >::plunder ( ) && {
	return new CompressedBitParallelTreeIndex ( std::move ( * this ) );
}

template < class SymbolType >
const std::map < SymbolType, common::SparseBoolVector > & CompressedBitParallelTreeIndex < SymbolType >::getData ( ) const {
	return m_vectors;
}

template < class SymbolType >
const std::vector < int > & CompressedBitParallelTreeIndex < SymbolType >::getJumps ( ) const {
	return m_jumpTable;
}

template < class SymbolType >
std::vector < SymbolType > CompressedBitParallelTreeIndex < SymbolType >::getString ( ) const {
	std::vector < SymbolType > res;

	unsigned index = 0;

	do {
		for ( const std::pair < const SymbolType, common::SparseBoolVector > & compressedBitVector : m_vectors )
			if ( compressedBitVector.second.size ( ) > index && compressedBitVector.second [ index ] ) {
				res.push_back ( compressedBitVector.first );
				continue;
			}

	} while ( res.size ( ) == index ++ + 1 );

	return res;
}

template < class SymbolType >
void CompressedBitParallelTreeIndex < SymbolType >::setCompressedBitVectorForSymbol ( SymbolType symbol, common::SparseBoolVector data ) {
	this->m_vectors [ symbol ] = std::move ( data );
}

template < class SymbolType >
void CompressedBitParallelTreeIndex < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(CompressedBitParallelTreeIndex " << this->m_vectors << ", " << m_jumpTable << ")";
}

template < class SymbolType >
int CompressedBitParallelTreeIndex < SymbolType >::compare ( const CompressedBitParallelTreeIndex & other ) const {
	auto first = std::tie ( getData ( ), getAlphabet ( ), getJumps ( ) );
	auto second = std::tie ( other.getData ( ), other.getAlphabet ( ), other.getJumps ( ) );

	static std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
CompressedBitParallelTreeIndex < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
CompressedBitParallelTreeIndex < SymbolType > CompressedBitParallelTreeIndex < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, CompressedBitParallelTreeIndex::getXmlTagName ( ) );
	std::set < SymbolType > alphabet = alib::xmlApi < std::set < SymbolType > >::parse ( input );
	std::map < SymbolType, common::SparseBoolVector > data = alib::xmlApi < std::map < SymbolType, common::SparseBoolVector > >::parse ( input );
	std::vector < int > jumps = alib::xmlApi < std::vector < int > >::parse ( input );

	CompressedBitParallelTreeIndex < SymbolType > res ( std::move ( alphabet ), std::move ( data ), std::move ( jumps ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, CompressedBitParallelTreeIndex::getXmlTagName ( ) );
	return res;
}

template < class SymbolType >
void CompressedBitParallelTreeIndex < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( CompressedBitParallelTreeIndex::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < std::set < SymbolType > >::compose ( out, getAlphabet ( ) );
	alib::xmlApi < std::map < SymbolType, common::SparseBoolVector > >::compose ( out, getData ( ) );
	alib::xmlApi < std::vector < int > >::compose ( out, getJumps ( ) );
	out.emplace_back ( CompressedBitParallelTreeIndex::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase * CompressedBitParallelTreeIndex < SymbolType >::inc ( ) && {
	return new alib::UniqueObject ( alib::Object ( std::move ( * this ) ), primitive::Integer ( 0 ) );
}

} /* namespace arbology */

} /* namespace indexes */

namespace std {

template < class SymbolType >
class ComponentConstraint < indexes::arbology::CompressedBitParallelTreeIndex < SymbolType >, SymbolType, indexes::arbology::GeneralAlphabet > {
public:
	static bool used ( const indexes::arbology::CompressedBitParallelTreeIndex < SymbolType > & index, const SymbolType & symbol ) {
		const std::map < SymbolType, common::SparseBoolVector > & content = index.getData ( );

		return content.find ( symbol ) != content.end ( );
	}

	static bool available ( const indexes::arbology::CompressedBitParallelTreeIndex < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::arbology::CompressedBitParallelTreeIndex < SymbolType > &, const SymbolType & ) {
	}

};

} /* namespace std */

#endif /* ARBOLOGY_COMPRESSED_BIT_PARALLEL_INDEX_H_ */
