/*
 * NonlinearFullAndLinearIndex.h
 *
 *  Created on: Apr 13, 2017
 *      Author: Jan Travnicek
 */

#ifndef ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_
#define ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_

#include <string>
#include <iostream>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsVector.h>

#include <indexes/stringology/PositionHeap.h>

#include <alphabet/common/SymbolNormalize.h>

namespace indexes {

namespace arbology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class NonlinearFullAndLinearIndex final : public alib::ObjectBase {
protected:
	indexes::stringology::PositionHeap < SymbolType > m_StringIndex;
	std::vector < int > m_JumpTable;
	std::vector < unsigned > m_Repeats;

public:
	/**
	 * @copydoc SuffixTrieNode::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc SuffixTrieNode::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit NonlinearFullAndLinearIndex ( indexes::stringology::PositionHeap < SymbolType > stringIndex, std::vector < int > jumpTable, std::vector < unsigned > repeats );

	/**
	 * @return Root node of the trie
	 */
	const indexes::stringology::PositionHeap < SymbolType > & getStringIndex ( ) const;

	/**
	 * @return subtree jump table
	 */
	const std::vector < int > & getJumps ( ) const;

	/**
	 * @return repeats
	 */
	const std::vector < unsigned > & getRepeats ( ) const;

	const std::vector < SymbolType > & getString ( ) const;

	const std::set < SymbolType > & getAlphabet ( ) const {
		return m_StringIndex.getAlphabet ( );
	}

	/**
	 * Sets the compressedBit vector for given symbol
	 * @param tree root node to set
	 */
	void setStringIndex ( indexes::stringology::PositionHeap < SymbolType > stringIndex );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromAlphabet ( const SymbolType & symbol ) {
		return m_StringIndex.removeSymbolFromAlphabet ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const NonlinearFullAndLinearIndex & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName ( ) {
		static std::string xmlTagName = "NonlinearFullAndLinearIndex";

		return xmlTagName;
	}

	static NonlinearFullAndLinearIndex parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef NonlinearFullAndLinearIndex < > normalized_type;

	virtual ObjectBase * normalize ( ) && {
		if ( typeid ( NonlinearFullAndLinearIndex < > ) == typeid ( NonlinearFullAndLinearIndex < SymbolType > ) )
			return this;

		indexes::stringology::PositionHeap < DefaultSymbolType > stringIndex = std::move ( m_StringIndex ).normalizeRaw ( );

		return new NonlinearFullAndLinearIndex < > ( std::move ( stringIndex ), std::move ( m_JumpTable ), std::move ( m_Repeats ) );
	}
};

} /* namespace arbology */

} /* namespace indexes */

namespace indexes {

namespace arbology {

template < class SymbolType >
NonlinearFullAndLinearIndex < SymbolType >::NonlinearFullAndLinearIndex ( indexes::stringology::PositionHeap < SymbolType > stringIndex, std::vector < int > jumpTable, std::vector < unsigned > repeats ) : m_StringIndex ( std::move ( stringIndex ) ), m_JumpTable ( jumpTable ), m_Repeats ( repeats ) {
}

template < class SymbolType >
alib::ObjectBase * NonlinearFullAndLinearIndex < SymbolType >::clone ( ) const {
	return new NonlinearFullAndLinearIndex ( * this );
}

template < class SymbolType >
alib::ObjectBase * NonlinearFullAndLinearIndex < SymbolType >::plunder ( ) && {
	return new NonlinearFullAndLinearIndex ( std::move ( * this ) );
}

template < class SymbolType >
const indexes::stringology::PositionHeap < SymbolType > & NonlinearFullAndLinearIndex < SymbolType >::getStringIndex ( ) const {
	return m_StringIndex;
}

template < class SymbolType >
const std::vector < int > & NonlinearFullAndLinearIndex < SymbolType >::getJumps ( ) const {
	return m_JumpTable;
}

template < class SymbolType >
const std::vector < unsigned > & NonlinearFullAndLinearIndex < SymbolType >::getRepeats ( ) const {
	return m_Repeats;
}

template < class SymbolType >
const std::vector < SymbolType > & NonlinearFullAndLinearIndex < SymbolType >::getString ( ) const {
	return m_StringIndex.getString ( );
}

template < class SymbolType >
void NonlinearFullAndLinearIndex < SymbolType >::setStringIndex ( indexes::stringology::PositionHeap < SymbolType > stringIndex ) {
	this->m_StringIndex = std::move ( stringIndex );
}

template < class SymbolType >
void NonlinearFullAndLinearIndex < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(NonlinearFullAndLinearIndex " << this->m_StringIndex << ", " << m_JumpTable << ")";
}

template < class SymbolType >
int NonlinearFullAndLinearIndex < SymbolType >::compare ( const NonlinearFullAndLinearIndex & other ) const {
	auto first = std::tie ( getStringIndex ( ), getJumps ( ), getRepeats ( ) );
	auto second = std::tie ( other.getStringIndex ( ), other.getJumps ( ), other.getRepeats ( ) );

	static std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
NonlinearFullAndLinearIndex < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
NonlinearFullAndLinearIndex < SymbolType > NonlinearFullAndLinearIndex < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, NonlinearFullAndLinearIndex::getXmlTagName ( ) );
	indexes::stringology::PositionHeap < SymbolType > stringIndex = alib::xmlApi < indexes::stringology::PositionHeap < SymbolType > >::parse ( input );
	std::vector < int > jumps = alib::xmlApi < std::vector < int > >::parse ( input );
	std::vector < unsigned > repeats = alib::xmlApi < std::vector < unsigned > >::parse ( input );

	NonlinearFullAndLinearIndex < SymbolType > res ( std::move ( stringIndex ), std::move ( jumps ), std::move ( repeats ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, NonlinearFullAndLinearIndex::getXmlTagName ( ) );
	return res;
}

template < class SymbolType >
void NonlinearFullAndLinearIndex < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( NonlinearFullAndLinearIndex::getXmlTagName ( ), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < indexes::stringology::PositionHeap < SymbolType > >::compose ( out, getStringIndex ( ) );
	alib::xmlApi < std::vector < int > >::compose ( out, getJumps ( ) );
	alib::xmlApi < std::vector < unsigned > >::compose ( out, getRepeats ( ) );
	out.emplace_back ( NonlinearFullAndLinearIndex::getXmlTagName ( ), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase * NonlinearFullAndLinearIndex < SymbolType >::inc ( ) && {
	return new alib::UniqueObject ( alib::Object ( std::move ( * this ) ), primitive::Integer ( 0 ) );
}

} /* namespace arbology */

} /* namespace indexes */

#endif /* ARBOLOGY_NONLINEAR_FULL_AND_LINEAR_INDEX_H_ */
