/*
 * PositionHeap.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef POSITION_HEAP_H_
#define POSITION_HEAP_H_

#include <string>
#include <set>
#include <trie>
#include <iostream>
#include <algorithm>
#include <sstream>

#include <common/DefaultSymbolType.h>

#include <core/components.hpp>
#include <exception/CommonException.h>

#include <object/Object.h>
#include <object/UniqueObject.h>
#include <object/ObjectBase.h>

#include <sax/FromXMLParserHelper.h>
#include <core/xmlApi.hpp>

#include <container/ObjectsSet.h>
#include <container/ObjectsTrie.h>
#include <container/ObjectsVector.h>

#include <primitive/Unsigned.h>

#include <alphabet/common/SymbolNormalize.h>
#include <indexes/common/IndexesNormalize.h>

namespace indexes {

namespace stringology {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of RegExpElement.
 */
template < class SymbolType = DefaultSymbolType >
class PositionHeap final : public alib::ObjectBase, public std::Components < PositionHeap < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
protected:
	std::trie < SymbolType, unsigned > m_trie;
	std::vector < SymbolType > m_string;

public:
	/**
	 * @copydoc PositionHeap::clone() const
	 */
	virtual ObjectBase * clone ( ) const;

	/**
	 * @copydoc PositionHeap::plunder() const
	 */
	virtual ObjectBase * plunder ( ) &&;

	explicit PositionHeap ( std::set < SymbolType > edgeAlphabet, std::trie < SymbolType, unsigned > trie, std::vector < SymbolType > string );
	explicit PositionHeap ( std::trie < SymbolType, unsigned > trie, std::vector < SymbolType > string );

	void checkTrie ( const std::trie < SymbolType, unsigned > & trie );

	/**
	 * @return Root node of the trie
	 */
	const std::trie < SymbolType, unsigned > & getRoot ( ) const;

	const std::vector < SymbolType > & getString ( ) const;

	const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	/**
	 * Sets the root node of the regular expression tree
	 * @param tree root node to set
	 */
	void setTree ( std::trie < SymbolType, unsigned > tree );

	/**
	 * Removes symbol from the alphabet of symbol available in the regular expression
	 * @param symbol removed symbol from the alphabet
	 */
	bool removeSymbolFromEdgeAlphabet ( const SymbolType & symbol ) {
		return this->template accessComponent < GeneralAlphabet > ( ).remove ( symbol );
	}

	/**
	 * Prints XML representation of the tree to the output stream.
	 * @param out output stream to which print the tree
	 * @param tree tree to print
	 */
	virtual void operator >>( std::ostream & out ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PositionHeap & other ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "PositionHeap";

		return xmlTagName;
	}

	static PositionHeap parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef PositionHeap < > normalized_type;

	PositionHeap < > normalizeRaw ( ) && {
		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		std::trie < DefaultSymbolType, unsigned > trie = IndexesNormalize::normalizeTrie ( std::move ( m_trie ) );
		std::vector < DefaultSymbolType > string = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( m_string ) );

		return PositionHeap < > ( std::move ( alphabet ), std::move ( trie ), std::move ( string ) );
	}

	virtual alib::ObjectBase * normalize ( ) && {
		if ( typeid ( PositionHeap < > ) == typeid ( PositionHeap < SymbolType > ) )
			return this;

		return std::move ( * this ).normalizeRaw ( ).plunder ( );
	}
};

} /* namespace stringology */

} /* namespace indexes */

namespace indexes {

namespace stringology {

template < class SymbolType >
PositionHeap < SymbolType >::PositionHeap ( std::set < SymbolType > edgeAlphabet, std::trie < SymbolType, unsigned > trie, std::vector < SymbolType > string ) : std::Components < PositionHeap, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( edgeAlphabet ) ), std::tuple < > ( ) ), m_trie ( std::move ( trie ) ), m_string ( std::move ( string ) ) {
	checkTrie ( this->m_trie );
	// TODO check validity of the string like in LinearString
}

template < class SymbolType >
PositionHeap < SymbolType >::PositionHeap ( std::trie < SymbolType, unsigned > trie, std::vector < SymbolType > string ) : PositionHeap ( computeMinimalEdgeAlphabet ( trie ), trie, std::move ( string ) ) {
}

template < class SymbolType >
alib::ObjectBase * PositionHeap < SymbolType >::clone ( ) const {
	return new PositionHeap ( * this );
}

template < class SymbolType >
alib::ObjectBase * PositionHeap < SymbolType >::plunder ( ) && {
	return new PositionHeap ( std::move ( * this ) );
}

template < class SymbolType >
void PositionHeap < SymbolType >::checkTrie ( const std::trie < SymbolType, unsigned > & trie ) {
	for ( const std::pair < const SymbolType, std::trie < SymbolType, unsigned > > & child : trie.getChildren ( ) ) {
		if ( ! getAlphabet ( ).count ( child.first ) )
			throw exception::CommonException ( "Symbol " + std::to_string ( child.first ) + "not in the alphabet." );
		checkTrie ( child.second );
	}
}

template < class SymbolType >
const std::trie < SymbolType, unsigned > & PositionHeap < SymbolType >::getRoot ( ) const {
	return m_trie;
}

template < class SymbolType >
const std::vector < SymbolType > & PositionHeap < SymbolType >::getString ( ) const {
	return m_string;
}

template < class SymbolType >
void PositionHeap < SymbolType >::setTree ( std::trie < SymbolType, unsigned > trie ) {
	checkTrie ( trie );
	this->m_trie = std::move ( trie ).plunder ( );
}

template < class SymbolType >
void PositionHeap < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(PositionHeap " << this->m_trie << ")";
}

template < class SymbolType >
int PositionHeap < SymbolType >::compare ( const PositionHeap & other ) const {
	auto first = std::tie ( getRoot ( ), getAlphabet ( ) );
	auto second = std::tie ( other.getRoot ( ), other.getAlphabet ( ) );

	static std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType >
PositionHeap < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

template < class SymbolType >
PositionHeap < SymbolType > PositionHeap < SymbolType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PositionHeap::getXmlTagName() );
	std::set < SymbolType > edgeAlphabet = alib::xmlApi < std::set < SymbolType > >::parse ( input );
	std::trie < SymbolType, unsigned > root = alib::xmlApi < std::trie < SymbolType, unsigned > >::parse ( input );
	std::vector < SymbolType > string = alib::xmlApi < std::vector < SymbolType > >::parse ( input );
	PositionHeap < SymbolType > trie ( std::move ( edgeAlphabet ), std::move ( root ), std::move ( string ) );

	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PositionHeap::getXmlTagName() );
	return trie;
}

template < class SymbolType >
void PositionHeap < SymbolType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PositionHeap::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	alib::xmlApi < std::set < SymbolType > >::compose ( out, getAlphabet ( ) );
	alib::xmlApi < std::trie < SymbolType, unsigned > >::compose ( out, getRoot ( ) );
	alib::xmlApi < std::vector < SymbolType > >::compose ( out, getString ( ) );
	out.emplace_back ( PositionHeap::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType >
alib::ObjectBase* PositionHeap < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace stringology */

} /* namespace indexes */

namespace std {

template < class SymbolType >
class ComponentConstraint < indexes::stringology::PositionHeap < SymbolType >, SymbolType, indexes::stringology::GeneralAlphabet > {

	static bool used ( const std::trie < SymbolType, unsigned > & trie, const SymbolType & symbol ) {
		for ( const std::pair < const SymbolType, std::trie < SymbolType, unsigned > > & child : trie.getChildren ( ) ) {
			if ( symbol == child.first || checkTrie ( trie, child.second ) )
				return true;
		}
		return false;
	}

public:
	static bool used ( const indexes::stringology::PositionHeap < SymbolType > & index, const SymbolType & symbol ) {
		const std::vector < SymbolType > & content = index.getString ( );
		return std::find ( content.begin(), content.end(), symbol ) != content.end() || used ( index.getRoot ( ), symbol );
	}

	static bool available ( const indexes::stringology::PositionHeap < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const indexes::stringology::PositionHeap < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* POSITION_HEAP_H_ */
