/*
 * FormalRegExpSymbol.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#ifndef FORMAL_REG_EXP_SYMBOL_H_
#define FORMAL_REG_EXP_SYMBOL_H_

#include <sstream>

#include "FormalRegExpElement.h"
#include <alphabet/common/SymbolNormalize.h>

namespace regexp {

/**
 * Represents symbol in the regular expression. Contains name of the symbol.
 */
template < class SymbolType >
class FormalRegExpSymbol : public FormalRegExpElement < SymbolType >, public std::NullaryNode < std::smart_ptr < FormalRegExpElement < SymbolType > >, std::smart_ptr < const FormalRegExpElement < SymbolType > >, FormalRegExpSymbol < SymbolType > > {
	SymbolType m_symbol;

public:
	void accept ( typename FormalRegExpElement < SymbolType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

	explicit FormalRegExpSymbol ( SymbolType symbol );

	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	FormalRegExpSymbol < SymbolType > * clone ( ) const override;

	/**
	 * @copydoc FormalRegExpElement::plunder() const
	 */
	FormalRegExpSymbol < SymbolType > * plunder ( ) && override;

	/**
	 * @copydoc FormalRegExpElement::clone() const
	 */
	std::smart_ptr < UnboundedRegExpElement < SymbolType > > asUnbounded ( ) const override;

	/**
	 * @copydoc FormalRegExpElement::testSymbol() const
	 */
	bool testSymbol ( const SymbolType & symbol ) const override;

	/**
	 * @copydoc FormalRegExpElement::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const override;

	/**
	 * @copydoc FormalRegExpElement::checkAlphabet()
	 */
	bool checkAlphabet ( const std::set < SymbolType > & alphabet ) const override;

	int compare ( const FormalRegExpElement < SymbolType > & other ) const override {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	int compare ( const FormalRegExpSymbol & ) const;

	/**
	 * @copydoc FormalRegExpElement::operator>>() const
	 */
	void operator >>( std::ostream & out ) const override;

	/**
	 * Returns the string representation of RegExp Symbol.
	 */
	const SymbolType & getSymbol ( ) const;

	explicit operator std::string ( ) const override;

	virtual std::smart_ptr < FormalRegExpElement < DefaultSymbolType > > normalize ( ) && override {
		return std::smart_ptr < FormalRegExpElement < DefaultSymbolType > > ( new FormalRegExpSymbol < DefaultSymbolType > ( alphabet::SymbolNormalize::normalizeSymbol ( std::move ( m_symbol ) ) ) );
	}
};

} /* namespace regexp */

#include "../unbounded/UnboundedRegExpSymbol.h"

namespace regexp {

template < class SymbolType >
FormalRegExpSymbol < SymbolType >::FormalRegExpSymbol ( SymbolType symbol ) : m_symbol ( std::move ( symbol ) ) {
}

template < class SymbolType >
FormalRegExpSymbol < SymbolType > * FormalRegExpSymbol < SymbolType >::clone ( ) const {
	return new FormalRegExpSymbol ( * this );
}

template < class SymbolType >
FormalRegExpSymbol < SymbolType > * FormalRegExpSymbol < SymbolType >::plunder ( ) && {
	return new FormalRegExpSymbol ( std::move ( * this ) );
}

template < class SymbolType >
std::smart_ptr < UnboundedRegExpElement < SymbolType > > FormalRegExpSymbol < SymbolType >::asUnbounded ( ) const {
	return std::smart_ptr < UnboundedRegExpElement < SymbolType > > ( new UnboundedRegExpSymbol < SymbolType > ( this->m_symbol ) );
}

template < class SymbolType >
int FormalRegExpSymbol < SymbolType >::compare ( const FormalRegExpSymbol & other ) const {
	static std::compare < decltype ( m_symbol ) > comp;

	return comp ( m_symbol, other.m_symbol );
}

template < class SymbolType >
void FormalRegExpSymbol < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(FormalRegExpSymbol " << m_symbol << ")";
}

template < class SymbolType >
bool FormalRegExpSymbol < SymbolType >::testSymbol ( const SymbolType & symbol ) const {
	return symbol == this->m_symbol;
}

template < class SymbolType >
void FormalRegExpSymbol < SymbolType >::computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const {
	alphabet.insert ( m_symbol );
}

template < class SymbolType >
bool FormalRegExpSymbol < SymbolType >::checkAlphabet ( const std::set < SymbolType > & alphabet ) const {
	return alphabet.count ( m_symbol );
}

template < class SymbolType >
const SymbolType & FormalRegExpSymbol < SymbolType >::getSymbol ( ) const {
	return m_symbol;
}

template < class SymbolType >
FormalRegExpSymbol < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */

#endif /* FORMAL_REG_EXP_SYMBOL_H_ */
