/*
 * UnboundedRegExpAlternation.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Martin Zak
 */

#ifndef UNBOUNDED_REG_EXP_ALTERNATION_H_
#define UNBOUNDED_REG_EXP_ALTERNATION_H_

#include <vector>
#include <sstream>

#include <exception/CommonException.h>

#include "UnboundedRegExpElement.h"

namespace regexp {

/**
 * Represents alternation operator in the regular expression. Contains list of UnboundedRegExpElement
 * as operands of the operator.
 */
template < class SymbolType >
class UnboundedRegExpAlternation : public UnboundedRegExpElement < SymbolType >, public std::VararyNode < std::smart_ptr < UnboundedRegExpElement < SymbolType > >, std::smart_ptr < const UnboundedRegExpElement < SymbolType > >, UnboundedRegExpAlternation < SymbolType > > {
public:
	void accept ( typename UnboundedRegExpElement < SymbolType >::Visitor & visitor ) const override {
		visitor.visit ( * this );
	}

	explicit UnboundedRegExpAlternation ( );

	/**
	 * @copydoc UnboundedRegExpElement::clone() const
	 */
	UnboundedRegExpAlternation < SymbolType > * clone ( ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::plunder() const
	 */
	UnboundedRegExpAlternation < SymbolType > * plunder ( ) && override;

	/**
	 * @copydoc UnboundedRegExpElement::cloneAsFormal() const
	 */
	std::smart_ptr < FormalRegExpElement < SymbolType > > asFormal ( ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::testSymbol() const
	 */
	bool testSymbol ( const SymbolType & symbol ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::computeMinimalAlphabet()
	 */
	void computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const override;

	/**
	 * @copydoc UnboundedRegExpElement::checkAlphabet()
	 */
	bool checkAlphabet ( const std::set < SymbolType > & alphabet ) const override;

	/**
	 * @return elements
	 */
	const std::vector < std::smart_ptr < const UnboundedRegExpElement < SymbolType > > > & getElements ( ) const;

	/**
	 * @return elements
	 */
	const std::vector < std::smart_ptr < UnboundedRegExpElement < SymbolType > > > & getElements ( );

	/**
	 * @param element to append
	 */
	void appendElement ( UnboundedRegExpElement < SymbolType > && element );
	void appendElement ( const UnboundedRegExpElement < SymbolType > & element );

	int compare ( const UnboundedRegExpElement < SymbolType > & other ) const override {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	int compare ( const UnboundedRegExpAlternation < SymbolType > & ) const;

	/**
	 * @copydoc UnboundedRegExpElement::operator>>() const
	 */
	void operator >>( std::ostream & out ) const override;

	explicit operator std::string ( ) const override;

	virtual std::smart_ptr < UnboundedRegExpElement < DefaultSymbolType > > normalize ( ) && override {
		UnboundedRegExpAlternation < DefaultSymbolType > * res = new UnboundedRegExpAlternation < DefaultSymbolType > ( );

		for ( std::smart_ptr < UnboundedRegExpElement < SymbolType > > & element : this->getChildren ( ) )
			res->appendElement ( std::move ( * std::move ( * element ).normalize ( ) ) );

		return std::smart_ptr < UnboundedRegExpElement < DefaultSymbolType > > ( res );
	}
};

} /* namespace regexp */

#include "../formal/FormalRegExpAlternation.h"
#include "../formal/FormalRegExpEmpty.h"

namespace regexp {

template < class SymbolType >
UnboundedRegExpAlternation < SymbolType >::UnboundedRegExpAlternation ( ) {
}

template < class SymbolType >
const std::vector < std::smart_ptr < const UnboundedRegExpElement < SymbolType > > > & UnboundedRegExpAlternation < SymbolType >::getElements ( ) const {
	return this->getChildren();
}

template < class SymbolType >
const std::vector < std::smart_ptr < UnboundedRegExpElement < SymbolType > > > & UnboundedRegExpAlternation < SymbolType >::getElements ( ) {
	return this->getChildren();
}

template < class SymbolType >
void UnboundedRegExpAlternation < SymbolType >::appendElement ( UnboundedRegExpElement < SymbolType > && element ) {
	this->pushBackChild ( std::smart_ptr < UnboundedRegExpElement < SymbolType > > ( std::move ( element ).plunder ( ) ) );
}

template < class SymbolType >
void UnboundedRegExpAlternation < SymbolType >::appendElement ( const UnboundedRegExpElement < SymbolType > & element ) {
	this->appendElement ( std::move_copy ( element ) );
}

template < class SymbolType >
UnboundedRegExpAlternation < SymbolType > * UnboundedRegExpAlternation < SymbolType >::clone ( ) const {
	return new UnboundedRegExpAlternation ( * this );
}

template < class SymbolType >
UnboundedRegExpAlternation < SymbolType > * UnboundedRegExpAlternation < SymbolType >::plunder ( ) && {
	return new UnboundedRegExpAlternation ( std::move ( * this ) );
}

template < class SymbolType >
std::smart_ptr < FormalRegExpElement < SymbolType > > UnboundedRegExpAlternation < SymbolType >::asFormal ( ) const {
	if ( getElements ( ).size ( ) == 0 ) return std::smart_ptr < FormalRegExpElement < SymbolType > > ( new FormalRegExpEmpty < SymbolType > ( ) );

	std::smart_ptr < FormalRegExpElement < SymbolType > > res = getElements ( )[getElements ( ).size ( ) - 1]->asFormal ( );

	for ( unsigned i = getElements ( ).size ( ) - 1; i >= 1; i-- )
		res = std::smart_ptr < FormalRegExpElement < SymbolType > > ( new FormalRegExpAlternation < SymbolType > ( std::move ( * getElements ( )[i - 1]->asFormal ( ) ), std::move ( * res ) ) );

	return res;
}

template < class SymbolType >
int UnboundedRegExpAlternation < SymbolType >::compare ( const UnboundedRegExpAlternation < SymbolType > & other ) const {
	static std::compare < typename std::decay < decltype ( getElements ( ) ) >::type > comp;

	return comp ( getElements ( ), other.getElements ( ) );
}

template < class SymbolType >
void UnboundedRegExpAlternation < SymbolType >::operator >>( std::ostream & out ) const {
	out << "(UnboundedRegExpAlternation";

	for ( const auto & child : getElements() )
		out << " " << * child;

	out << ")";
}

template < class SymbolType >
bool UnboundedRegExpAlternation < SymbolType >::testSymbol ( const SymbolType & symbol ) const {
	for ( const auto & child : getElements() )
		if ( child->testSymbol ( symbol ) ) return true;

	return false;
}

template < class SymbolType >
bool UnboundedRegExpAlternation < SymbolType >::checkAlphabet ( const std::set < SymbolType > & alphabet ) const {
	for ( const auto & child : getElements() )
		if ( !child->checkAlphabet ( alphabet ) ) return false;

	return true;
}

template < class SymbolType >
void UnboundedRegExpAlternation < SymbolType >::computeMinimalAlphabet ( std::set < SymbolType > & alphabet ) const {
	for ( const auto & child : getElements() )
		child->computeMinimalAlphabet ( alphabet );
}

template < class SymbolType >
UnboundedRegExpAlternation < SymbolType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

} /* namespace regexp */

#endif /* UNBOUNDED_REG_EXP_ALTERNATION_H_ */
