/*
 * CyclicString.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef CYCLIC_STRING_H_
#define CYCLIC_STRING_H_

#include <set>
#include <vector>
#include <sstream>
#include <algorithm>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>

#include "StringBase.h"
#include "StringFeatures.h"
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include <exception/CommonException.h>
#include <alphabet/common/SymbolNormalize.h>

namespace string {

class GeneralAlphabet;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of CyclicStringElement.
 */
template < class SymbolType >
class CyclicString final : public StringBase, public std::Components < CyclicString < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
	std::vector < SymbolType > m_Data;

public:
	explicit CyclicString ( );
	explicit CyclicString ( std::set < SymbolType > alphabet, std::vector < SymbolType > str );
	explicit CyclicString ( std::vector < SymbolType > str );
	explicit CyclicString ( const std::string & str );
	explicit CyclicString ( const char * str );
	explicit CyclicString ( const Epsilon < SymbolType > & epsilon );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	virtual const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < SymbolType > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add( symbols );
	}

	/**
	 * @return List of symbols forming string (const version).
	 */
	const std::vector < SymbolType > & getContent ( ) const;

	void setContent ( std::vector < SymbolType > str );

	/**
	 * @return true if string is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const CyclicString & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "CyclicString";

		return xmlTagName;
	}

	static CyclicString parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef CyclicString < > normalized_type;

	virtual StringBase * normalize ( ) && {
		if ( typeid ( CyclicString < > ) == typeid ( CyclicString < SymbolType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		std::vector < DefaultSymbolType > content = alphabet::SymbolNormalize::normalizeSymbols ( std::move ( m_Data ) );
		return new CyclicString < > ( std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace string */

#include "Epsilon.h"

namespace string {

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(std::set<SymbolType> alphabet, std::vector<SymbolType> str) : std::Components < CyclicString < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ) {
	setContent(std::move(str));
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString() : CyclicString ( std::set < SymbolType > ( ), std::vector < SymbolType > ( ) ) {
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(std::vector<SymbolType> str) : CyclicString ( std::set < SymbolType> ( str.begin(), str.end() ), str ) {
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(const std::string& str) : CyclicString ( std::vector < SymbolType > ( str.begin ( ), str.end ( ) ) ) {
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(const char* str) : CyclicString((std::string) str) {
}

template < class SymbolType >
CyclicString < SymbolType >::CyclicString(const Epsilon < SymbolType > & epsilon) : CyclicString ( epsilon.getAlphabet( ), std::vector < SymbolType > ( ) ) {
}

template < class SymbolType >
StringBase* CyclicString < SymbolType >::clone() const {
	return new CyclicString(*this);
}

template < class SymbolType >
StringBase* CyclicString < SymbolType >::plunder() && {
	return new CyclicString(std::move(*this));
}

template < class SymbolType >
const std::vector<SymbolType>& CyclicString < SymbolType >::getContent() const {
	return this->m_Data;
}

//serves as both move and copy content setter
template < class SymbolType >
void CyclicString < SymbolType >::setContent(std::vector<SymbolType> str) {
	std::set<SymbolType> minimalAlphabet(str.begin(), str.end());
	std::set<SymbolType> unknownSymbols;
	std::set_difference(minimalAlphabet.begin(), minimalAlphabet.end(), getAlphabet().begin(), getAlphabet().end(), std::inserter(unknownSymbols, unknownSymbols.end()));

	if(unknownSymbols.size() > 0)
		throw exception::CommonException("Input symbols not in the alphabet.");

	m_Data = std::move(str);
}

template < class SymbolType >
bool CyclicString < SymbolType >::isEmpty() const {
	return this->m_Data.size() == 0;
}

template < class SymbolType >
int CyclicString < SymbolType >::compare(const CyclicString& other) const {
	auto first = std::tie ( m_Data, getAlphabet ( ) );
	auto second = std::tie ( other.m_Data, other.getAlphabet ( ) );

	static std::compare<decltype(first)> comp;
	return comp(first, second);
}

template < class SymbolType >
void CyclicString < SymbolType >::operator >>(std::ostream& out) const {
	out << "(CyclicString ";
	for(const SymbolType& symbol : this->m_Data)
		out << symbol;
	out << ")";
}

template < class SymbolType >
CyclicString < SymbolType >::operator std::string () const {
	std::stringstream ss;
	ss << "<";
	for(const SymbolType& symbol : this->m_Data)
		ss << symbol;
	ss << ">";
	return std::move(ss).str();
}

template < class SymbolType >
CyclicString < SymbolType > CyclicString < SymbolType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, CyclicString < >::getXmlTagName());
	std::set<SymbolType> alphabet = StringFromXMLParser::parseAlphabet < SymbolType > (input);
	std::vector<SymbolType> content = StringFromXMLParser::parseContent < SymbolType > (input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, CyclicString < >::getXmlTagName());
	return CyclicString < SymbolType > ( std::move ( alphabet ), std::move ( content ) );
}

template < class SymbolType >
void CyclicString < SymbolType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(CyclicString::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);
	StringToXMLComposer::compose(out, getAlphabet());
	StringToXMLComposer::compose(out, m_Data);
	out.emplace_back(CyclicString::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
alib::ObjectBase* CyclicString < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace string */

namespace std {

template < class SymbolType >
class ComponentConstraint< ::string::CyclicString < SymbolType >, SymbolType, ::string::GeneralAlphabet > {
public:
	static bool used ( const ::string::CyclicString < SymbolType > & str, const SymbolType & symbol ) {
		const std::vector<SymbolType>& content = str.getContent ( );
		return std::find(content.begin(), content.end(), symbol) != content.end();
	}

	static bool available ( const ::string::CyclicString < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const ::string::CyclicString < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* CYCLIC_STRING_H_ */
