/*
 * Epsilon.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef EPSILON_H_
#define EPSILON_H_

#include <set>
#include <vector>

#include <sax/FromXMLParserHelper.h>
#include <core/components.hpp>
#include <object/UniqueObject.h>

#include "StringBase.h"
#include "StringFeatures.h"
#include "common/StringFromXMLParser.h"
#include "common/StringToXMLComposer.h"
#include <alphabet/common/SymbolNormalize.h>

namespace string {

class GeneralAlphabet;

/**
 * Represents epsilon. Regular expression is stored
 * as a tree of EpsilonElement.
 */
template < class SymbolType >
class Epsilon final : public StringBase, public std::Components < Epsilon < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > {
	static const std::vector < SymbolType > content;

public:
	Epsilon ( );

	Epsilon ( std::set < SymbolType > alphabet );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	virtual const std::set < SymbolType > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < SymbolType > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add( symbols );
	}

	/**
	 * @return List of symbols forming string (const version).
	 */
	const std::vector < SymbolType > & getContent ( ) const;

	/**
	 * @return true
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const Epsilon & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static Epsilon EPSILON;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "Epsilon";

		return xmlTagName;
	}

	static Epsilon parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef Epsilon < > normalized_type;

	virtual StringBase * normalize ( ) && {
		if ( typeid ( Epsilon < > ) == typeid ( Epsilon < SymbolType > ) )
			return this;

		std::set < DefaultSymbolType > alphabet = alphabet::SymbolNormalize::normalizeAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		return new Epsilon < > ( std::move ( alphabet ) );
	}
};

template < class SymbolType >
Epsilon < SymbolType >::Epsilon(std::set<SymbolType> alphabet) : std::Components < Epsilon < SymbolType >, SymbolType, std::tuple < GeneralAlphabet >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ) ), std::tuple < > ( ) ) {
}

template < class SymbolType >
Epsilon < SymbolType >::Epsilon() : Epsilon ( std::set< SymbolType > ( ) ) {
}

template < class SymbolType >
StringBase* Epsilon < SymbolType >::clone() const {
	return new Epsilon(*this);
}

template < class SymbolType >
StringBase* Epsilon < SymbolType >::plunder() && {
	return new Epsilon(std::move(*this));
}

template < class SymbolType >
const std::vector<SymbolType> & Epsilon < SymbolType >::getContent() const {
	return Epsilon::content;
}

template < class SymbolType >
bool Epsilon < SymbolType >::isEmpty() const {
	return true;
}

template < class SymbolType >
int Epsilon < SymbolType >::compare(const Epsilon& other) const {
	static std::compare<std::set<SymbolType>> comp;
	return comp(getAlphabet(), other.getAlphabet());
}

template < class SymbolType >
void Epsilon < SymbolType >::operator>>(std::ostream& out) const {
	out << "(Epsilon)";
}

template < class SymbolType >
Epsilon < SymbolType >::operator std::string() const {
	return "E";
}

template < class SymbolType >
Epsilon < SymbolType > Epsilon < SymbolType >::EPSILON = Epsilon < SymbolType >();

template < class SymbolType >
const std::vector<SymbolType> Epsilon < SymbolType >::content;

template < class SymbolType >
Epsilon < SymbolType > Epsilon < SymbolType >::parse(std::deque<sax::Token>::iterator& input) {
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::START_ELEMENT, Epsilon::getXmlTagName());
	std::set<SymbolType> alphabet = StringFromXMLParser::parseAlphabet < SymbolType > (input);
	sax::FromXMLParserHelper::popToken(input, sax::Token::TokenType::END_ELEMENT, Epsilon::getXmlTagName());
	return Epsilon < SymbolType > (alphabet);
}

template < class SymbolType >
void Epsilon < SymbolType >::compose(std::deque<sax::Token>& out) const {
	out.emplace_back(Epsilon::getXmlTagName(), sax::Token::TokenType::START_ELEMENT);
	StringToXMLComposer::compose(out, getAlphabet());
	out.emplace_back(Epsilon::getXmlTagName(), sax::Token::TokenType::END_ELEMENT);
}

template < class SymbolType >
alib::ObjectBase* Epsilon < SymbolType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace string */

namespace std {

template < class SymbolType >
class ComponentConstraint< ::string::Epsilon < SymbolType >, SymbolType, ::string::GeneralAlphabet > {
public:
	static bool used ( const ::string::Epsilon < SymbolType > &, const SymbolType & ) {
		return false;
	}

	static bool available ( const ::string::Epsilon < SymbolType > &, const SymbolType & ) {
		return true;
	}

	static void valid ( const ::string::Epsilon < SymbolType > &, const SymbolType & ) {
	}
};

} /* namespace std */

#endif /* EPSILON_H_ */
