/*
 * PrefixRankedBarTree.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef PREFIX_RANKED_BAR_TREE_H_
#define PREFIX_RANKED_BAR_TREE_H_

#include <set>
#include <vector>
#include <tree>
#include <sstream>
#include <algorithm>
#include <deque>

#include <core/components.hpp>
#include <sax/FromXMLParserHelper.h>
#include <object/UniqueObject.h>
#include "alphabet/ranked_symbol.hpp"

#include "../TreeFeatures.h"
#include "../TreeException.h"
#include "../RankedTreeBase.h"
#include "../common/TreeFromXMLParser.h"
#include "../common/TreeToXMLComposer.h"
#include "../common/TreeAuxiliary.h"
#include <tree/common/TreeNormalize.h>

#include <alphabet/BarSymbol.h>

namespace tree {

class GeneralAlphabet;
class BarSymbols;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
template < class SymbolType, class RankType >
class PrefixRankedBarTree final : public RankedTreeBase, public std::Components < PrefixRankedBarTree < SymbolType, RankType >, std::ranked_symbol < SymbolType, RankType >, std::tuple < GeneralAlphabet, BarSymbols >, std::tuple < > > {
	std::vector < std::ranked_symbol < SymbolType, RankType > > m_Data;

	static std::vector < std::ranked_symbol < SymbolType, RankType > > toPrefixRankedBar ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & node, const SymbolType & barBase );

public:
	explicit PrefixRankedBarTree ( std::set < std::ranked_symbol < SymbolType, RankType > > bars, std::set < std::ranked_symbol < SymbolType, RankType > > alphabet, std::vector < std::ranked_symbol < SymbolType, RankType > > data );
	explicit PrefixRankedBarTree ( std::set < std::ranked_symbol < SymbolType, RankType > > bars, std::vector < std::ranked_symbol < SymbolType, RankType > > data );
	explicit PrefixRankedBarTree ( SymbolType barBase, const RankedTree < SymbolType, RankType > & tree );
	explicit PrefixRankedBarTree ( const RankedTree < > & tree );

	virtual RankedTreeBase * clone ( ) const;
	virtual RankedTreeBase * plunder ( ) &&;

	/**
	 * @return List of symbols forming tree (const version).
	 */
	const std::vector < std::ranked_symbol < SymbolType, RankType > > & getContent ( ) const;

	const std::set < std::ranked_symbol < SymbolType, RankType > > & getAlphabet ( ) const {
		return this->template accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < std::ranked_symbol < SymbolType, RankType > > & symbols ) {
		this->template accessComponent < GeneralAlphabet > ( ).add ( symbols );
	}

	const std::set < std::ranked_symbol < SymbolType, RankType > > & getBars ( ) const {
		return this->template accessComponent < BarSymbols > ( ).get ( );
	}

	void setContent ( std::vector < std::ranked_symbol < SymbolType, RankType > > data );

	void arityChecksum ( const std::vector < std::ranked_symbol < SymbolType, RankType > > & data );

	/**
	 * @return true if tree is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const PrefixRankedBarTree & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "PrefixRankedBarTree";

		return xmlTagName;
	}

	static PrefixRankedBarTree parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef PrefixRankedBarTree < > normalized_type;

	virtual RankedTreeBase * normalize ( ) && {
		if ( typeid ( PrefixRankedBarTree < > ) == typeid ( PrefixRankedBarTree < SymbolType, RankType > ) )
			return this;

		std::set < std::ranked_symbol < DefaultSymbolType, DefaultRankType > > bars = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( this->template accessComponent < BarSymbols > ( ).get ( ) ) );
		std::set < std::ranked_symbol < DefaultSymbolType, DefaultRankType > > alphabet = alphabet::SymbolNormalize::normalizeRankedAlphabet ( std::move ( this->template accessComponent < GeneralAlphabet > ( ).get ( ) ) );
		std::vector < std::ranked_symbol < DefaultSymbolType, DefaultRankType > > content = alphabet::SymbolNormalize::normalizeRankedSymbols ( std::move ( m_Data ) );

		return new PrefixRankedBarTree < > ( std::move ( bars ), std::move ( alphabet ), std::move ( content ) );
	}
};

} /* namespace tree */

#include "RankedTree.h"

namespace tree {

template < class SymbolType, class RankType >
PrefixRankedBarTree < SymbolType, RankType >::PrefixRankedBarTree ( std::set < std::ranked_symbol < SymbolType, RankType > > bars, std::set < std::ranked_symbol < SymbolType, RankType > > alphabet, std::vector < std::ranked_symbol < SymbolType, RankType > > data ) : std::Components < PrefixRankedBarTree, std::ranked_symbol < SymbolType, RankType >, std::tuple < GeneralAlphabet, BarSymbols >, std::tuple < > > ( std::make_tuple ( std::move ( alphabet ), std::move ( bars ) ), std::tuple < > ( ) ) {
	setContent ( std::move ( data ) );
}

template < class SymbolType, class RankType >
PrefixRankedBarTree < SymbolType, RankType >::PrefixRankedBarTree ( std::set < std::ranked_symbol < SymbolType, RankType > > bars, std::vector < std::ranked_symbol < SymbolType, RankType > > data ) : PrefixRankedBarTree ( bars, bars + std::set < std::ranked_symbol < SymbolType, RankType > > ( data.begin ( ), data.end ( ) ), data ) {
}

template < class SymbolType, class RankType >
PrefixRankedBarTree < SymbolType, RankType >::PrefixRankedBarTree ( SymbolType barBase, const RankedTree < SymbolType, RankType > & tree ) : PrefixRankedBarTree ( TreeAuxiliary::computeBars ( tree.getAlphabet ( ), barBase ), tree.getAlphabet ( ) + TreeAuxiliary::computeBars ( tree.getAlphabet ( ), barBase ), toPrefixRankedBar ( tree.getContent ( ), barBase ) ) {
}

template < class SymbolType, class RankType >
PrefixRankedBarTree < SymbolType, RankType >::PrefixRankedBarTree ( const RankedTree < > & tree ) : PrefixRankedBarTree ( alphabet::BarSymbol::instance < SymbolType > ( ), tree ) {
}

template < class SymbolType, class RankType >
std::vector < std::ranked_symbol < SymbolType, RankType > > PrefixRankedBarTree < SymbolType, RankType >::toPrefixRankedBar ( const std::tree < std::ranked_symbol < SymbolType, RankType > > & tree, const SymbolType & barBase ) {
	std::vector < std::ranked_symbol < SymbolType, RankType > > res;

	for ( typename std::tree < std::ranked_symbol < SymbolType, RankType > >::const_structure_iterator iter = tree.structure_begin ( ); iter != tree.structure_end ( ); ++iter )
		if ( iter.getVirtual ( ) )
			res.push_back ( std::ranked_symbol < SymbolType, RankType > ( barBase, iter->getRank ( ) ) );
		else
			res.push_back ( * iter );

	return res;
}

template < class SymbolType, class RankType >
RankedTreeBase * PrefixRankedBarTree < SymbolType, RankType >::clone ( ) const {
	return new PrefixRankedBarTree ( * this );
}

template < class SymbolType, class RankType >
RankedTreeBase * PrefixRankedBarTree < SymbolType, RankType >::plunder ( ) && {
	return new PrefixRankedBarTree ( std::move ( * this ) );
}

template < class SymbolType, class RankType >
const std::vector < std::ranked_symbol < SymbolType, RankType > > & PrefixRankedBarTree < SymbolType, RankType >::getContent ( ) const {
	return this->m_Data;
}

template < class SymbolType, class RankType >
void PrefixRankedBarTree < SymbolType, RankType >::setContent ( std::vector < std::ranked_symbol < SymbolType, RankType > > data ) {
	arityChecksum ( data );

	std::set < std::ranked_symbol < SymbolType, RankType > > minimalAlphabet ( data.begin ( ), data.end ( ) );
	std::set < std::ranked_symbol < SymbolType, RankType > > unknownSymbols;
	std::set_difference ( minimalAlphabet.begin ( ), minimalAlphabet.end ( ), getAlphabet().begin ( ), getAlphabet().end ( ), std::inserter ( unknownSymbols, unknownSymbols.end ( ) ) );

	if ( unknownSymbols.size ( ) > 0 )
		throw TreeException ( "Input symbols not in the alphabet." );

	this->m_Data = std::move ( data );
}

template < class SymbolType, class RankType >
void PrefixRankedBarTree < SymbolType, RankType >::arityChecksum ( const std::vector < std::ranked_symbol < SymbolType, RankType > > & data ) {
	int arityChecksumTerminals = 1;
	int arityChecksumBars  = 1;
	int arityChecksumTypes = 0;

	for ( const std::ranked_symbol < SymbolType, RankType > & symbol : data ) {
		if ( getBars ( ).count ( symbol ) ) {
			arityChecksumBars += ( unsigned ) symbol.getRank ( );
			arityChecksumBars -= 1;
			arityChecksumTypes -= 1;
		} else {
			arityChecksumTerminals += ( unsigned ) symbol.getRank ( );
			arityChecksumTerminals -= 1;
			arityChecksumTypes += 1;
		}
	}

	if ( ( arityChecksumTerminals != 0 ) || ( arityChecksumBars != 0 ) || ( arityChecksumTypes != 0 ) ) throw TreeException ( "The string does not form a tree" );
}

template < class SymbolType, class RankType >
bool PrefixRankedBarTree < SymbolType, RankType >::isEmpty ( ) const {
	return this->m_Data.size ( ) == 0;
}

template < class SymbolType, class RankType >
int PrefixRankedBarTree < SymbolType, RankType >::compare ( const PrefixRankedBarTree & other ) const {
	auto first = std::tie ( m_Data, getAlphabet(), getBars() );
	auto second = std::tie ( other.m_Data, other.getAlphabet(), other.getBars() );

	static std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

template < class SymbolType, class RankType >
void PrefixRankedBarTree < SymbolType, RankType >::operator >>( std::ostream & out ) const {
	out << "(PrefixRankedBarTree ";

	for ( const std::ranked_symbol < SymbolType, RankType > & symbol : this->m_Data )
		out << symbol;

	out << ")";
}

template < class SymbolType, class RankType >
PrefixRankedBarTree < SymbolType, RankType >::operator std::string ( ) const {
	std::stringstream ss;
	ss << "\"";

	for ( const std::ranked_symbol < SymbolType, RankType > & symbol : this->m_Data )
		ss << symbol;

	ss << "\"";
	return std::move ( ss ).str ( );
}

template < class SymbolType, class RankType >
PrefixRankedBarTree < SymbolType, RankType > PrefixRankedBarTree < SymbolType, RankType >::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, PrefixRankedBarTree::getXmlTagName() );
	std::set < std::ranked_symbol < SymbolType, RankType > > bars = TreeFromXMLParser::parseBars < std::ranked_symbol < SymbolType, RankType > > ( input );
	std::set < std::ranked_symbol < SymbolType, RankType > > rankedAlphabet = TreeFromXMLParser::parseAlphabet < std::ranked_symbol < SymbolType, RankType > > ( input );
	std::vector < std::ranked_symbol < SymbolType, RankType > > data = TreeFromXMLParser::parseLinearContent < std::ranked_symbol < SymbolType, RankType > > ( input );
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, PrefixRankedBarTree::getXmlTagName() );

	return PrefixRankedBarTree < SymbolType, RankType > ( std::move ( bars ), std::move ( rankedAlphabet ), std::move ( data ) );
}

template < class SymbolType, class RankType >
void PrefixRankedBarTree < SymbolType, RankType >::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( PrefixRankedBarTree::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	TreeToXMLComposer::composeBars ( out, getBars() );
	TreeToXMLComposer::composeAlphabet ( out, getAlphabet() );
	TreeToXMLComposer::composeContent ( out, m_Data );
	out.emplace_back ( PrefixRankedBarTree::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

template < class SymbolType, class RankType >
alib::ObjectBase* PrefixRankedBarTree < SymbolType, RankType >::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace tree */

namespace std {

template < class SymbolType, class RankType >
class ComponentConstraint< ::tree::PrefixRankedBarTree < SymbolType, RankType >, std::ranked_symbol < SymbolType, RankType >, ::tree::GeneralAlphabet > {
public:
	static bool used ( const ::tree::PrefixRankedBarTree < SymbolType, RankType > & tree, const std::ranked_symbol < SymbolType, RankType > & symbol ) {
		const std::vector < std::ranked_symbol < SymbolType, RankType > > & content = tree.getContent ( );

		return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( );
	}

	static bool available ( const ::tree::PrefixRankedBarTree < SymbolType, RankType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
		return true;
	}

	static void valid ( const ::tree::PrefixRankedBarTree < SymbolType, RankType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
	}
};

template < class SymbolType, class RankType >
class ComponentConstraint< ::tree::PrefixRankedBarTree < SymbolType, RankType >, std::ranked_symbol < SymbolType, RankType >, ::tree::BarSymbols > {
public:
	static bool used ( const ::tree::PrefixRankedBarTree < SymbolType, RankType > & tree, const std::ranked_symbol < SymbolType, RankType > & symbol ) {
		const std::vector < std::ranked_symbol < SymbolType, RankType > > & content = tree.getContent ( );

		return std::find ( content.begin ( ), content.end ( ), symbol ) != content.end ( );
	}

	static bool available ( const ::tree::PrefixRankedBarTree < SymbolType, RankType > & tree, const std::ranked_symbol < SymbolType, RankType > & symbol ) {
		return tree.template accessComponent < ::tree::GeneralAlphabet > ( ).get ( ).count ( symbol );
	}

	static void valid ( const ::tree::PrefixRankedBarTree < SymbolType, RankType > &, const std::ranked_symbol < SymbolType, RankType > & ) {
	}
};

} /* namespace std */

#endif /* PREFIX_RANKED_BAR_TREE_H_ */
