/*
 * SuffixTrieTerminatingSymbol.cpp
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#include "SuffixTrieTerminatingSymbol.h"
#include <exception/CommonException.h>

#include <iostream>
#include <algorithm>
#include <sstream>

#include <sax/FromXMLParserHelper.h>
#include "../common/IndexFromXMLParser.h"
#include "../common/IndexToXMLComposer.h"
#include <object/Object.h>
#include <registration/TypeRegistration.hpp>
#include <object/UniqueObject.h>

namespace indexes {

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( std::set < DefaultSymbolType > alphabet, DefaultSymbolType terminatingSymbol ) : SuffixTrieTerminatingSymbol ( std::move ( alphabet ), std::move ( terminatingSymbol ), SuffixTrieNodeTerminatingSymbol ( { } ) ) {
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( std::set < DefaultSymbolType > alphabet, DefaultSymbolType terminatingSymbol, SuffixTrieNodeTerminatingSymbol tree ) : std::Components < SuffixTrieTerminatingSymbol, DefaultSymbolType, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > ( std::make_tuple ( std::move ( alphabet ) ), std::make_tuple ( std::move ( terminatingSymbol ) ) ), m_tree ( NULL ) {
	setTree ( std::move ( tree ) );
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( DefaultSymbolType terminatingSymbol, SuffixTrieNodeTerminatingSymbol tree ) : SuffixTrieTerminatingSymbol ( tree.computeMinimalAlphabet ( ) + std::set < DefaultSymbolType > { terminatingSymbol }, terminatingSymbol, tree ) {
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( const SuffixTrieTerminatingSymbol & other ) : std::Components < SuffixTrieTerminatingSymbol, DefaultSymbolType, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > ( std::make_tuple ( other.getAlphabet ( ) ), std::make_tuple ( other.getTerminatingSymbol ( ) ) ), m_tree ( other.m_tree->clone ( ) ) {
	this->m_tree->attachTree ( this );
}

SuffixTrieTerminatingSymbol::SuffixTrieTerminatingSymbol ( SuffixTrieTerminatingSymbol && other ) noexcept : std::Components < SuffixTrieTerminatingSymbol, DefaultSymbolType, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > ( std::make_tuple ( std::move ( other.accessComponent < GeneralAlphabet > ( ).get ( ) ) ), std::make_tuple ( std::move ( other.accessElement < TerminatingSymbol > ( ).get ( ) ) ) ), m_tree ( other.m_tree ) {
	this->m_tree->attachTree ( this );
	other.m_tree = NULL;
}

alib::ObjectBase * SuffixTrieTerminatingSymbol::clone ( ) const {
	return new SuffixTrieTerminatingSymbol ( * this );
}

alib::ObjectBase * SuffixTrieTerminatingSymbol::plunder ( ) && {
	return new SuffixTrieTerminatingSymbol ( std::move ( * this ) );
}

SuffixTrieTerminatingSymbol & SuffixTrieTerminatingSymbol::operator =( const SuffixTrieTerminatingSymbol & other ) {
	if ( this == & other )
		return * this;

	* this = SuffixTrieTerminatingSymbol ( other );

	return * this;
}

SuffixTrieTerminatingSymbol & SuffixTrieTerminatingSymbol::operator =( SuffixTrieTerminatingSymbol && other ) noexcept {
	std::swap ( this->m_tree, other.m_tree );
	std::swap ( accessComponent < GeneralAlphabet > ( ).get ( ), other.accessComponent < GeneralAlphabet > ( ).get ( ) );
	std::swap ( accessElement < TerminatingSymbol > ( ).get ( ), other.accessElement < TerminatingSymbol > ( ).get ( ) );

	return * this;
}

SuffixTrieTerminatingSymbol::~SuffixTrieTerminatingSymbol ( ) noexcept {
	delete m_tree;
}

const SuffixTrieNodeTerminatingSymbol & SuffixTrieTerminatingSymbol::getRoot ( ) const {
	return * m_tree;
}

SuffixTrieNodeTerminatingSymbol & SuffixTrieTerminatingSymbol::getRoot ( ) {
	return * m_tree;
}

void SuffixTrieTerminatingSymbol::setTree ( SuffixTrieNodeTerminatingSymbol tree ) {
	delete this->m_tree;
	this->m_tree = std::move ( tree ).plunder ( );

	if ( !this->m_tree->attachTree ( this ) ) {
		delete this->m_tree;
		throw exception::CommonException ( "Input symbols not in the alphabet." );
	}
}

void SuffixTrieTerminatingSymbol::operator >>( std::ostream & out ) const {
	out << "(SuffixTrieTerminatingSymbol " << * ( this->m_tree ) << ")";
}

std::ostream & operator <<( std::ostream & out, const SuffixTrieTerminatingSymbol & instance ) {
	instance >> out;
	return out;
}

int SuffixTrieTerminatingSymbol::compare ( const SuffixTrieTerminatingSymbol & other ) const {
	auto first = std::tie ( * m_tree, getAlphabet(), getTerminatingSymbol() );
	auto second = std::tie ( * other.m_tree, other.getAlphabet(), other.getTerminatingSymbol() );

	static std::compare < decltype ( first ) > comp;

	return comp ( first, second );
}

SuffixTrieTerminatingSymbol::operator std::string ( ) const {
	std::stringstream ss;
	ss << * this;
	return ss.str ( );
}

SuffixTrieTerminatingSymbol SuffixTrieTerminatingSymbol::parse ( std::deque < sax::Token >::iterator & input ) {
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::START_ELEMENT, SuffixTrieTerminatingSymbol::getXmlTagName() );
	std::set < DefaultSymbolType > rankedAlphabet = IndexFromXMLParser::parseAlphabet ( input );
	DefaultSymbolType terminatingSymbol = alib::xmlApi < DefaultSymbolType >::parse ( input );
	SuffixTrieNodeTerminatingSymbol * root = IndexFromXMLParser::parseSuffixTrieNodeTerminatingSymbol ( input );
	SuffixTrieTerminatingSymbol tree ( std::move ( rankedAlphabet ), terminatingSymbol, std::move ( * root ) );

	delete root;
	sax::FromXMLParserHelper::popToken ( input, sax::Token::TokenType::END_ELEMENT, SuffixTrieTerminatingSymbol::getXmlTagName() );
	return tree;
}

void SuffixTrieTerminatingSymbol::compose ( std::deque < sax::Token > & out ) const {
	out.emplace_back ( SuffixTrieTerminatingSymbol::getXmlTagName(), sax::Token::TokenType::START_ELEMENT );
	IndexToXMLComposer::composeAlphabet ( out, getAlphabet ( ) );
	alib::xmlApi < DefaultSymbolType >::compose ( out, getTerminatingSymbol ( ) );
	IndexToXMLComposer::composeNode ( out, * m_tree );
	out.emplace_back ( SuffixTrieTerminatingSymbol::getXmlTagName(), sax::Token::TokenType::END_ELEMENT );
}

alib::ObjectBase* SuffixTrieTerminatingSymbol::inc() && {
	return new alib::UniqueObject(alib::Object(std::move(*this)), primitive::Integer(0));
}

} /* namespace indexes */

namespace alib {

auto suffixTreeTerminatingSymbolType = registration::TypeRegister < alib::Object, indexes::SuffixTrieTerminatingSymbol > ( );

} /* namespace alib */
