/*
 * LinearStringTerminatingSymbol.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef LINEAR_STRING_TERMINATING_SYMBOL_H_
#define LINEAR_STRING_TERMINATING_SYMBOL_H_

#include <iostream>
#include <set>
#include <vector>

#include <string/StringBase.h>
#include <string/StringFeatures.h>

#include <alphabet/Symbol.h>
#include <core/components.hpp>

namespace string {

class GeneralAlphabet;
class TerminatingSymbol;

/**
 * Represents regular expression parsed from the XML. Regular expression is stored
 * as a tree of LinearStringElement.
 */
class LinearStringTerminatingSymbol final : public StringBase, public std::Components < LinearStringTerminatingSymbol, DefaultSymbolType, std::tuple < GeneralAlphabet >, std::tuple < TerminatingSymbol > > {
	std::vector < DefaultSymbolType > m_Data;

public:
	explicit LinearStringTerminatingSymbol ( DefaultSymbolType terminatingSymbol );
	explicit LinearStringTerminatingSymbol ( std::set < DefaultSymbolType > alphabet, DefaultSymbolType terminatingSymbol, std::vector < DefaultSymbolType > data );
	explicit LinearStringTerminatingSymbol ( DefaultSymbolType terminatingSymbol, std::vector < DefaultSymbolType > data );
	explicit LinearStringTerminatingSymbol ( DefaultSymbolType terminatingSymbol, const LinearString < > & str );

	virtual StringBase * clone ( ) const;
	virtual StringBase * plunder ( ) &&;

	virtual const std::set < DefaultSymbolType > & getAlphabet ( ) const {
		return accessComponent < GeneralAlphabet > ( ).get ( );
	}

	void extendAlphabet ( const std::set < DefaultSymbolType > & symbols ) {
		accessComponent < GeneralAlphabet > ( ).add( symbols );
	}

	const DefaultSymbolType & getTerminatingSymbol ( ) const {
		return accessElement < TerminatingSymbol > ( ).get ( );
	}

	/**
	 * @param element to append
	 */
	void appendSymbol ( DefaultSymbolType symbol );

	/**
	 * @return List of symbols forming string (const version).
	 */
	const std::vector < DefaultSymbolType > & getContent ( ) const;

	void setContent ( std::vector < DefaultSymbolType > data );

	/**
	 * @return true if string is an empty word (vector length is 0)
	 */
	bool isEmpty ( ) const;

	virtual int compare ( const ObjectBase & other ) const {
		if ( std::type_index ( typeid ( * this ) ) == std::type_index ( typeid ( other ) ) ) return this->compare ( ( decltype ( * this ) )other );

		return std::type_index ( typeid ( * this ) ) - std::type_index ( typeid ( other ) );
	}

	virtual int compare ( const LinearStringTerminatingSymbol & other ) const;

	virtual void operator >>( std::ostream & out ) const;

	virtual explicit operator std::string ( ) const;

	static const std::string & getXmlTagName() {
		static std::string xmlTagName = "LinearStringTerminatingSymbol";

		return xmlTagName;
	}

	static LinearStringTerminatingSymbol parse ( std::deque < sax::Token >::iterator & input );

	void compose ( std::deque < sax::Token > & out ) const;

	virtual alib::ObjectBase * inc ( ) &&;

	typedef LinearStringTerminatingSymbol normalized_type;

	virtual LinearStringTerminatingSymbol * normalize ( ) && {
		return this;
	}
};

} /* namespace string */

namespace std {

template < >
class ComponentConstraint< ::string::LinearStringTerminatingSymbol, DefaultSymbolType, ::string::GeneralAlphabet > {
public:
	static bool used ( const ::string::LinearStringTerminatingSymbol & str, const DefaultSymbolType & symbol ) {
		const std::vector<DefaultSymbolType>& content = str.getContent ( );
		return std::find(content.begin(), content.end(), symbol) != content.end();
	}

	static bool available ( const ::string::LinearStringTerminatingSymbol &, const DefaultSymbolType & ) {
		return true;
	}

	static void valid ( const ::string::LinearStringTerminatingSymbol &, const DefaultSymbolType & ) {
	}
};

template < >
class ElementConstraint< ::string::LinearStringTerminatingSymbol, DefaultSymbolType, ::string::TerminatingSymbol > {
public:
	static bool available ( const ::string::LinearStringTerminatingSymbol & str, const DefaultSymbolType & symbol ) {
		return str.getAlphabet ( ).count ( symbol );
	}

	static void valid ( const ::string::LinearStringTerminatingSymbol &, const DefaultSymbolType & ) {
	}
};

} /* namespace std */

#endif /* LINEAR_STRING_TERMINATING_SYMBOL_H_ */
