/*
 * fdstream.hpp
 *
 * Created on: Feb 12, 2016
 * Author: Radovan Cerveny
 */

/*
 * ofdstream
 * ofdstream takes a primary file descriptor (fd) and fallback file descriptor by constructor to use for output
 * if the primary fd is not valid for any reason, ofdstream will try to use fallback fd, if even fallback fd is not valid, ofdstream sets its state to ios::fail
 * if the primary fd (or fallback fd) is valid , it is used for successive buffered output
 *
 * supplied file descriptor has to be already open in order for ofdstream to work
 * ofdstream does not close the file descriptor and does not change internal flags of the file descriptor during the lifetime of ofdstream and during destruction of ofdstream
 *
 * typical use is to use as primary fd some arbitrary fd (5) and for fallback fd some standard fd (2) if we wish to see the output
 */

#ifndef __FDSTREAM_HPP_
#define __FDSTREAM_HPP_

#include <unistd.h>
#include <fcntl.h>
#include <ostream>
#include <istream>
#include <array>

namespace std {

extern const int FAIL_FD;

class fdstreambuf : public streambuf {
	static const size_t buff_sz = 512;

	int fd;
	array < char_type, buff_sz > buff;

	bool flush ( );

protected:
	int_type pbackfail ( int_type );

	int_type underflow ( );
	int_type overflow ( int_type );
	int sync ( );

public:
	explicit fdstreambuf ( int );
	virtual ~fdstreambuf ( );
	fdstreambuf ( const fdstreambuf & ) = delete;
	fdstreambuf & operator =( const fdstreambuf & ) = delete;
};

class fdaccessor {
	int fd;
	bool redirected;

public:
	explicit fdaccessor ( int, int );
	int get_fd ( ) const;
	bool is_redirected ( ) const;
};

class ofdstream : public ostream {
	fdaccessor fda;
	fdstreambuf fdbuf;

public:
	explicit ofdstream ( int, int = FAIL_FD);
	virtual ~ofdstream ( );
	bool is_redirected ( ) const;
};

class ifdstream : public istream {
	fdaccessor fda;
	fdstreambuf fdbuf;

public:
	explicit ifdstream ( int, int = FAIL_FD);
	virtual ~ifdstream ( );
	bool is_redirected ( ) const;
};

} /* namespace std */

#endif /* __FDSTREAM_HPP_ */
