/*
 * ObjectToStringComposer.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnciek
 */

#ifndef OBJECT_TO_STRING_COMPOSER_H_
#define OBJECT_TO_STRING_COMPOSER_H_

#include <ostream>
#include <core/multipleDispatch.hpp>
#include <object/Object.h>
#include <object/Void.h>
#include <container/ObjectsPair.h>
#include <container/ObjectsSet.h>
#include <container/ObjectsVector.h>

#include "alphabet/LabeledSymbol.h"
#include "alphabet/StartSymbol.h"
#include "alphabet/EndSymbol.h"
#include "alphabet/BottomOfTheStackSymbol.h"
#include "alphabet/BlankSymbol.h"
#include "alphabet/BarSymbol.h"
#include "alphabet/VariablesBarSymbol.h"
#include "alphabet/SubtreeWildcardSymbol.h"
#include "alphabet/NonlinearVariableSymbol.h"
#include "alphabet/RankedSymbol.h"
#include "alphabet/UniqueSymbol.h"

#include "label/Label.h"
#include "label/PrimitiveLabel.h"
#include "label/HexavigesimalLabel.h"
#include "label/ObjectLabel.h"
#include "label/UniqueLabel.h"
#include "label/InitialStateLabel.h"
#include "label/FinalStateLabel.h"
#include "label/FailStateLabel.h"

#include "primitive/Primitive.h"
#include "primitive/Integer.h"
#include "primitive/String.h"
#include "primitive/Character.h"
#include "primitive/Unsigned.h"
#include "primitive/Bool.h"

namespace alib {

class ObjectToStringComposer;

typedef std::SingleDispatchFirstStaticParam<ObjectToStringComposer, void, std::ostream&, const CommonBaseBase &> ObjectBaseDispatcher;
typedef std::SingleDispatchFirstStaticParam<ObjectToStringComposer, void, std::ostream&, const WrapperBaseBase &> WrapperBaseDispatcher;

/**
 * This class contains methods to print XML representation of string to the output stream.
 */
class ObjectToStringComposer : public ObjectBaseDispatcher, public WrapperBaseDispatcher {
public:
	static void compose(std::ostream& output, const primitive::Integer& primitive);
	static void compose(std::ostream& output, const primitive::String& primitive);
	static void compose(std::ostream& output, const primitive::Character& primitive);
	static void compose(std::ostream& output, const primitive::Unsigned& primitive);
	static void compose(std::ostream& output, unsigned primitive);
	static void compose(std::ostream& output, const primitive::Bool& primitive);

	/**
	 * Prints text representation of Primitive to the output stream.
	 * @param string String to print
	 * @param out output stream to which print the String
	 */
	static void compose(std::ostream& output, const primitive::Primitive& string);

	static void compose(std::ostream& output, const label::PrimitiveLabel& label);
	static void compose(std::ostream& output, const label::HexavigesimalLabel& label);
	static void compose(std::ostream& output, const label::ObjectLabel& label);
	static void compose(std::ostream& output, const label::UniqueLabel& label);
	static void compose(std::ostream& output, const label::FailStateLabel& label);
	static void compose(std::ostream& output, const label::InitialStateLabel& label);
	static void compose(std::ostream& output, const label::FinalStateLabel& label);

	/**
	 * Prints text representation of Label to the output stream.
	 * @param string String to print
	 * @param out output stream to which print the String
	 */
	static void compose(std::ostream& output, const label::Label& string);

	static void compose(std::ostream& output, const alphabet::LabeledSymbol& symbol);
	static void compose(std::ostream& output, const alphabet::BlankSymbol& symbol);
	static void compose(std::ostream& output, const alphabet::BottomOfTheStackSymbol& symbol);
	static void compose(std::ostream& output, const alphabet::EndSymbol& symbol);
	static void compose(std::ostream& output, const alphabet::StartSymbol& symbol);
	template < class SymbolType, class RankType >
	static void compose(std::ostream& output, const alphabet::RankedSymbol < SymbolType, RankType > & symbol);
	static void compose(std::ostream& output, const alphabet::BarSymbol& symbol);
	static void compose(std::ostream& output, const alphabet::VariablesBarSymbol& symbol);
	static void compose(std::ostream& output, const alphabet::SubtreeWildcardSymbol& symbol);
	template < class SymbolType >
	static void compose(std::ostream& output, const alphabet::NonlinearVariableSymbol < SymbolType > & symbol);
	static void compose(std::ostream& output, const alphabet::UniqueSymbol& symbol);

	/**
	 * Prints text representation of Symbol to the output stream.
	 * @param string String to print
	 * @param out output stream to which print the String
	 */
	static void compose(std::ostream& output, const alphabet::Symbol& string);

	/**
	 * Prints text representation of Object to the output stream.
	 * @param string Object to print
	 * @param out output stream to which print the String
	 */
	static void compose(std::ostream& output, const Object& object);
	static void compose(std::ostream& output, const Void& object);
	template < class T >
	static void compose(std::ostream& output, const AnyObject < T > & object);
	template < class T, class R >
	static void compose(std::ostream& output, const container::ObjectsPair < T, R > & object);
	template < class T, class R >
	static void compose(std::ostream& output, const std::pair < T, R > & object);
	template < class T >
	static void compose(std::ostream& output, const container::ObjectsSet < T > & object);
	template < class T >
	static void compose(std::ostream& output, const std::set < T > & object);
	template < class T >
	static void compose(std::ostream& output, const container::ObjectsVector < T > & object);
	template < class T >
	static void compose(std::ostream& output, const std::vector < T > & object);

	static void compose(std::ostream& output, const CommonBaseBase& object);
	static void compose(std::ostream& output, const WrapperBaseBase& object);
};

template < class SymbolType, class RankType >
void ObjectToStringComposer::compose(std::ostream& out, const alphabet::RankedSymbol < SymbolType, RankType > & symbol) {
	compose(out, symbol.getSymbol());
	compose(out, symbol.getRank());
}

template < class SymbolType >
void ObjectToStringComposer::compose(std::ostream& out, const alphabet::NonlinearVariableSymbol < SymbolType > & symbol ) {
	out << "$";
	compose(out, symbol.getSymbol());
}

template < class T >
void ObjectToStringComposer::compose(std::ostream& out, const AnyObject < T > & object) {
	compose ( out, object.getData ( ) );
}

template < class T, class R >
void ObjectToStringComposer::compose(std::ostream& out, const container::ObjectsPair < T, R > & object) {
	compose ( out, static_cast < const std::pair < T, R > > ( object ) );
}

template < class T, class R >
void ObjectToStringComposer::compose(std::ostream& out, const std::pair < T, R > & object) {
	out << '<';
	compose(out, object.first);
	out << ", ";
	compose(out, object.second);
	out << '>';
}

template < class T >
void ObjectToStringComposer::compose(std::ostream& out, const container::ObjectsSet < T > & object) {
	compose ( out, static_cast < const std::set < T > > ( object ) );
}

template < class T >
void ObjectToStringComposer::compose(std::ostream& out, const std::set < T > & object) {
	out << '[';
	bool first = true;
	for(const T & innerObject : object) {
		if(!first)
			out << ", ";
		else
			first = false;
		compose(out, innerObject);
	}
	out << ']';
}

template < class T >
void ObjectToStringComposer::compose(std::ostream& out, const container::ObjectsVector < T > & object) {
	compose ( out, static_cast < const std::vector < T > > ( object ) );
}

template < class T >
void ObjectToStringComposer::compose(std::ostream& out, const std::vector < T > & object) {
	out << '[';
	bool first = true;
	for(const T & innerObject : object) {
		if(!first)
			out << ", ";
		else
			first = false;
		compose(out, innerObject);
	}
	out << ']';
}

} /* namespace alib */

#endif /* OBJECT_TO_STRING_COMPOSER_H_ */
