/*
 * RegExpToStringComposer.h
 *
 *  Created on: Nov 23, 2013
 *      Author: Jan Travnicek
 */

#ifndef REX_EXP_TO_STRING_COMPOSER_H_
#define REX_EXP_TO_STRING_COMPOSER_H_

#include <string>
#include <core/multipleDispatch.hpp>
#include "regexp/RegExp.h"
#include "regexp/unbounded/UnboundedRegExp.h"
#include "regexp/unbounded/UnboundedRegExpElement.h"
#include "regexp/formal/FormalRegExp.h"
#include "regexp/formal/FormalRegExpElement.h"

#include "regexp/RegExpClasses.h"

#include "regexp/unbounded/UnboundedRegExpElements.h"
#include "regexp/formal/FormalRegExpElements.h"

#include <object/ObjectToStringComposer.h>

namespace regexp {

class RegExpToStringComposer : public std::SingleDispatchFirstStaticParam<RegExpToStringComposer, void, std::ostream&, const RegExpBase &> {
private:
	enum class Priority {
		ALTERNATION,
		CONCATENATION,
		FACTOR
	};

public:
	/**
	 * Composes string representation of RegExp.
	 * @param regexp RegExp to print
	 * @returns string representation of regexp
	 */
	static void compose(std::ostream& out, const RegExp& regexp);

	template < class SymbolType >
	static void compose(std::ostream& out, const UnboundedRegExp < SymbolType > & regexp);
	template < class SymbolType >
	static void compose(std::ostream& out, const UnboundedRegExpStructure < SymbolType > & regexp);
	template < class SymbolType >
	static void compose(std::ostream& out, const FormalRegExp < SymbolType > & regexp);
	template < class SymbolType >
	static void compose(std::ostream& out, const FormalRegExpStructure < SymbolType > & regexp);

	template < class SymbolType >
	class Unbounded {
	public:
		static void visit( const UnboundedRegExpAlternation < SymbolType > & alternation, std::tuple<Priority&, std::ostream &> & out );
		static void visit( const UnboundedRegExpConcatenation < SymbolType > & concatenation, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const UnboundedRegExpIteration < SymbolType > & iteration, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const UnboundedRegExpSymbol < SymbolType > & symbol, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const UnboundedRegExpEpsilon < SymbolType > & epsilon, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const UnboundedRegExpEmpty < SymbolType > & empty, std::tuple<Priority&, std::ostream &> & output);
	};

	template < class SymbolType >
	class Formal {
	public:
		static void visit( const FormalRegExpAlternation < SymbolType > & alternation, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const FormalRegExpConcatenation < SymbolType > & concatenation, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const FormalRegExpIteration < SymbolType > & iteration, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const FormalRegExpSymbol < SymbolType > & symbol, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const FormalRegExpEpsilon < SymbolType > & epsilon, std::tuple<Priority&, std::ostream &> & output);
		static void visit( const FormalRegExpEmpty < SymbolType > & empty, std::tuple<Priority&, std::ostream &> & output);
	};
};

template < class SymbolType >
void RegExpToStringComposer::compose(std::ostream& output, const UnboundedRegExpStructure < SymbolType > & regexp) {
	Priority tmp = Priority::ALTERNATION;
	std::tuple<Priority&, std::ostream&> out = std::tie(tmp, output);
	regexp.getStructure().template accept<void, RegExpToStringComposer::Unbounded < SymbolType > > (out);
}

template < class SymbolType >
void RegExpToStringComposer::compose(std::ostream& output, const FormalRegExpStructure < SymbolType > & regexp) {
	Priority tmp = Priority::ALTERNATION;
	std::tuple<Priority&, std::ostream&> out = std::tie(tmp, output);
	regexp.getStructure().template accept<void, RegExpToStringComposer::Formal < SymbolType > > (out);
}

template < class SymbolType >
void RegExpToStringComposer::Unbounded < SymbolType >::visit( const UnboundedRegExpAlternation < SymbolType > & alternation, std::tuple<Priority&, std::ostream &> & out ) {
	if(alternation.getElements().size() == 0) {
		std::get<1>(out) << "#0";
	} else if(alternation.getElements().size() == 1) {
		alternation.getElements()[0]->template accept<void, RegExpToStringComposer::Unbounded < SymbolType > > (out);
	} else {
		Priority outerPriorityMinimum = std::get<0>(out);
		if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << '(';
		bool first = true;
		for (const auto& element : alternation.getElements()) {
			if(first) {
				first = false;
			} else {
				std::get<1>(out) << '+';
			}
			std::get<0>(out) = Priority::ALTERNATION;
			element->template accept<void, RegExpToStringComposer::Unbounded < SymbolType > > (out);
		}
		if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << ')';
	}
}

template < class SymbolType >
void RegExpToStringComposer::Unbounded < SymbolType >::visit( const UnboundedRegExpConcatenation < SymbolType > & concatenation, std::tuple<Priority&, std::ostream &> & out ) {
	Priority outerPriorityMinimum = std::get<0>(out);
	if(concatenation.getElements().size() == 0) {
		std::get<1>(out) << "#E";
	} else if(concatenation.getElements().size() == 1) {
		concatenation.getElements()[0]->template accept<void, RegExpToStringComposer::Unbounded < SymbolType > > (out);
	} else {
		if(outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << '(';
		bool first = true;
		for (auto element : concatenation.getElements()) {
			if(first) {
				first = false;
			} else {
				std::get<1>(out) << ' ';
			}
			std::get<0>(out) = Priority::CONCATENATION;
			element->template accept<void, RegExpToStringComposer::Unbounded < SymbolType > > (out);
		}
		if(outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << ')';
	}
}

template < class SymbolType >
void RegExpToStringComposer::Unbounded < SymbolType >::visit( const UnboundedRegExpIteration < SymbolType > & iteration, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<0>(out) = Priority::FACTOR;
	iteration.getElement().template accept<void, RegExpToStringComposer::Unbounded < SymbolType > > (out);
	std::get<1>(out) << "*";
}

template < class SymbolType >
void RegExpToStringComposer::Unbounded < SymbolType >::visit( const UnboundedRegExpSymbol < SymbolType > & symbol, std::tuple<Priority&, std::ostream &> & out ) {
	alib::ObjectToStringComposer::compose(std::get<1>(out), symbol.getSymbol());
}

template < class SymbolType >
void RegExpToStringComposer::Unbounded < SymbolType >::visit( const UnboundedRegExpEpsilon < SymbolType > &, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<1>(out) << "#E";
}

template < class SymbolType >
void RegExpToStringComposer::Unbounded < SymbolType >::visit( const UnboundedRegExpEmpty < SymbolType > &, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<1>(out) << "#0";
}

template < class SymbolType >
void RegExpToStringComposer::Formal < SymbolType >::visit( const FormalRegExpAlternation < SymbolType > & alternation, std::tuple<Priority&, std::ostream &> & out ) {
	Priority outerPriorityMinimum = std::get<0>(out);
	if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << '(';

	std::get<0>(out) = Priority::ALTERNATION;
	alternation.getLeftElement().template accept<void, RegExpToStringComposer::Formal < SymbolType > > (out);
	std::get<1>(out) << '+';
	alternation.getRightElement().template accept<void, RegExpToStringComposer::Formal < SymbolType > > (out);

	if(outerPriorityMinimum == Priority::CONCATENATION || outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << ')';
}

template < class SymbolType >
void RegExpToStringComposer::Formal < SymbolType >::visit( const FormalRegExpConcatenation < SymbolType > & concatenation, std::tuple<Priority&, std::ostream &> & out ) {
	Priority outerPriorityMinimum = std::get<0>(out);
	if(outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << '(';

	std::get<0>(out) = Priority::CONCATENATION;
	concatenation.getLeftElement().template accept<void, RegExpToStringComposer::Formal < SymbolType > > (out);
	std::get<1>(out) << ' ';
	concatenation.getRightElement().template accept<void, RegExpToStringComposer::Formal < SymbolType > > (out);

	if(outerPriorityMinimum == Priority::FACTOR) std::get<1>(out) << ')';
}

template < class SymbolType >
void RegExpToStringComposer::Formal < SymbolType >::visit( const FormalRegExpIteration < SymbolType > & iteration, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<0>(out) = Priority::FACTOR;
	iteration.getElement().template accept<void, RegExpToStringComposer::Formal < SymbolType > > (out);
	std::get<1>(out) << "*";
}

template < class SymbolType >
void RegExpToStringComposer::Formal < SymbolType >::visit( const FormalRegExpSymbol < SymbolType > & symbol, std::tuple<Priority&, std::ostream &> & out ) {
	alib::ObjectToStringComposer::compose(std::get<1>(out), symbol.getSymbol());
}

template < class SymbolType >
void RegExpToStringComposer::Formal < SymbolType >::visit( const FormalRegExpEpsilon < SymbolType > &, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<1>(out) << "#E";
}

template < class SymbolType >
void RegExpToStringComposer::Formal < SymbolType >::visit( const FormalRegExpEmpty < SymbolType > &, std::tuple<Priority&, std::ostream &> & out ) {
	std::get<1>(out) << "#0";
}

} /* namespace regexp */

#endif /* REX_EXP_TO_STRING_COMPOSER_H_ */
