/*
 * arand.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Tomas Pecka
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>
#include <random>
#include <sax/FromXMLParserHelper.h>

#include <string/String.h>
#include <factory/XmlDataFactory.hpp>
#include "automaton/generate/RandomAutomatonFactory.h"
#include "grammar/generate/RandomGrammarFactory.h"
#include "regexp/generate/RandomRegExpFactory.h"
#include "string/generate/RandomStringFactory.h"
#include "string/generate/RandomSubstringFactory.h"
#include "tree/generate/RandomTreeFactory.h"
#include <automaton/generate/RandomizeAutomaton.h>
#include <grammar/generate/RandomizeGrammar.h>

int main ( int argc, char * argv[] ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Random generator binary", ' ', "0.01" );

		std::vector < std::string > allowed;
		allowed.push_back ( "FSM" );
		allowed.push_back ( "RFSM" );
		allowed.push_back ( "CFG" );
		allowed.push_back ( "RCFG" );
		allowed.push_back ( "RE" );
		allowed.push_back ( "ST" );
		allowed.push_back ( "SST" );
		allowed.push_back ( "UT" );
		allowed.push_back ( "UP" );
		allowed.push_back ( "RT" );
		allowed.push_back ( "RP" );
		allowed.push_back ( "RNP" );
		TCLAP::ValuesConstraint < std::string > allowedVals ( allowed );

		TCLAP::ValueArg < std::string > type ( "t", "type", "Type of generated structure", true, "FSM", & allowedVals );
		cmd.add ( type );
	
		TCLAP::ValueArg < unsigned > seed ( "s", "seed", "Seed random number", false, 0, "integer" );
		cmd.add ( seed );

		TCLAP::ValueArg < int > alphabetSize ( "", "terminals", "Number of terminals/alphabet size", false, 3, "integer" );
		cmd.add ( alphabetSize );

		TCLAP::SwitchArg integerSymbols ( "", "integerSymbols", "Use integer symbols instead of char symbols", false );
		cmd.add ( integerSymbols );

		TCLAP::SwitchArg randomizedAlphabet ( "", "randomizedTerminals", "Number of terminals/alphabet size", false );
		cmd.add ( randomizedAlphabet );

		TCLAP::ValueArg < double > density ( "", "density", "Automaton's/grammar's transition density", false, 10, "double" );
		cmd.add ( density );

		TCLAP::ValueArg < int > nonterminals ( "", "nonterminals", "Number of grammar's nononterminals", false, 5, "integer" );
		cmd.add ( nonterminals );

		TCLAP::ValueArg < int > nodes ( "", "nodes", "Number of automaton's or tree nodes", false, 6, "integer" );
		cmd.add ( nodes );

		TCLAP::ValueArg < int > states ( "", "states", "Number of automaton's states", false, 5, "integer" );
		cmd.add ( states );

		TCLAP::ValueArg < int > terminalNodes ( "", "leaves", "Number of tree's/regexp's terminal nodes", false, 5, "integer" );
		cmd.add ( terminalNodes );

		TCLAP::ValueArg < int > height ( "", "height", "Height of the tree/regexp", false, 3, "integer" );
		cmd.add ( height );

		TCLAP::ValueArg < int > length ( "", "length", "Length of the string", false, 5, "integer" );
		cmd.add ( length );

		TCLAP::ValueArg < int > maxRank ( "", "rank", "Maximal rank of tree nodes", false, 5, "integer" );
		cmd.add ( maxRank );

		TCLAP::SwitchArg singleNonlinearVariable ( "", "single_nonlinear_variable", "Generate single nonlinear variable", false );
		cmd.add ( singleNonlinearVariable );

		TCLAP::ValueArg < std::string > input ( "i", "input", "Input to use in generation", false, "-", "file" );
		cmd.add ( input );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if ( verbose.isSet ( ) )
			common::GlobalData::verbose = true;

		if ( measure.isSet ( ) )
			common::GlobalData::measure = true;

		if ( seed.isSet ( ) )
			std::random_devices::semirandom.seed ( seed.getValue ( ) );

		measurements::start ( "Overal", measurements::Type::OVERALL );

		if ( !type.isSet ( ) ) throw exception::CommonException ( "Type is not defined." );

		if ( type.getValue ( ) == "FSM" ) {
			if ( ( density.getValue ( ) < 1 ) || ( density.getValue ( ) > 100 ) )
				 // TODO: floating point arithmetic
				throw exception::CommonException ( "You must specify density as a number between 1 and 100." );

			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::NFA < > res = automaton::generate::RandomAutomatonFactory::generateNFA ( states.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ), density.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "RFSM" ) {
			automaton::Automaton baseAutomaton = alib::XmlDataFactory::fromTokens ( sax::FromXMLParserHelper::parseInput ( input ) );

			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton res = automaton::generate::RandomizeAutomaton::randomize ( baseAutomaton );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "CFG" ) {
			if ( ( density.getValue ( ) < 1 ) || ( density.getValue ( ) > 100 ) )
				 // TODO: floating point arithmetic
				throw exception::CommonException ( "You must specify density as a number between 1 and 100." );

			measurements::start ( "Algorithm", measurements::Type::MAIN );

			grammar::CFG < > res = grammar::generate::RandomGrammarFactory::generateCFG ( nonterminals.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ), density.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "RCFG" ) {
			grammar::Grammar baseGrammar = alib::XmlDataFactory::fromTokens ( sax::FromXMLParserHelper::parseInput ( input ) );

			measurements::start ( "Algorithm", measurements::Type::MAIN );

			grammar::Grammar res = grammar::generate::RandomizeGrammar::randomize ( baseGrammar );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "RE" ) {
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			regexp::UnboundedRegExp < > res = regexp::generate::RandomRegExpFactory::generateUnboundedRegExp ( terminalNodes.getValue ( ), height.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "ST" ) {
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			string::LinearString < > res = string::generate::RandomStringFactory::generateLinearString ( length.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ), integerSymbols.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "SST" ) {
			string::String baseString = alib::XmlDataFactory::fromTokens ( sax::FromXMLParserHelper::parseInput ( input ) );

			measurements::start ( "Algorithm", measurements::Type::MAIN );

			string::String res = string::generate::RandomSubstringFactory::generateSubstring ( length.getValue ( ), baseString );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "UT" ) {
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			tree::UnrankedTree < > res = tree::generate::RandomTreeFactory::generateUnrankedTree ( height.getValue ( ), nodes.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ), maxRank.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "UP" ) {
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			tree::UnrankedPattern < > res = tree::generate::RandomTreeFactory::generateUnrankedPattern ( height.getValue ( ), nodes.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ), maxRank.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "RT" ) {
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			tree::RankedTree < > res = tree::generate::RandomTreeFactory::generateRankedTree ( height.getValue ( ), nodes.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ), maxRank.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "RP" ) {
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			tree::RankedPattern < > res = tree::generate::RandomTreeFactory::generateRankedPattern ( height.getValue ( ), nodes.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ), maxRank.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( type.getValue ( ) == "RNP" ) {
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			tree::RankedNonlinearPattern < > res = tree::generate::RandomTreeFactory::generateRankedNonlinearPattern ( height.getValue ( ), nodes.getValue ( ), alphabetSize.getValue ( ), randomizedAlphabet.getValue ( ), singleNonlinearVariable.getValue ( ), maxRank.getValue ( ) );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else {
			throw exception::CommonException ( "Invalid type." );
		}

		measurements::end ( );
		measurements::end ( );

		if ( measure.getValue ( ) ) std::cmeasure << measurements::results ( ) << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		std::cout << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		std::cerr << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
