/*
 * astringology.cpp
 *
 *  Created on: 26. 3. 2014
 *	  Author: Jan Travnicek
 */

#include <tclap/CmdLine.h>
#include <global/GlobalData.h>
#include <measure>
#include <vector>
#include <sax/FromXMLParserHelper.h>

#include <factory/XmlDataFactory.hpp>
#include <exception/CommonException.h>
#include <string/String.h>
#include <automaton/Automaton.h>
#include <container/Container.h>
#include <container/ObjectsSet.h>
#include <container/ObjectsVector.h>

#include <stringology/exact/ExactFactorMatch.h>
#include <stringology/exact/BoyerMooreHorspool.h>
#include <stringology/exact/BoyerMoore.h>
#include <stringology/exact/ReversedBoyerMooreHorspool.h>
#include <stringology/exact/DeadZoneUsingBadCharacterShift.h>
#include <stringology/query/BNDMOccurrences.h>
#include <stringology/query/BackwardOracleMatching.h>
#include <stringology/query/BackwardDAWGMatching.h>
#include <stringology/matching/ExactMatchingAutomaton.h>
#include <stringology/indexing/NondeterministicExactFactorAutomaton.h>
#include <stringology/indexing/ExactSubsequenceAutomaton.h>
#include <stringology/indexing/ExactNondeterministicSubsequenceAutomaton.h>
#include <stringology/indexing/ExactMultiNondeterministicSubsequenceAutomaton.h>
#include <stringology/indexing/ExactFactorOracleAutomaton.h>
#include <stringology/indexing/NondeterministicExactSuffixAutomaton.h>
#include <string/properties/BorderArray.h>
#include <stringology/indexing/SuffixTrieNaive.h>
#include <stringology/indexing/PositionHeapNaive.h>
#include <stringology/indexing/SuffixArrayNaive.h>
#include <stringology/indexing/BitParallelIndexConstruction.h>
#include <stringology/matching/BNDMMatcherConstruction.h>
#include <stringology/indexing/CompressedBitParallelIndexConstruction.h>

int main ( int argc, char * argv[] ) {
	try {
		common::GlobalData::argc = argc;
		common::GlobalData::argv = argv;

		TCLAP::CmdLine cmd ( "Stringology algorithm access binary", ' ', "0.01" );

		std::vector < std::string > allowed;
		allowed.push_back ( "exactMatchingAutomaton" );
		allowed.push_back ( "exactNondeterministicFactorAutomaton" );
		allowed.push_back ( "exactSubsequenceAutomaton" );
		allowed.push_back ( "exactNondeterministicSubsequenceAutomaton" );
		allowed.push_back ( "exactMultiNondeterministicSubsequenceAutomaton" );
		allowed.push_back ( "exactFactorOracleAutomaton" );
		allowed.push_back ( "exactNondeterministicSuffixAutomaton" );
		allowed.push_back ( "exactFactorMatch" );
		allowed.push_back ( "boyerMooreHorspool" );
		allowed.push_back ( "boyerMoore" );
		allowed.push_back ( "reversedBoyerMooreHorspool" );
		allowed.push_back ( "deadZoneUsingBadCharacterShift" );
		allowed.push_back ( "bndmOccurrences" );
		allowed.push_back ( "backwardOracleMatching" );
		allowed.push_back ( "backwardDAWGMatching" );
		allowed.push_back ( "suffixTrie" );
		allowed.push_back ( "positionHeap" );
		allowed.push_back ( "bitParallelIndex" );
		allowed.push_back ( "bndmMatcher" );
		allowed.push_back ( "compressedBitParallelIndex" );
		allowed.push_back ( "suffixArray" );

		allowed.push_back ( "borderArray" );
		allowed.push_back ( "badCharacterShiftTable" );
		allowed.push_back ( "goodSuffixShiftTable" );
		TCLAP::ValuesConstraint < std::string > allowedVals ( allowed );

		TCLAP::ValueArg < std::string > algorithm ( "a", "algorithm", "Execute algorithm", false, "exactFactorMatch", & allowedVals );
		cmd.add ( algorithm );

		TCLAP::MultiArg < std::string > subjectInput ( "s", "subject", "Subject string from file", false, "file" );
		cmd.add ( subjectInput );

		TCLAP::MultiArg < std::string > patternInput ( "p", "pattern", "Pattern string from file", false, "file" );
		cmd.add ( patternInput );

		TCLAP::SwitchArg measure ( "m", "measure", "Measure times", false );
		cmd.add ( measure );

		TCLAP::SwitchArg verbose ( "v", "verbose", "Be verbose", false );
		cmd.add ( verbose );

		cmd.parse ( argc, argv );

		if(verbose.isSet())
			common::GlobalData::verbose = true;
		if(measure.isSet())
			common::GlobalData::measure = true;

		measurements::start ( "Overal", measurements::Type::OVERALL );
		measurements::start ( "Input read", measurements::Type::AUXILIARY );

		if ( algorithm.getValue ( ) == "exactFactorMatch" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::ExactFactorMatch::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "boyerMooreHorspool" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::BoyerMooreHorspool::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "boyerMoore" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::BoyerMoore::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "reversedBoyerMooreHorspool" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::ReversedBoyerMooreHorspool::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "deadZoneUsingBadCharacterShift" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::exact::DeadZoneUsingBadCharacterShift::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "bndmOccurrences" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );
			indexes::stringology::BNDMMatcher < DefaultSymbolType > pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::query::BNDMOccurrences::query ( pattern, subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "backwardOracleMatching" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );
			automaton::Automaton pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::query::BackwardOracleMatching::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "backwardDAWGMatching" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );
			automaton::Automaton pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::set < unsigned > res = stringology::query::BackwardDAWGMatching::match ( subject, pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( res );
		} else if ( algorithm.getValue ( ) == "exactMatchingAutomaton" ) {
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::matching::ExactMatchingAutomaton::construct ( pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactNondeterministicFactorAutomaton" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::indexing::NondeterministicExactFactorAutomaton::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactSubsequenceAutomaton" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::indexing::ExactSubsequenceAutomaton::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactNondeterministicSubsequenceAutomaton" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::indexing::ExactNondeterministicSubsequenceAutomaton::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactMultiNondeterministicSubsequenceAutomaton" ) {
			std::set < string::String > subjects;

			for ( std::deque < sax::Token > & tokens : sax::FromXMLParserHelper::parseInput(false, subjectInput) ) {
				string::String string = alib::XmlDataFactory::fromTokens ( std::move ( tokens ) );
				subjects.insert ( std::move ( string ) );
			}

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::indexing::ExactMultiNondeterministicSubsequenceAutomaton::construct ( subjects );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactFactorOracleAutomaton" ) {
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::indexing::ExactFactorOracleAutomaton::construct ( pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "exactNondeterministicSuffixAutomaton" ) {
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			automaton::Automaton automaton = stringology::indexing::NondeterministicExactSuffixAutomaton::construct ( pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( automaton );
		} else if ( algorithm.getValue ( ) == "borderArray" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::vector < unsigned > borderArray = string::properties::BorderArray::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( borderArray );
		} else if ( algorithm.getValue ( ) == "suffixTrie" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			indexes::stringology::SuffixTrie < DefaultSymbolType > suffixTrie = stringology::indexing::SuffixTrieNaive::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( suffixTrie );
		} else if ( algorithm.getValue ( ) == "positionHeap" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			indexes::stringology::PositionHeap < DefaultSymbolType > positionHeap = stringology::indexing::PositionHeapNaive::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( positionHeap );
		} else if ( algorithm.getValue ( ) == "bitParallelIndex" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			indexes::stringology::BitParallelIndex < DefaultSymbolType > bitParallelIndex = stringology::indexing::BitParallelIndexConstruction::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( bitParallelIndex );
		} else if ( algorithm.getValue ( ) == "bndmMatcher" ) {
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			indexes::stringology::BNDMMatcher < DefaultSymbolType > bndmMatcher = stringology::matching::BNDMMatcherConstruction::construct ( pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( bndmMatcher );
		} else if ( algorithm.getValue ( ) == "compressedBitParallelIndex" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			indexes::stringology::CompressedBitParallelIndex < DefaultSymbolType > compressedBitParallelIndex = stringology::indexing::CompressedBitParallelIndexConstruction::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( compressedBitParallelIndex );
		} else if ( algorithm.getValue ( ) == "suffixArray" ) {
			string::String subject = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, subjectInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			indexes::stringology::SuffixArray < > suffixArray = stringology::indexing::SuffixArrayNaive::construct ( subject );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( suffixArray );
		} else if ( algorithm.getValue ( ) == "badCharacterShiftTable" ) {
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::map < DefaultSymbolType, size_t > bcs = string::properties::BadCharacterShiftTable::bcs ( pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( bcs );
		} else if ( algorithm.getValue ( ) == "goodSuffixShiftTable" ) {
			string::String pattern = alib::XmlDataFactory::fromTokens ( std::move ( sax::FromXMLParserHelper::parseInput(true, patternInput).front ( ) ) );

			measurements::end ( );
			measurements::start ( "Algorithm", measurements::Type::MAIN );

			std::vector < size_t > gss = string::properties::GoodSuffixShiftTable::gss ( pattern );

			measurements::end ( );
			measurements::start ( "Output write", measurements::Type::AUXILIARY );

			alib::XmlDataFactory::toStdout ( gss );
		} else {
			throw exception::CommonException ( "Invalid algorithm" );
		}

		measurements::end ( );
		measurements::end ( );

		if ( measure.getValue ( ) ) std::cmeasure << measurements::results ( ) << std::endl;

		return 0;
	} catch ( const exception::CommonException & exception ) {
		alib::XmlDataFactory::toStdout ( exception );
		return 1;
	} catch ( const TCLAP::ArgException & exception ) {
		std::cout << exception.error ( ) << std::endl;
		return 2;
	} catch ( const std::exception & exception ) {
		std::cerr << "Exception caught: " << exception.what ( ) << std::endl;
		return 3;
	} catch ( ... ) {
		std::cerr << "Unknown exception caught." << std::endl;
		return 127;
	}
}
